/*
 *  open_vznet.c
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

/*
 * Virtual Networking device used to change VE ownership on packets
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/seq_file.h>

#include <linux/inet.h>
#include <net/ip.h>
#include <linux/skbuff.h>
#include <linux/venet.h>

void veip_stop(struct ve_struct *ve)
{
	struct list_head *p, *tmp;

	write_lock_irq(&veip_hash_lock);
	if (ve->veip == NULL)
		goto unlock;
	list_for_each_safe(p, tmp, &ve->veip->ip_lh) {
		struct ip_entry_struct *ptr;
		ptr = list_entry(p, struct ip_entry_struct, ve_list);
		ptr->active_env = NULL;
		list_del(&ptr->ve_list);
		list_del(&ptr->ip_hash);
		kfree(ptr);
	}
	veip_put(ve->veip);
	ve->veip = NULL;
unlock:
	write_unlock_irq(&veip_hash_lock);
}

int veip_start(struct ve_struct *ve)
{
	int err;

	err = 0;
	write_lock_irq(&veip_hash_lock);
	ve->veip = veip_findcreate(ve->veid);
	if (ve->veip == NULL)
		err = -ENOMEM;
	write_unlock_irq(&veip_hash_lock);
	return err;
}

int veip_entry_add(struct ve_struct *ve, struct sockaddr_in *addr)
{
	struct ip_entry_struct *entry, *found;
	int err;

	entry = kmalloc(sizeof(struct ip_entry_struct), GFP_KERNEL);
	if (entry == NULL)
		return -ENOMEM;

	memset(entry, 0, sizeof(struct ip_entry_struct));
	entry->ip = addr->sin_addr.s_addr;

	write_lock_irq(&veip_hash_lock);
	err = -EADDRINUSE;
	found = ip_entry_lookup(entry->ip);
	if (found != NULL)
		goto out_unlock;
	else {
		ip_entry_hash(entry, ve->veip);
		found = entry;
		entry = NULL;
	}
	err = 0;
	found->active_env = ve;
out_unlock:
	write_unlock_irq(&veip_hash_lock);
	if (entry != NULL)
		kfree(entry);
	return err;
}

int veip_entry_del(envid_t veid, struct sockaddr_in *addr)
{
	struct ip_entry_struct *found;
	int err;

	err = -EADDRNOTAVAIL;
	write_lock_irq(&veip_hash_lock);
	found = ip_entry_lookup(addr->sin_addr.s_addr);
	if (found == NULL)
		goto out;
	if (found->active_env->veid != veid)
		goto out;

	err = 0;
	found->active_env = NULL;

	list_del(&found->ip_hash);
	list_del(&found->ve_list);
	kfree(found);
out:
	write_unlock_irq(&veip_hash_lock);
	return err;
}

static struct ve_struct *venet_find_ve(__u32 ip)
{
	struct ip_entry_struct *entry;

	entry = ip_entry_lookup(ip);
	if (entry == NULL)
		return NULL;

	return entry->active_env;
}

int venet_change_skb_owner(struct sk_buff *skb)
{
	struct ve_struct *ve, *ve_old;
	struct iphdr *iph;

	ve_old = skb->owner_env;
	iph = skb->nh.iph;

	read_lock(&veip_hash_lock);
	if (!ve_is_super(ve_old)) {
		/* from VE to host */
		ve = venet_find_ve(iph->saddr);
		if (ve == NULL)
			goto out_drop;
		if (!ve_accessible_strict(ve, ve_old))
			goto out_source;
		skb->owner_env = get_ve0();
	} else {
		/* from host to VE */
		ve = venet_find_ve(iph->daddr);
		if (ve == NULL)
			goto out_drop;
		skb->owner_env = ve;
	}
	read_unlock(&veip_hash_lock);

	return 0;

out_drop:
	read_unlock(&veip_hash_lock);
	return -ESRCH;

out_source:
	read_unlock(&veip_hash_lock);
	if (net_ratelimit()) {
		printk(KERN_WARNING "Dropped packet, source wrong "
		       "veid=%u src-IP=%u.%u.%u.%u "
		       "dst-IP=%u.%u.%u.%u\n",
		       skb->owner_env->veid,
		       NIPQUAD(skb->nh.iph->saddr),
		       NIPQUAD(skb->nh.iph->daddr));
	}
	return -EACCES;
}

#ifdef CONFIG_PROC_FS
int veip_seq_show(struct seq_file *m, void *v)
{
	struct list_head *p;
	struct ip_entry_struct *entry;
	char s[16];

	p = (struct list_head *)v;
	if (p == ip_entry_hash_table) {
		seq_puts(m, "Version: 2.5\n");
		return 0;
	}
	entry = list_entry(p, struct ip_entry_struct, ip_hash);
	sprintf(s, "%u.%u.%u.%u", NIPQUAD(entry->ip));
	seq_printf(m, "%15s %10u\n", s, 0);
	return 0;
}
#endif

MODULE_AUTHOR("SWsoft <info@sw-soft.com>");
MODULE_DESCRIPTION("Virtuozzo Virtual Network Device");
MODULE_LICENSE("GPL v2");
