/*
 *  venet_core.c
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

/*
 * Common part for Virtuozzo virtual network devices
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/tcp.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/unistd.h>

#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <net/ip.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/if_ether.h>	/* For the statistics structure. */
#include <linux/if_arp.h>	/* For ARPHRD_ETHER */
#include <linux/venet.h>
#include <linux/ve_proto.h>
#include <linux/vzctl.h>
#include <linux/vzctl_venet.h>

struct list_head ip_entry_hash_table[VEIP_HASH_SZ];
rwlock_t veip_hash_lock = RW_LOCK_UNLOCKED;
LIST_HEAD(veip_lh);

#define ip_entry_hash_function(ip)  (ntohl(ip) & (VEIP_HASH_SZ - 1))

void ip_entry_hash(struct ip_entry_struct *entry, struct veip_struct *veip)
{
	list_add(&entry->ip_hash,
		 ip_entry_hash_table + ip_entry_hash_function(entry->ip));
	list_add(&entry->ve_list, &veip->ip_lh);
}

void veip_put(struct veip_struct *veip)
{
	if (!list_empty(&veip->ip_lh))
		return;
	if (!list_empty(&veip->src_lh))
		return;
	if (!list_empty(&veip->dst_lh))
		return;

	list_del(&veip->list);
	kfree(veip);
}

struct ip_entry_struct *ip_entry_lookup(u32 addr)
{
	struct ip_entry_struct *entry;
	struct list_head *tmp;

	list_for_each(tmp, ip_entry_hash_table + ip_entry_hash_function(addr)) {
		entry = list_entry(tmp, struct ip_entry_struct, ip_hash);
		if (entry->ip != addr)
			continue;
		return entry;
	}
	return NULL;
}

struct veip_struct *veip_find(envid_t veid)
{
	struct veip_struct *ptr;
	list_for_each_entry(ptr, &veip_lh, list) {
		if (ptr->veid != veid)
			continue;
		return ptr;
	}
	return NULL;
}

struct veip_struct *veip_findcreate(envid_t veid)
{
	struct veip_struct *ptr;

	ptr = veip_find(veid);
	if (ptr != NULL)
		return ptr;

	ptr = kmalloc(sizeof(struct veip_struct), GFP_ATOMIC);
	if (ptr == NULL)
		return NULL;
	memset(ptr, 0, sizeof(struct veip_struct));
	INIT_LIST_HEAD(&ptr->ip_lh);
	INIT_LIST_HEAD(&ptr->src_lh);
	INIT_LIST_HEAD(&ptr->dst_lh);
	list_add(&ptr->list, &veip_lh);
	ptr->veid = veid;
	return ptr;
}

/*
 * Device functions
 */

static int venet_open(struct net_device *dev)
{
	if (!try_module_get(THIS_MODULE))
		return -EBUSY;
	return 0;
}

static int venet_close(struct net_device *master)
{
	module_put(THIS_MODULE);
	return 0;
}

static void venet_destructor(struct net_device *dev)
{
	kfree(dev->priv);
	dev->priv = NULL;
}

/*
 * The higher levels take care of making this non-reentrant (it's
 * called with bh's disabled).
 */
static int venet_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct net_device_stats *stats = (struct net_device_stats *)dev->priv;
	struct net_device *rcv = NULL;
	struct iphdr *iph;
	int length;

	/*
	 *	Optimise so buffers with skb->free=1 are not copied but
	 *	instead are lobbed from tx queue to rx queue
	 */
	if (atomic_read(&skb->users) != 1) {
	  	struct sk_buff *skb2 = skb;
	  	skb = skb_clone(skb, GFP_ATOMIC);	/* Clone the buffer */
	  	if (skb == NULL) {
			kfree_skb(skb2);
			goto out;
		}
	  	kfree_skb(skb2);
	} else
		skb_orphan(skb);

	if (skb->protocol != __constant_htons(ETH_P_IP))
		goto outf;

	iph = skb->nh.iph;
	if (MULTICAST(iph->daddr))
		goto outf;

	if (venet_change_skb_owner(skb) < 0)
		goto outf;

	rcv = VE_OWNER_SKB(skb)->_venet_dev;
	if (!rcv)
		/* VE going down */
		goto outf;

	dev_hold(rcv);

	if (!(rcv->flags & IFF_UP)) {
		/* Target VE does not want to receive packets */
		dev_put(rcv);
		goto outf;
	}

	skb->pkt_type = PACKET_HOST;
	skb->dev = rcv;

	skb->mac.raw = skb->data;
	memset(skb->data - dev->hard_header_len, 0, dev->hard_header_len);

	dst_release(skb->dst);
	skb->dst = NULL;
#ifdef CONFIG_NETFILTER
	nf_conntrack_put(skb->nfct);
	skb->nfct = NULL;
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif
#endif
	length = skb->len;

	netif_rx(skb);

	stats->tx_bytes += length;
	stats->tx_packets++;
	if (rcv) {
		struct net_device_stats *rcv_stats =
			(struct net_device_stats *)rcv->priv;
		rcv_stats->rx_bytes += length;
		rcv_stats->rx_packets++;
		dev_put(rcv);
	}

	return 0;

outf:
	kfree_skb(skb);
	++stats->tx_dropped;
out:
	return 0;
}

static struct net_device_stats *get_stats(struct net_device *dev)
{
	return (struct net_device_stats *)dev->priv;
}

/* Initialize the rest of the LOOPBACK device. */
int venet_init_dev(struct net_device *dev)
{
	dev->hard_start_xmit	= venet_xmit;
	dev->priv = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
	if (dev->priv == NULL)
		return -ENOMEM;
	memset(dev->priv, 0, sizeof(struct net_device_stats));
	dev->get_stats = get_stats;
	dev->open = venet_open;
	dev->stop = venet_close;
	dev->destructor = venet_destructor;

	/*
	 *	Fill in the generic fields of the device structure.
	 */
	dev->type		= ARPHRD_VOID;
	dev->hard_header_len 	= ETH_HLEN;
	dev->mtu		= 1500; /* eth_mtu */
	dev->tx_queue_len	= 0;

	memset(dev->broadcast, 0xFF, ETH_ALEN);

	/* New-style flags. */
	dev->flags		= IFF_BROADCAST|IFF_NOARP|IFF_POINTOPOINT;
	return 0;
}

static void venet_setup(struct net_device *dev)
{
	dev->init = venet_init_dev;
	/*
	 * No other features, as they are:
	 *  - checksumming is required, and nobody else will done our job
	 */
	dev->features |= NETIF_F_VENET | NETIF_F_VIRTUAL;
}

#ifdef CONFIG_PROC_FS
static int veinfo_seq_show(struct seq_file *m, void *v)
{
	struct ve_struct *ve = (struct ve_struct *)v;
	struct list_head *tmp;

	seq_printf(m, "%10u %5u %5u", ve->veid,
                                ve->class_id, atomic_read(&ve->pcounter));
	read_lock(&veip_hash_lock);
	if (ve->veip == NULL)
		goto unlock;
	list_for_each(tmp, &ve->veip->ip_lh) {
		char ip[16];
		struct ip_entry_struct *entry;

		entry = list_entry(tmp, struct ip_entry_struct, ve_list);
		if (entry->active_env == NULL)
			continue;

		sprintf(ip, "%u.%u.%u.%u", NIPQUAD(entry->ip));
		seq_printf(m, " %15s", ip);
	}
unlock:
	read_unlock(&veip_hash_lock);
	seq_putc(m, '\n');
	return 0;
}

static void *ve_seq_start(struct seq_file *m, loff_t *pos)
{
	struct ve_struct *ve, *curve;
	loff_t l;

	curve = get_exec_env();
	read_lock(&ve_list_guard);
	if (!ve_is_super(curve)) {
		if (*pos != 0)
			return NULL;
		return curve;
	}
	for (ve = ve_list_head, l = *pos;
	     ve != NULL && l > 0;
	     ve = ve->next, l--);
	return ve;
}

static void *ve_seq_next(struct seq_file *m, void *v, loff_t *pos)
{
	struct ve_struct *ve = (struct ve_struct *)v;

	if (!ve_is_super(get_exec_env()))
		return NULL;
	(*pos)++;
	return ve->next;
}

static void ve_seq_stop(struct seq_file *m, void *v)
{
	read_unlock(&ve_list_guard);
}


static struct seq_operations veinfo_seq_op = {
        start:  ve_seq_start,
        next:   ve_seq_next,
        stop:   ve_seq_stop,
        show:   veinfo_seq_show
};

static int veinfo_open(struct inode *inode, struct file *file)
{
        return seq_open(file, &veinfo_seq_op);
}

static struct file_operations proc_veinfo_operations = {
        open:           veinfo_open,
        read:           seq_read,
        llseek:         seq_lseek,
        release:        seq_release
};

static void *veip_seq_start(struct seq_file *m, loff_t *pos)
{
	loff_t l;
	struct list_head *p;
	int i;

	l = *pos;
	write_lock_irq(&veip_hash_lock);
	if (l == 0)
		return ip_entry_hash_table;
	for (i = 0; i < VEIP_HASH_SZ; i++) {
		list_for_each(p, ip_entry_hash_table + i) {
			if (--l == 0)
				return p;
		}
	}
	return NULL;
}

static void *veip_seq_next(struct seq_file *m, void *v, loff_t *pos)
{
	struct list_head *p;

	p = (struct list_head *)v;
	while (1) {
		p = p->next;
		if (p < ip_entry_hash_table ||
		    p >= ip_entry_hash_table + VEIP_HASH_SZ) {
			(*pos)++;
			return p;
		}
		if (++p >= ip_entry_hash_table + VEIP_HASH_SZ)
			return NULL;
	}
	return NULL;
}

static void veip_seq_stop(struct seq_file *m, void *v)
{
	write_unlock_irq(&veip_hash_lock);
}

static struct seq_operations veip_seq_op = {
        start:  veip_seq_start,
        next:   veip_seq_next,
        stop:   veip_seq_stop,
        show:   veip_seq_show
};

static int veip_open(struct inode *inode, struct file *file)
{
        return seq_open(file, &veip_seq_op);
}

static struct file_operations proc_veip_operations = {
        open:           veip_open,
        read:           seq_read,
        llseek:         seq_lseek,
        release:        seq_release
};
#endif

int real_ve_ip_map(envid_t veid, int op, struct sockaddr *uservaddr, int addrlen)
{
	int err;
	struct sockaddr_in addr;
	struct ve_struct *ve;

	err = -EPERM;
	if (!capable(CAP_SETVEID))
		goto out;

	err = -EINVAL;
	if (addrlen != sizeof(struct sockaddr_in))
		goto out;

	err = move_addr_to_kernel(uservaddr, addrlen, &addr);
	if (err < 0)
		goto out;

	switch (op)
	{
		case VE_IP_ADD:
			ve = get_ve_by_id(veid);
			err = -ESRCH;
			if (!ve)
				goto out;

			down_read(&ve->op_sem);
			if (ve->is_running)
				err = veip_entry_add(ve, &addr);
			up_read(&ve->op_sem);
			put_ve(ve);
			break;

		case VE_IP_DEL:
			err = veip_entry_del(veid, &addr);
			break;
		default:
			err = -EINVAL;
	}

out:
	return err;
}

int venet_ioctl(struct inode *ino, struct file *file, unsigned int cmd,
		unsigned long arg)
{
	int err;

	err = -ENOTTY;
	switch(cmd) {
	    case VENETCTL_VE_IP_MAP: {
			struct vzctl_ve_ip_map s;
			err = -EFAULT;
			if (copy_from_user(&s, (void *)arg, sizeof(s)))
				break;
			err = real_ve_ip_map(s.veid, s.op, s.addr, s.addrlen);
		}
		break;
	}
	return err;
}

static struct vzioctlinfo venetcalls = {
	type: VENETCTLTYPE,
	func: venet_ioctl,
	owner: THIS_MODULE,
};

int venet_dev_start(struct ve_struct *env)
{
	struct net_device *dev_venet;
	int err;

	dev_venet = alloc_netdev(0, "venet%d", venet_setup);
	if (!dev_venet)
		return -ENOMEM;
	err = dev_alloc_name(dev_venet, dev_venet->name);
	if (err<0)
		goto err;
	if ((err = register_netdev(dev_venet)) != 0)
		goto err;
	env->_venet_dev = dev_venet;
	return 0;
err:
	free_netdev(dev_venet);
	printk(KERN_ERR "VENET initialization error err=%d\n", err);
	return err;
}

static int venet_start(unsigned int hooknum, void *data)
{
	struct ve_struct *env;
	int err;

	env = (struct ve_struct *)data;
	if (env->veip)
		return -EEXIST;
	if (!ve_is_super(env) && !try_module_get(THIS_MODULE))
		return 0;

	err = veip_start(env);
	if (err)
		goto err;

	err = venet_dev_start(env);
	if (err)
		goto err_free;
	return 0;

err_free:
	veip_stop(env);
err:
	if (!ve_is_super(env))
		module_put(THIS_MODULE);
	return err;
}

static int venet_stop(unsigned int hooknum, void *data)
{
	struct ve_struct *env;

	env = (struct ve_struct *)data;
	veip_stop(env);
	if (!ve_is_super(env))
		module_put(THIS_MODULE);
	return 0;
}

#define VE_HOOK_PRI_NET		0

static struct ve_hook venet_ve_hook_init = {
	hook:	venet_start,
	undo:	venet_stop,
	hooknum: VE_HOOK_INIT,
	priority: VE_HOOK_PRI_NET
};

static struct ve_hook venet_ve_hook_fini = {
	hook:	venet_stop,
	hooknum: VE_HOOK_FINI,
	priority: VE_HOOK_PRI_NET
};

__init int venet_init(void)
{
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *de;
#endif
	int i, err;

	if (get_ve0()->_venet_dev != NULL)
		return -EEXIST;

	for (i = 0; i < VEIP_HASH_SZ; i++)
		INIT_LIST_HEAD(ip_entry_hash_table + i);

	err = venet_start(VE_HOOK_INIT, (void *)get_ve0());
	if (err)
		return err;

#ifdef CONFIG_PROC_FS
	de = create_proc_glob_entry("vz/veinfo",
			S_IFREG|S_IRUSR, NULL);
	if (de)
		de->proc_fops = &proc_veinfo_operations;
	else
		printk(KERN_WARNING "venet: can't make veinfo proc entry\n");

	de = create_proc_entry("vz/veip", S_IFREG|S_IRUSR, NULL);
	if (de)
		de->proc_fops = &proc_veip_operations;
	else
		printk(KERN_WARNING "venet: can't make veip proc entry\n");
#endif

	ve_hook_register(&venet_ve_hook_init);
	ve_hook_register(&venet_ve_hook_fini);
	vzioctl_register(&venetcalls);
	return 0;
}

__exit void venet_exit(void)
{
	struct net_device *dev_venet;

	vzioctl_unregister(&venetcalls);
	ve_hook_unregister(&venet_ve_hook_fini);
	ve_hook_unregister(&venet_ve_hook_init);
#ifdef CONFIG_PROC_FS
	remove_proc_entry("vz/veip", NULL);
	remove_proc_entry("vz/veinfo", NULL);
#endif

	dev_venet = get_ve0()->_venet_dev;
	if (dev_venet != NULL) {
		get_ve0()->_venet_dev = NULL;
		unregister_netdev(dev_venet);
		free_netdev(dev_venet);
	}
	veip_stop(get_ve0());
}

module_init(venet_init);
module_exit(venet_exit);
