/*
 *  fs/simfs.c
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/config.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/init.h>
#include <linux/namei.h>
#include <linux/err.h>
#include <linux/module.h>
#include <linux/mount.h>
#include <linux/vzquota.h>
#include <linux/statfs.h>
#include <linux/virtinfo.h>
#include <linux/faudit.h>
#include <linux/genhd.h>

#include <asm/unistd.h>
#include <asm/uaccess.h>

#define SIMFS_GET_LOWER_FS_SB(sb) sb->s_root->d_sb

static struct super_operations sim_super_ops;

static int sim_getattr(struct vfsmount *mnt, struct dentry *dentry,
		struct kstat *stat)
{
	struct super_block *sb;
	struct inode *inode;

	inode = dentry->d_inode;
	if (!inode->i_op->getattr) {
		generic_fillattr(inode, stat);
		if (!stat->blksize) {
			unsigned blocks;

			sb = inode->i_sb;
			blocks = (stat->size + sb->s_blocksize-1) >>
				sb->s_blocksize_bits;
			stat->blocks = (sb->s_blocksize / 512) * blocks;
			stat->blksize = sb->s_blocksize;
		}
	} else {
		int err;

		err = inode->i_op->getattr(mnt, dentry, stat);
		if (err)
			return err;
	}

	sb = mnt->mnt_sb;
	if (sb->s_op == &sim_super_ops)
		stat->dev = sb->s_dev;
	return 0;
}

static void quota_get_stat(struct super_block *sb, struct kstatfs *buf)
{
	int err;
	struct dq_stat qstat;
	struct virt_info_quota q;
	long free_file, adj_file;
	s64 blk, free_blk, adj_blk;
	int bsize_bits;

	q.super = sb;
	q.qstat = &qstat;
	err = virtinfo_notifier_call(VITYPE_QUOTA, VIRTINFO_QUOTA_GETSTAT, &q);
	if (err != NOTIFY_OK)
		return;

	bsize_bits = ffs(buf->f_bsize) - 1;
	free_blk = (s64)(qstat.bsoftlimit - qstat.bcurrent) >> bsize_bits;
	if (free_blk < 0)
		free_blk = 0;
	/*
	 * In the regular case, we always set buf->f_bfree and buf->f_blocks to
	 * the values reported by quota.  In case of real disk space shortage,
	 * we adjust the values.  We want this adjustment to look as if the
	 * total disk space were reduced, not as if the usage were increased.
	 *    -- SAW
	 */
	adj_blk = 0;
	if (buf->f_bfree < free_blk)
		adj_blk = free_blk - buf->f_bfree;
	buf->f_bfree = (long)(free_blk - adj_blk);

	if (free_blk < buf->f_bavail)
		buf->f_bavail = (long)free_blk; /* min(f_bavail, free_blk) */

	blk = (qstat.bsoftlimit >> bsize_bits) - adj_blk;
	buf->f_blocks = blk > LONG_MAX ? LONG_MAX : blk;

	free_file = qstat.isoftlimit - qstat.icurrent;
	if (free_file < 0)
		free_file = 0;
	if (buf->f_ffree == -1)
		/*
		 * One filesystem uses -1 to represent the fact that it doesn't
		 * have a detached limit for inode number.
		 * May be, because -1 is a good pretendent for the maximum value
		 * of signed long type, may be, because it's just nice to have
		 * an exceptional case...  Guess what that filesystem is :-)
		 *    -- SAW
		 */
		buf->f_ffree = free_file;
	adj_file = 0;
	if (buf->f_ffree < free_file)
		adj_file = free_file - buf->f_ffree;
	buf->f_ffree = free_file - adj_file;
	buf->f_files = qstat.isoftlimit - adj_file;
}

static int sim_statfs(struct super_block *sb, struct statfs *buf)
{
	int err;
	struct super_block *lsb;
	struct kstatfs statbuf;

	err = 0;
	if (sb->s_op != &sim_super_ops)
		goto out;

	lsb = SIMFS_GET_LOWER_FS_SB(sb);

	err = -ENOSYS;
	if (lsb && lsb->s_op && lsb->s_op->statfs)
		err = lsb->s_op->statfs(lsb, &statbuf);
	if (err)
		goto out;

	quota_get_stat(sb, &statbuf);

	buf->f_files    = statbuf.f_files;
	buf->f_ffree    = statbuf.f_ffree;
	buf->f_blocks   = statbuf.f_blocks;
	buf->f_bfree    = statbuf.f_bfree;
	buf->f_bavail   = statbuf.f_bavail;
out:
	return err;
}

static int sim_statfs64(struct super_block *sb, struct statfs64 *buf)
{
	int err;
	struct super_block *lsb;
	struct kstatfs statbuf;

	err = 0;
	if (sb->s_op != &sim_super_ops)
		goto out;

	lsb = SIMFS_GET_LOWER_FS_SB(sb);

	err = -ENOSYS;
	if (lsb && lsb->s_op && lsb->s_op->statfs)
		err = lsb->s_op->statfs(lsb, &statbuf);
	if (err)
		goto out;

	quota_get_stat(sb, &statbuf);

	buf->f_files    = (__u64)statbuf.f_files;
	buf->f_ffree    = (__u64)statbuf.f_ffree;
	buf->f_blocks   = (__u64)statbuf.f_blocks;
	buf->f_bfree    = (__u64)statbuf.f_bfree;
	buf->f_bavail   = (__u64)statbuf.f_bavail;
out:
	return err;
}

static int sim_systemcall(struct vnotifier_block *me, unsigned long n,
		void *d, int old_ret)
{
	int err;
	struct faudit_stat_arg *arg;

	arg = (struct faudit_stat_arg *)d;
	switch (n) {
	case VIRTINFO_FAUDIT_STAT:
		err = sim_getattr(arg->mnt, arg->dentry,
				(struct kstat *)arg->stat);
		break;
	case VIRTINFO_FAUDIT_STATFS:
		err = sim_statfs(arg->mnt->mnt_sb,
				(struct statfs *)arg->stat);
		break;
	case VIRTINFO_FAUDIT_STATFS64:
		err = sim_statfs64(arg->mnt->mnt_sb,
				(struct statfs64 *)arg->stat);
		break;
	default:
		return old_ret;
	}
	arg->err = err;
	return (err ? NOTIFY_BAD : NOTIFY_OK);
}

static struct inode *sim_quota_root(struct super_block *sb)
{
	return sb->s_root->d_inode;
}

void sim_put_super(struct super_block *sb)
{
	struct virt_info_quota viq;

	viq.super = sb;
	virtinfo_notifier_call(VITYPE_QUOTA, VIRTINFO_QUOTA_OFF, &viq);
	bdput(sb->s_bdev);
}

static struct super_operations sim_super_ops = {
	.get_quota_root	= sim_quota_root,
	.put_super = sim_put_super,
};

static int sim_fill_super(struct super_block *s, void *data)
{
	int err;
	struct nameidata *nd;

	err = set_anon_super(s, NULL);
	if (err)
		goto out;

	err = 0;
	nd = (struct nameidata *)data;
	s->s_root = dget(nd->dentry);
	s->s_op = &sim_super_ops;
out:
	return err;
}

struct super_block *sim_get_sb(struct file_system_type *type,
		int flags, const char *dev_name, void *opt)
{
	int err;
	struct nameidata nd;
	struct super_block *sb;
	struct block_device *bd;
	struct virt_info_quota viq;
	static struct hd_struct fake_hds;

	sb = ERR_PTR(-EINVAL);
	if (opt == NULL)
		goto out;

	err = path_lookup(opt, LOOKUP_FOLLOW|LOOKUP_DIRECTORY, &nd);
	sb = ERR_PTR(err);
	if (err)
		goto out;

	sb = sget(type, NULL, sim_fill_super, &nd);
	if (IS_ERR(sb))
		goto out_path;

	bd = bdget(sb->s_dev);
	if (!bd)
		goto out_killsb;

	sb->s_bdev = bd;
	bd->bd_part = &fake_hds;
	viq.super = sb;
	virtinfo_notifier_call(VITYPE_QUOTA, VIRTINFO_QUOTA_ON, &viq);
out_path:
	path_release(&nd);
out:
	return sb;

out_killsb:
	up_write(&sb->s_umount);
	deactivate_super(sb);
	sb = ERR_PTR(-ENODEV);
	goto out_path;
}

static struct file_system_type sim_fs_type = {
	.owner		= THIS_MODULE,
	.name		= "simfs",
	.get_sb		= sim_get_sb,
	.kill_sb	= kill_anon_super,
};

static struct vnotifier_block sim_syscalls = {
	.notifier_call = sim_systemcall,
};

static int __init init_simfs(void)
{
	int err;

	err = register_filesystem(&sim_fs_type);
	if (err)
		return err;

	virtinfo_notifier_register(VITYPE_FAUDIT, &sim_syscalls);
	return 0;
}

static void __exit exit_simfs(void)
{
	virtinfo_notifier_unregister(VITYPE_FAUDIT, &sim_syscalls);
	unregister_filesystem(&sim_fs_type);
}

MODULE_AUTHOR("SWsoft <info@sw-soft.com>");
MODULE_DESCRIPTION("Open Virtuozzo Simulation of File System");
MODULE_LICENSE("GPL v2");

module_init(init_simfs);
module_exit(exit_simfs);
