/*
 * Copyright (C) 2002 SWsoft
 * All rights reserved.
 *
 * Licensing governed by "linux/COPYING.SWsoft" file.
 *
 * This file contains Virtuozzo UID/GID disk quota implementation
 */

#include <linux/config.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <linux/rcupdate.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/quota.h>
#include <linux/quotaio_v2.h>
#include <linux/virtinfo.h>

#include <linux/vzctl.h>
#include <linux/vzctl_quota.h>
#include <linux/vzquota.h>

/*
 * XXX
 * may be something is needed for sb->s_dquot->info[]?
 */

#define USRQUOTA_MASK		(1 << USRQUOTA)
#define GRPQUOTA_MASK		(1 << GRPQUOTA)
#define QTYPE2MASK(type)	(1 << (type))

static kmem_cache_t *vz_quota_ugid_cachep;

/* guard to protect vz_quota_master from destroy in quota_on/off. Also protects
 * list on the hash table */
extern struct semaphore vz_quota_sem;

inline struct vz_quota_ugid *vzquota_get_ugid(struct vz_quota_ugid *qugid)
{
	if (qugid != VZ_QUOTA_UGBAD)
		atomic_inc(&qugid->qugid_count);
	return qugid;
}

/* we don't limit users with zero limits */
static inline int vzquota_fake_stat(struct dq_stat *stat)
{
	return stat->bhardlimit == 0 && stat->bsoftlimit == 0 &&
		stat->ihardlimit == 0 && stat->isoftlimit == 0;
}

/* callback function for quotatree_free() */
static inline void vzquota_free_qugid(void *ptr)
{
	kmem_cache_free(vz_quota_ugid_cachep, ptr);
}

/*
 * destroy ugid, if it have zero refcount, limits and usage
 * must be called under qmblk->dq_sem
 */
void vzquota_put_ugid(struct vz_quota_master *qmblk,
		struct vz_quota_ugid *qugid)
{
	if (qugid == VZ_QUOTA_UGBAD)
		return;
	qmblk_data_read_lock(qmblk);
	if (atomic_dec_and_test(&qugid->qugid_count) &&
	    (qmblk->dq_flags & VZDQUG_FIXED_SET) == 0 &&
	    vzquota_fake_stat(&qugid->qugid_stat) &&
	    qugid->qugid_stat.bcurrent == 0 &&
	    qugid->qugid_stat.icurrent == 0) {
		quotatree_remove(QUGID_TREE(qmblk, qugid->qugid_type),
				qugid->qugid_id);
		qmblk->dq_ugid_count--;
		vzquota_free_qugid(qugid);
	}
	qmblk_data_read_unlock(qmblk);
}

/*
 * Get ugid block by its index, like it would present in array.
 * In reality, this is not array - this is leafs chain of the tree.
 * NULL if index is out of range.
 * qmblk semaphore is required to protect the tree.
 */
static inline struct vz_quota_ugid *
vzquota_get_byindex(struct vz_quota_master *qmblk, unsigned int index, int type)
{
	return quotatree_leaf_byindex(QUGID_TREE(qmblk, type), index);
}

/*
 * get next element from ugid "virtual array"
 * ugid must be in current array and this array may not be changed between
 * two accesses (quaranteed by "stopped" quota state and quota semaphore)
 * qmblk semaphore is required to protect the tree
 */
static inline struct vz_quota_ugid *
vzquota_get_next(struct vz_quota_master *qmblk, struct vz_quota_ugid *qugid)
{
	return quotatree_get_next(QUGID_TREE(qmblk, qugid->qugid_type),
			qugid->qugid_id);
}

/*
 * requires dq_sem
 */
struct vz_quota_ugid *__vzquota_find_ugid(struct vz_quota_master *qmblk,
			unsigned int quota_id, int type, int flags)
{
	struct vz_quota_ugid *qugid;
	struct quotatree_tree *tree;
	struct quotatree_find_state st;

	tree = QUGID_TREE(qmblk, type);
	qugid = quotatree_find(tree, quota_id, &st);
	if (qugid)
		goto success;

	/* caller does not want alloc */
	if (flags & VZDQUG_FIND_DONT_ALLOC)
		goto fail;

	if (flags & VZDQUG_FIND_FAKE)
		goto doit;

	/* check limit */
	if (qmblk->dq_ugid_count >= qmblk->dq_ugid_max)
		goto fail;

	/* see comment at VZDQUG_FIXED_SET define */
	if (qmblk->dq_flags & VZDQUG_FIXED_SET)
		goto fail;

doit:
	/* alloc new structure */
	qugid = kmem_cache_alloc(vz_quota_ugid_cachep,
			SLAB_NOFS | __GFP_NOFAIL);
	if (qugid == NULL)
		goto fail;

	/* initialize new structure */
	qugid->qugid_id = quota_id;
	memset(&qugid->qugid_stat, 0, sizeof(qugid->qugid_stat));
	qugid->qugid_type = type;
	atomic_set(&qugid->qugid_count, 0);

	/* insert in tree */
	if (quotatree_insert(tree, quota_id, &st, qugid) < 0)
		goto fail_insert;
	qmblk->dq_ugid_count++;

success:
	vzquota_get_ugid(qugid);
	return qugid;

fail_insert:
	vzquota_free_qugid(qugid);
fail:
	return VZ_QUOTA_UGBAD;
}

/*
 * takes dq_sem, may schedule
 */
struct vz_quota_ugid *vzquota_find_ugid(struct vz_quota_master *qmblk,
			unsigned int quota_id, int type, int flags)
{
	struct vz_quota_ugid *qugid;

	down(&qmblk->dq_sem);
	qugid = __vzquota_find_ugid(qmblk, quota_id, type, flags);
	up(&qmblk->dq_sem);

	return qugid;
}

/*
 * destroy all ugid records on given quota master
 */
void vzquota_kill_ugid(struct vz_quota_master *qmblk)
{
	BUG_ON((qmblk->dq_gid_tree == NULL && qmblk->dq_uid_tree != NULL) ||
		(qmblk->dq_uid_tree == NULL && qmblk->dq_gid_tree != NULL));

	if (qmblk->dq_uid_tree != NULL) {
		quotatree_free(qmblk->dq_uid_tree, vzquota_free_qugid);
		quotatree_free(qmblk->dq_gid_tree, vzquota_free_qugid);
	}
}


/* ----------------------------------------------------------------------
 * Management interface to ugid quota for (super)users.
 * --------------------------------------------------------------------- */

/**
 * vzquota_find_qmblk - helper to emulate quota on virtual filesystems
 *
 * This function finds a quota master block corresponding to the root of
 * a virtual filesystem.
 * Returns a quota master block with reference taken, or %NULL if not under
 * quota, or %VZ_QUOTA_BAD if quota inconsistency is found (and all allocation
 * operations will fail).
 *
 * Note: this function uses vzquota_inode_qmblk().
 * The latter is a rather confusing function: it returns qmblk that used to be
 * on the inode some time ago (without guarantee that it still has any
 * relations to the inode).  So, vzquota_find_qmblk() leaves it up to the
 * caller to think whether the inode could have changed its qmblk and what to
 * do in that case.
 * Currently, the callers appear to not care :(
 */
struct vz_quota_master *vzquota_find_qmblk(struct super_block *sb)
{
	struct inode *qrinode;
	struct vz_quota_master *qmblk;

	qmblk = NULL;
	qrinode = NULL;
	if (sb->s_op->get_quota_root != NULL)
		qrinode = sb->s_op->get_quota_root(sb);
	if (qrinode != NULL)
		qmblk = vzquota_inode_qmblk(qrinode);
	return qmblk;
}

static int vzquota_initialize2(struct inode *inode, int type)
{
	return QUOTA_OK;
}

static int vzquota_drop2(struct inode *inode)
{
	return QUOTA_OK;
}

static int vzquota_alloc_space2(struct inode *inode,
			     qsize_t number, int prealloc)
{
	inode_add_bytes(inode, number);
	return QUOTA_OK;
}

static int vzquota_alloc_inode2(const struct inode *inode, unsigned long number)
{
	return QUOTA_OK;
}

static int vzquota_free_space2(struct inode *inode, qsize_t number)
{
	inode_sub_bytes(inode, number);
	return QUOTA_OK;
}

static int vzquota_free_inode2(const struct inode *inode, unsigned long number)
{
	return QUOTA_OK;
}

static int vzquota_transfer2(struct inode *inode, struct iattr *iattr)
{
	return QUOTA_OK;
}

struct dquot_operations vz_quota_operations2 = {
	initialize:	vzquota_initialize2,
	drop:		vzquota_drop2,
	alloc_space:	vzquota_alloc_space2,
	alloc_inode:	vzquota_alloc_inode2,
	free_space:	vzquota_free_space2,
	free_inode:	vzquota_free_inode2,
	transfer:	vzquota_transfer2
};

static int vz_quota_on(struct super_block *sb, int type,
		int format_id, char *path)
{
	struct vz_quota_master *qmblk;
	int mask, mask2;
	int err;

	qmblk = vzquota_find_qmblk(sb);
	down(&vz_quota_sem);
	err = -ESRCH;
	if (qmblk == NULL)
		goto out;
	err = -EIO;
	if (qmblk == VZ_QUOTA_BAD)
		goto out;

	mask = 0;
	mask2 = 0;
	sb->dq_op = &vz_quota_operations2;
	sb->s_qcop = &vz_quotactl_operations;
	if (type == USRQUOTA) {
		mask = DQUOT_USR_ENABLED;
		mask2 = VZDQ_USRQUOTA;
	}
	if (type == GRPQUOTA) {
		mask = DQUOT_GRP_ENABLED;
		mask2 = VZDQ_GRPQUOTA;
	}
	err = -EBUSY;
	if (qmblk->dq_flags & mask2)
		goto out;

	err = 0;
	qmblk->dq_flags |= mask2;
	sb->s_dquot.flags |= mask;

out:
	up(&vz_quota_sem);
	if (qmblk != NULL && qmblk != VZ_QUOTA_BAD)
		qmblk_put(qmblk);
	return err;
}

static int vz_quota_off(struct super_block *sb, int type)
{
	struct vz_quota_master *qmblk;
	int mask2;
	int err;

	qmblk = vzquota_find_qmblk(sb);
	down(&vz_quota_sem);
	err = -ESRCH;
	if (qmblk == NULL)
		goto out;
	err = -EIO;
	if (qmblk == VZ_QUOTA_BAD)
		goto out;

	mask2 = 0;
	if (type == USRQUOTA)
		mask2 = VZDQ_USRQUOTA;
	if (type == GRPQUOTA)
		mask2 = VZDQ_GRPQUOTA;
	err = -EINVAL;
	if (!(qmblk->dq_flags & mask2))
		goto out;

	qmblk->dq_flags &= ~mask2;
	err = 0;

out:
	up(&vz_quota_sem);
	if (qmblk != NULL && qmblk != VZ_QUOTA_BAD)
		qmblk_put(qmblk);
	return err;
}

static int vz_quota_sync(struct super_block *sb, int type)
{
	return 0;	/* vz quota is always uptodate */
}

static int vz_get_dqblk(struct super_block *sb, int type,
		qid_t id, struct if_dqblk *di)
{
	struct vz_quota_master *qmblk;
	struct vz_quota_ugid *ugid;
	int err;

	qmblk = vzquota_find_qmblk(sb);
	down(&vz_quota_sem);
	err = -ESRCH;
	if (qmblk == NULL)
		goto out;
	err = -EIO;
	if (qmblk == VZ_QUOTA_BAD)
		goto out;

	err = 0;
	ugid = vzquota_find_ugid(qmblk, id, type, VZDQUG_FIND_DONT_ALLOC);
	if (ugid != VZ_QUOTA_UGBAD) {
		qmblk_data_read_lock(qmblk);
		di->dqb_bhardlimit = ugid->qugid_stat.bhardlimit >> 10;
		di->dqb_bsoftlimit = ugid->qugid_stat.bsoftlimit >> 10;
		di->dqb_curspace = ugid->qugid_stat.bcurrent;
		di->dqb_ihardlimit = ugid->qugid_stat.ihardlimit;
		di->dqb_isoftlimit = ugid->qugid_stat.isoftlimit;
		di->dqb_curinodes = ugid->qugid_stat.icurrent;
		di->dqb_btime = ugid->qugid_stat.btime;
		di->dqb_itime = ugid->qugid_stat.itime;
		qmblk_data_read_unlock(qmblk);
		di->dqb_valid = QIF_ALL;
		vzquota_put_ugid(qmblk, ugid);
	} else {
		memset(di, 0, sizeof(*di));
		di->dqb_valid = QIF_ALL;
	}

out:
	up(&vz_quota_sem);
	if (qmblk != NULL && qmblk != VZ_QUOTA_BAD)
		qmblk_put(qmblk);
	return err;
}

/* must be called under vz_quota_sem */
static int __vz_set_dqblk(struct vz_quota_master *qmblk,
		int type, qid_t id, struct if_dqblk *di)
{
	struct vz_quota_ugid *ugid;

	ugid = vzquota_find_ugid(qmblk, id, type, 0);
	if (ugid == VZ_QUOTA_UGBAD)
		return -ESRCH;

	qmblk_data_write_lock(qmblk);
	/*
	 * Subtle compatibility breakage.
	 *
	 * Some old non-vz kernel quota didn't start grace period
	 * if the new soft limit happens to be below the usage.
	 * Non-vz kernel quota in 2.4.20 starts the grace period
	 * (if it hasn't been started).
	 * Current non-vz kernel performs even more complicated
	 * manipulations...
	 *
	 * Also, current non-vz kernels have inconsistency related to
	 * the grace time start.  In regular operations the grace period
	 * is started if the usage is greater than the soft limit (and,
	 * strangely, is cancelled if the usage is less).
	 * However, set_dqblk starts the grace period if the usage is greater
	 * or equal to the soft limit.
	 *
	 * Here we try to mimic the behavior of the current non-vz kernel.
	 */
	if (di->dqb_valid & QIF_BLIMITS) {
		ugid->qugid_stat.bhardlimit =
			(__u64)di->dqb_bhardlimit << 10;
		ugid->qugid_stat.bsoftlimit =
			(__u64)di->dqb_bsoftlimit << 10;
		if (di->dqb_bsoftlimit == 0 ||
		    ugid->qugid_stat.bcurrent < ugid->qugid_stat.bsoftlimit)
			ugid->qugid_stat.btime = 0;
		else if (!(di->dqb_valid & QIF_BTIME))
			ugid->qugid_stat.btime = CURRENT_TIME_SECONDS
				+ qmblk->dq_ugid_info[type].bexpire;
		else
			ugid->qugid_stat.btime = di->dqb_btime;
	}
	if (di->dqb_valid & QIF_ILIMITS) {
		ugid->qugid_stat.ihardlimit = di->dqb_ihardlimit;
		ugid->qugid_stat.isoftlimit = di->dqb_isoftlimit;
		if (di->dqb_isoftlimit == 0 ||
		    ugid->qugid_stat.icurrent < ugid->qugid_stat.isoftlimit)
			ugid->qugid_stat.itime = 0;
		else if (!(di->dqb_valid & QIF_ITIME))
			ugid->qugid_stat.itime = CURRENT_TIME_SECONDS
				+ qmblk->dq_ugid_info[type].iexpire;
		else
			ugid->qugid_stat.itime = di->dqb_itime;
	}
	qmblk_data_write_unlock(qmblk);
	vzquota_put_ugid(qmblk, ugid);

	return 0;
}

static int vz_set_dqblk(struct super_block *sb, int type,
		qid_t id, struct if_dqblk *di)
{
	struct vz_quota_master *qmblk;
	int err;

	qmblk = vzquota_find_qmblk(sb);
	down(&vz_quota_sem);
	err = -ESRCH;
	if (qmblk == NULL)
		goto out;
	err = -EIO;
	if (qmblk == VZ_QUOTA_BAD)
		goto out;
	err = __vz_set_dqblk(qmblk, type, id, di);
out:
	up(&vz_quota_sem);
	if (qmblk != NULL && qmblk != VZ_QUOTA_BAD)
		qmblk_put(qmblk);
	return err;
}

static int vz_get_dqinfo(struct super_block *sb, int type,
		struct if_dqinfo *ii)
{
	struct vz_quota_master *qmblk;
	int err;

	qmblk = vzquota_find_qmblk(sb);
	down(&vz_quota_sem);
	err = -ESRCH;
	if (qmblk == NULL)
		goto out;
	err = -EIO;
	if (qmblk == VZ_QUOTA_BAD)
		goto out;

	err = 0;
	ii->dqi_bgrace = qmblk->dq_ugid_info[type].bexpire;
	ii->dqi_igrace = qmblk->dq_ugid_info[type].iexpire;
	ii->dqi_flags = 0;
	ii->dqi_valid = IIF_ALL;

out:
	up(&vz_quota_sem);
	if (qmblk != NULL && qmblk != VZ_QUOTA_BAD)
		qmblk_put(qmblk);
	return err;
}

/* must be called under vz_quota_sem */
static int __vz_set_dqinfo(struct vz_quota_master *qmblk,
		int type, struct if_dqinfo *ii)
{
	if (ii->dqi_valid & IIF_FLAGS)
		if (ii->dqi_flags & DQF_MASK)
			return -EINVAL;

	if (ii->dqi_valid & IIF_BGRACE)
		qmblk->dq_ugid_info[type].bexpire = ii->dqi_bgrace;
	if (ii->dqi_valid & IIF_IGRACE)
		qmblk->dq_ugid_info[type].iexpire = ii->dqi_igrace;
	return 0;
}

static int vz_set_dqinfo(struct super_block *sb, int type,
		struct if_dqinfo *ii)
{
	struct vz_quota_master *qmblk;
	int err;

	qmblk = vzquota_find_qmblk(sb);
	down(&vz_quota_sem);
	err = -ESRCH;
	if (qmblk == NULL)
		goto out;
	err = -EIO;
	if (qmblk == VZ_QUOTA_BAD)
		goto out;
	err = __vz_set_dqinfo(qmblk, type, ii);
out:
	up(&vz_quota_sem);
	if (qmblk != NULL && qmblk != VZ_QUOTA_BAD)
		qmblk_put(qmblk);
	return err;
}

#ifdef CONFIG_QUOTA_COMPAT

#define Q_GETQUOTI_SIZE 1024

#define UGID2DQBLK(dst, src)						\
	do {								\
		(dst).dqb_ihardlimit = (src)->qugid_stat.ihardlimit;	\
		(dst).dqb_isoftlimit = (src)->qugid_stat.isoftlimit;	\
		(dst).dqb_curinodes = (src)->qugid_stat.icurrent;	\
		/* in 1K blocks */					\
		(dst).dqb_bhardlimit = (src)->qugid_stat.bhardlimit >> 10; \
		/* in 1K blocks */					\
		(dst).dqb_bsoftlimit = (src)->qugid_stat.bsoftlimit >> 10; \
		/* in bytes, 64 bit */					\
		(dst).dqb_curspace = (src)->qugid_stat.bcurrent;	\
		(dst).dqb_btime = (src)->qugid_stat.btime;		\
		(dst).dqb_itime = (src)->qugid_stat.itime;		\
	} while (0)

static int vz_get_quoti(struct super_block *sb, int type, qid_t idx,
		struct v2_disk_dqblk *dqblk)
{
	struct vz_quota_master *qmblk;
	struct v2_disk_dqblk data;
	struct vz_quota_ugid *ugid;
	int count;
	int err;

	qmblk = vzquota_find_qmblk(sb);
	down(&vz_quota_sem);
	err = -ESRCH;
	if (qmblk == NULL)
		goto out;
	err = -EIO;
	if (qmblk == VZ_QUOTA_BAD)
		goto out;

	down(&qmblk->dq_sem);
	for (ugid = vzquota_get_byindex(qmblk, idx, type), count = 0;
		ugid != NULL && count < Q_GETQUOTI_SIZE;
		count++)
	{
		qmblk_data_read_lock(qmblk);
		UGID2DQBLK(data, ugid);
		qmblk_data_read_unlock(qmblk);
		data.dqb_id = ugid->qugid_id;
		if (copy_to_user(dqblk, &data, sizeof(data)))
			goto fault;
		dqblk++;

		/* Find next entry */
		ugid = vzquota_get_next(qmblk, ugid);
		BUG_ON(ugid != NULL && ugid->qugid_type != type);
	}
	err = count;
out_ugid:
	up(&qmblk->dq_sem);
out:
	up(&vz_quota_sem);
	if (qmblk != NULL && qmblk != VZ_QUOTA_BAD)
		qmblk_put(qmblk);

	return err;

fault:
	err = count ? count : -EFAULT;
	goto out_ugid;
}

#endif

struct quotactl_ops vz_quotactl_operations = {
	quota_on:	vz_quota_on,
	quota_off:	vz_quota_off,
	quota_sync:	vz_quota_sync,
	get_info:	vz_get_dqinfo,
	set_info:	vz_set_dqinfo,
	get_dqblk:	vz_get_dqblk,
	set_dqblk:	vz_set_dqblk,
#ifdef CONFIG_QUOTA_COMPAT
	get_quoti:	vz_get_quoti
#endif
};


/* ----------------------------------------------------------------------
 * Management interface for host system admins.
 * --------------------------------------------------------------------- */

static int quota_ugid_addstat(unsigned int quota_id, unsigned int ugid_size,
		struct vz_quota_iface *u_ugid_buf)
{
	struct vz_quota_master *qmblk;
	int ret;

	down(&vz_quota_sem);

	ret = -ENOENT;
	qmblk = vzquota_find_master(quota_id);
	if (qmblk == NULL)
		goto out;

	ret = -EBUSY;
	if (qmblk->dq_state != VZDQ_STARTING)
		goto out; /* working quota doesn't accept new ugids */

	ret = 0;
	/* start to add ugids */
	for (ret = 0; ret < ugid_size; ret++) {
		struct vz_quota_iface ugid_buf;
		struct vz_quota_ugid *ugid;

		if (copy_from_user(&ugid_buf, u_ugid_buf, sizeof(ugid_buf)))
			break;

		if (ugid_buf.qi_type >= MAXQUOTAS)
			break; /* bad quota type - this is the only check */

		ugid = vzquota_find_ugid(qmblk,
				ugid_buf.qi_id, ugid_buf.qi_type, 0);
		if (ugid == VZ_QUOTA_UGBAD) {
			qmblk->dq_flags |= VZDQUG_FIXED_SET;
			break; /* limit reached */
		}

		/* update usage/limits
		 * we can copy the data without the lock, because the data
		 * cannot be modified in VZDQ_STARTING state */
		ugid->qugid_stat = ugid_buf.qi_stat;

		vzquota_put_ugid(qmblk, ugid);

		u_ugid_buf++; /* next user buffer */
	}
out:
	up(&vz_quota_sem);

	return ret;
}

static int quota_ugid_setgrace(unsigned int quota_id,
		struct dq_info u_dq_info[])
{
	struct vz_quota_master *qmblk;
	struct dq_info dq_info[MAXQUOTAS];
	struct dq_info *target;
	int err, type;

	down(&vz_quota_sem);

	err = -ENOENT;
	qmblk = vzquota_find_master(quota_id);
	if (qmblk == NULL)
		goto out;

	err = -EBUSY;
	if (qmblk->dq_state != VZDQ_STARTING)
		goto out; /* working quota doesn't accept changing options */

	err = -EFAULT;
	if (copy_from_user(dq_info, u_dq_info, sizeof(dq_info)))
		goto out;

	err = 0;

	/* update in qmblk */
	for (type = 0; type < MAXQUOTAS; type ++) {
		target = &qmblk->dq_ugid_info[type];
		target->bexpire = dq_info[type].bexpire;
		target->iexpire = dq_info[type].iexpire;
	}
out:
	up(&vz_quota_sem);

	return err;
}

static int do_quota_ugid_getstat(struct vz_quota_master *qmblk, int index, int size,
		struct vz_quota_iface *u_ugid_buf)
{
	int type, count;
	struct vz_quota_ugid *ugid;

	if (QTREE_LEAFNUM(qmblk->dq_uid_tree) +
	    QTREE_LEAFNUM(qmblk->dq_gid_tree)
	    		<= index)
		return 0;

	count = 0;

	type = index < QTREE_LEAFNUM(qmblk->dq_uid_tree) ? USRQUOTA : GRPQUOTA;
	if (type == GRPQUOTA)
		index -= QTREE_LEAFNUM(qmblk->dq_uid_tree);

	/* loop through ugid and then qgid quota */
repeat:
	for (ugid = vzquota_get_byindex(qmblk, index, type);
		ugid != NULL && count < size;
		ugid = vzquota_get_next(qmblk, ugid), count++)
	{
		struct vz_quota_iface ugid_buf;

		/* form interface buffer and send in to user-level */
		qmblk_data_read_lock(qmblk);
		memcpy(&ugid_buf.qi_stat, &ugid->qugid_stat,
				sizeof(ugid_buf.qi_stat));
		qmblk_data_read_unlock(qmblk);
		ugid_buf.qi_id = ugid->qugid_id;
		ugid_buf.qi_type = ugid->qugid_type;

		if (copy_to_user(u_ugid_buf, &ugid_buf, sizeof(ugid_buf)))
			goto fault;
		u_ugid_buf++; /* next portion of user buffer */
	}

	if (type == USRQUOTA && count < size) {
		type = GRPQUOTA;
		index = 0;
		goto repeat;
	}

	return count;

fault:
	return count ? count : -EFAULT;
}

static int quota_ugid_getstat(unsigned int quota_id,
		int index, int size, struct vz_quota_iface *u_ugid_buf)
{
	struct vz_quota_master *qmblk;
	int err;

	if (index < 0 || size < 0)
		return -EINVAL;

	down(&vz_quota_sem);

	err = -ENOENT;
	qmblk = vzquota_find_master(quota_id);
	if (qmblk == NULL)
		goto out;

	down(&qmblk->dq_sem);
	err = do_quota_ugid_getstat(qmblk, index, size, u_ugid_buf);
	up(&qmblk->dq_sem);

out:
	up(&vz_quota_sem);
	return err;
}

static int quota_ugid_getgrace(unsigned int quota_id,
		struct dq_info u_dq_info[])
{
	struct vz_quota_master *qmblk;
	struct dq_info dq_info[MAXQUOTAS];
	struct dq_info *target;
	int err, type;

	down(&vz_quota_sem);

	err = -ENOENT;
	qmblk = vzquota_find_master(quota_id);
	if (qmblk == NULL)
		goto out;

	err = 0;
	/* update from qmblk */
	for (type = 0; type < MAXQUOTAS; type ++) {
		target = &qmblk->dq_ugid_info[type];
		dq_info[type].bexpire = target->bexpire;
		dq_info[type].iexpire = target->iexpire;
		dq_info[type].flags = target->flags;
	}

	if (copy_to_user(u_dq_info, dq_info, sizeof(dq_info)))
		err = -EFAULT;
out:
	up(&vz_quota_sem);

	return err;
}

static int quota_ugid_getconfig(unsigned int quota_id,
		struct vz_quota_ugid_stat *info)
{
	struct vz_quota_master *qmblk;
	struct vz_quota_ugid_stat kinfo;
	int err;

	down(&vz_quota_sem);

	err = -ENOENT;
	qmblk = vzquota_find_master(quota_id);
	if (qmblk == NULL)
		goto out;

	err = 0;
	kinfo.limit = qmblk->dq_ugid_max;
	kinfo.count = qmblk->dq_ugid_count;
	kinfo.flags = qmblk->dq_flags;

	if (copy_to_user(info, &kinfo, sizeof(kinfo)))
		err = -EFAULT;
out:
	up(&vz_quota_sem);

	return err;
}

static int quota_ugid_setconfig(unsigned int quota_id,
		struct vz_quota_ugid_stat *info)
{
	struct vz_quota_master *qmblk;
	struct vz_quota_ugid_stat kinfo;
	int err;

	down(&vz_quota_sem);

	err = -ENOENT;
	qmblk = vzquota_find_master(quota_id);
	if (qmblk == NULL)
		goto out;

	err = -EFAULT;
	if (copy_from_user(&kinfo, info, sizeof(kinfo)))
		goto out;

	err = 0;
	qmblk->dq_ugid_max = kinfo.limit;
	if (qmblk->dq_state == VZDQ_STARTING) {
		qmblk->dq_flags = kinfo.flags;
		if (qmblk->dq_flags & VZDQUG_ON)
			qmblk->dq_flags |= VZDQ_USRQUOTA | VZDQ_GRPQUOTA;
	}

out:
	up(&vz_quota_sem);

	return err;
}

static int quota_ugid_setlimit(unsigned int quota_id,
		struct vz_quota_ugid_setlimit *u_lim)
{
	struct vz_quota_master *qmblk;
	struct vz_quota_ugid_setlimit lim;
	int err;

	down(&vz_quota_sem);

	err = -ESRCH;
	qmblk = vzquota_find_master(quota_id);
	if (qmblk == NULL)
		goto out;

	err = -EFAULT;
	if (copy_from_user(&lim, u_lim, sizeof(lim)))
		goto out;

	err = __vz_set_dqblk(qmblk, lim.type, lim.id, &lim.dqb);

out:
	up(&vz_quota_sem);

	return err;
}

static int quota_ugid_setinfo(unsigned int quota_id,
		struct vz_quota_ugid_setinfo *u_info)
{
	struct vz_quota_master *qmblk;
	struct vz_quota_ugid_setinfo info;
	int err;

	down(&vz_quota_sem);

	err = -ESRCH;
	qmblk = vzquota_find_master(quota_id);
	if (qmblk == NULL)
		goto out;

	err = -EFAULT;
	if (copy_from_user(&info, u_info, sizeof(info)))
		goto out;

	err = __vz_set_dqinfo(qmblk, info.type, &info.dqi);

out:
	up(&vz_quota_sem);

	return err;
}

/*
 * This is a system call to maintain UGID quotas
 * Note this call is allowed to run ONLY from VE0
 */
long do_vzquotaugidctl(struct vzctl_quotaugidctl *qub)
{
	int ret;

	ret = -EPERM;
	/* access allowed only from root of VE0 */
	if (!capable(CAP_SYS_RESOURCE) ||
	    !capable(CAP_SYS_ADMIN))
		goto out;

	switch (qub->cmd) {
		case VZ_DQ_UGID_GETSTAT:
			ret = quota_ugid_getstat(qub->quota_id,
					qub->ugid_index, qub->ugid_size,
				       	(struct vz_quota_iface *)qub->addr);
			break;
		case VZ_DQ_UGID_ADDSTAT:
			ret = quota_ugid_addstat(qub->quota_id, qub->ugid_size,
				       	(struct vz_quota_iface *)qub->addr);
			break;
		case VZ_DQ_UGID_GETGRACE:
			ret = quota_ugid_getgrace(qub->quota_id,
					(struct dq_info *)qub->addr);
			break;
		case VZ_DQ_UGID_SETGRACE:
			ret = quota_ugid_setgrace(qub->quota_id,
					(struct dq_info *)qub->addr);
			break;
		case VZ_DQ_UGID_GETCONFIG:
			ret = quota_ugid_getconfig(qub->quota_id,
					(struct vz_quota_ugid_stat *)qub->addr);
			break;
		case VZ_DQ_UGID_SETCONFIG:
			ret = quota_ugid_setconfig(qub->quota_id,
					(struct vz_quota_ugid_stat *)qub->addr);
			break;
		case VZ_DQ_UGID_SETLIMIT:
			ret = quota_ugid_setlimit(qub->quota_id,
					(struct vz_quota_ugid_setlimit *)
								qub->addr);
			break;
		case VZ_DQ_UGID_SETINFO:
			ret = quota_ugid_setinfo(qub->quota_id,
					(struct vz_quota_ugid_setinfo *)
								qub->addr);
			break;
		default:
			ret = -EINVAL;
			goto out;
	}
out:
	return ret;
}

static void ugid_quota_on_sb(struct super_block *sb)
{
	struct super_block *real_sb;
	struct vz_quota_master *qmblk;

	if (!sb->s_op->get_quota_root)
		return;

	real_sb = sb->s_op->get_quota_root(sb)->i_sb;
	if (real_sb->dq_op != &vz_quota_operations)
		return;

	sb->dq_op = &vz_quota_operations2;
	sb->s_qcop = &vz_quotactl_operations;
	INIT_LIST_HEAD(&sb->s_dquot.info[USRQUOTA].dqi_dirty_list);
	INIT_LIST_HEAD(&sb->s_dquot.info[GRPQUOTA].dqi_dirty_list);
	sb->s_dquot.info[USRQUOTA].dqi_format = &vz_quota_empty_v2_format;
	sb->s_dquot.info[GRPQUOTA].dqi_format = &vz_quota_empty_v2_format;

	qmblk = vzquota_find_qmblk(sb);
	if ((qmblk == NULL) || (qmblk == VZ_QUOTA_BAD))
		return;
	down(&vz_quota_sem);
	if (qmblk->dq_flags & VZDQ_USRQUOTA)
		sb->s_dquot.flags |= DQUOT_USR_ENABLED;
	if (qmblk->dq_flags & VZDQ_GRPQUOTA)
		sb->s_dquot.flags |= DQUOT_GRP_ENABLED;
	up(&vz_quota_sem);
	qmblk_put(qmblk);
}

static void ugid_quota_off_sb(struct super_block *sb)
{
	/* can't make quota off on mounted super block */
	BUG_ON(sb->s_root != NULL);
}

static int ugid_notifier_call(struct vnotifier_block *self,
		unsigned long n, void *data, int old_ret)
{
	struct virt_info_quota *viq;

	viq = (struct virt_info_quota *)data;

	switch (n) {
	case VIRTINFO_QUOTA_ON:
		ugid_quota_on_sb(viq->super);
		break;
	case VIRTINFO_QUOTA_OFF:
		ugid_quota_off_sb(viq->super);
		break;
	case VIRTINFO_QUOTA_GETSTAT:
		break;
	default:
		return old_ret;
	}
	return NOTIFY_OK;
}

static struct vnotifier_block ugid_notifier_block = {
	.notifier_call = ugid_notifier_call,
};

/* ----------------------------------------------------------------------
 * Init/exit.
 * --------------------------------------------------------------------- */

struct quota_format_type vz_quota_empty_v2_format = {
	qf_fmt_id:	QFMT_VFS_V0,
	qf_ops:		NULL,
	qf_owner:	THIS_MODULE
};

int vzquota_ugid_init()
{
	int err;

	vz_quota_ugid_cachep = kmem_cache_create("vz_quota_ugid",
				      sizeof(struct vz_quota_ugid),
				      0, SLAB_HWCACHE_ALIGN,
				      NULL, NULL);
	if (vz_quota_ugid_cachep == NULL)
		goto err_slab;

	err = register_quota_format(&vz_quota_empty_v2_format);
	if (err)
		goto err_reg;

	virtinfo_notifier_register(VITYPE_QUOTA, &ugid_notifier_block);
	return 0;

err_reg:
	kmem_cache_destroy(vz_quota_ugid_cachep);
	return err;

err_slab:
	printk(KERN_ERR "Cannot create VZ_QUOTA SLAB cache\n");
	return -ENOMEM;
}

void vzquota_ugid_release()
{
	virtinfo_notifier_unregister(VITYPE_QUOTA, &ugid_notifier_block);
	unregister_quota_format(&vz_quota_empty_v2_format);

	if (kmem_cache_destroy(vz_quota_ugid_cachep))
		printk(KERN_ERR "VZQUOTA: kmem_cache_destroy failed\n");
}
