/*
 * Generic pidhash and scalable, time-bounded PID allocator
 *
 * (C) 2002-2003 William Irwin, IBM
 * (C) 2004 William Irwin, Oracle
 * (C) 2002-2004 Ingo Molnar, Red Hat
 *
 * pid-structures are backing objects for tasks sharing a given ID to chain
 * against. There is very little to them aside from hashing them and
 * parking tasks using given ID's on a list.
 *
 * The hash is always changed with the tasklist_lock write-acquired,
 * and the hash is only accessed with the tasklist_lock at least
 * read-acquired, so there's no additional SMP locking needed here.
 *
 * We have a list of bitmap pages, which bitmaps represent the PID space.
 * Allocating and freeing PIDs is completely lockless. The worst-case
 * allocation scenario when all but one out of 1 million PIDs possible are
 * allocated already: the scanning of 32 list entries and at most PAGE_SIZE
 * bytes. The typical fastpath is a single successful setbit. Freeing is O(1).
 */

#include <linux/mm.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/bootmem.h>
#include <linux/hash.h>

#ifdef CONFIG_VE
static void __free_vpid(int vpid, struct ve_struct *ve);
#endif

#define pid_hashfn(nr) hash_long((unsigned long)nr, pidhash_shift)
static struct hlist_head *pid_hash[PIDTYPE_MAX];
static int pidhash_shift;

int pid_max = PID_MAX_DEFAULT;
int last_pid;

#define RESERVED_PIDS		300

int pid_max_min = RESERVED_PIDS + 1;
int pid_max_max = PID_MAX_LIMIT;

#define PIDMAP_ENTRIES		((PID_MAX_LIMIT + 8*PAGE_SIZE - 1)/PAGE_SIZE/8)
#define BITS_PER_PAGE		(PAGE_SIZE*8)
#define BITS_PER_PAGE_MASK	(BITS_PER_PAGE-1)
#define mk_pid(map, off)	(((map) - pidmap_array)*BITS_PER_PAGE + (off))
#define find_next_offset(map, off)					\
		find_next_zero_bit((map)->page, BITS_PER_PAGE, off)

/*
 * PID-map pages start out as NULL, they get allocated upon
 * first use and are never deallocated. This way a low pid_max
 * value does not cause lots of bitmaps to be allocated, but
 * the scheme scales to up to 4 million PIDs, runtime.
 */
typedef struct pidmap {
	atomic_t nr_free;
	void *page;
} pidmap_t;

#ifdef CONFIG_VE
#define PIDMAP_NRFREE (BITS_PER_PAGE/2)
#else
#define PIDMAP_NRFREE BITS_PER_PAGE
#endif

static pidmap_t pidmap_array[PIDMAP_ENTRIES] =
	 { [ 0 ... PIDMAP_ENTRIES-1 ] = { ATOMIC_INIT(PIDMAP_NRFREE), NULL } };

static  __cacheline_aligned_in_smp DEFINE_SPINLOCK(pidmap_lock);

fastcall void free_pidmap(int pid)
{
	pidmap_t *map = pidmap_array + pid / BITS_PER_PAGE;
	int offset = pid & BITS_PER_PAGE_MASK;

	BUG_ON(__is_virtual_pid(pid) || pid == 1);

	clear_bit(offset, map->page);
	atomic_inc(&map->nr_free);
}

int alloc_pidmap(void)
{
	int i, offset, max_scan, pid, last = last_pid;
	pidmap_t *map;

	pid = last + 1;
	if (__is_virtual_pid(pid))
		pid += VPID_DIV;
	if (pid >= pid_max)
		pid = RESERVED_PIDS;
	offset = pid & BITS_PER_PAGE_MASK;
	map = &pidmap_array[pid/BITS_PER_PAGE];
	max_scan = (pid_max + BITS_PER_PAGE - 1)/BITS_PER_PAGE - !offset;
	for (i = 0; i <= max_scan; ++i) {
		if (unlikely(!map->page)) {
			unsigned long page = get_zeroed_page(GFP_KERNEL);
			/*
			 * Free the page if someone raced with us
			 * installing it:
			 */
			spin_lock(&pidmap_lock);
			if (map->page)
				free_page(page);
			else
				map->page = (void *)page;
			spin_unlock(&pidmap_lock);
			if (unlikely(!map->page))
				break;
		}
		if (likely(atomic_read(&map->nr_free))) {
			do {
				if (!test_and_set_bit(offset, map->page)) {
					atomic_dec(&map->nr_free);
					last_pid = pid;
					return pid;
				}
				offset = find_next_offset(map, offset);
				if (__is_virtual_pid(offset))
					offset += VPID_DIV;
				pid = mk_pid(map, offset);
			/*
			 * find_next_offset() found a bit, the pid from it
			 * is in-bounds, and if we fell back to the last
			 * bitmap block and the final block was the same
			 * as the starting point, pid is before last_pid.
			 */
			} while (offset < BITS_PER_PAGE && pid < pid_max &&
					(i != max_scan || pid < last ||
					    !((last+1) & BITS_PER_PAGE_MASK)));
		}
		if (map < &pidmap_array[(pid_max-1)/BITS_PER_PAGE]) {
			++map;
			offset = 0;
		} else {
			map = &pidmap_array[0];
			offset = RESERVED_PIDS;
			if (unlikely(last == offset))
				break;
		}
		pid = mk_pid(map, offset);
	}
	return -1;
}

struct pid * fastcall find_pid(enum pid_type type, int nr)
{
	struct hlist_node *elem;
	struct pid *pid;

	hlist_for_each_entry_rcu(pid, elem,
			&pid_hash[type][pid_hashfn(nr)], pid_chain) {
		if (pid->nr == nr)
			return pid;
	}
	return NULL;
}
EXPORT_SYMBOL(find_pid);

int fastcall attach_pid(task_t *task, enum pid_type type, int nr)
{
	struct pid *pid, *task_pid;

	task_pid = &task->pids[type];
	pid = find_pid(type, nr);
	task_pid->nr = nr;
	if (pid == NULL) {
		INIT_LIST_HEAD(&task_pid->pid_list);
		hlist_add_head_rcu(&task_pid->pid_chain,
				   &pid_hash[type][pid_hashfn(nr)]);
	} else {
		INIT_HLIST_NODE(&task_pid->pid_chain);
		list_add_tail_rcu(&task_pid->pid_list, &pid->pid_list);
	}

	return 0;
}

static fastcall int __detach_pid(task_t *task, enum pid_type type)
{
	struct pid *pid, *pid_next;
	int nr = 0;

	pid = &task->pids[type];
	if (!hlist_unhashed(&pid->pid_chain)) {

		if (list_empty(&pid->pid_list)) {
			nr = pid->nr;
			hlist_del_rcu(&pid->pid_chain);
		} else {
			pid_next = list_entry(pid->pid_list.next,
						struct pid, pid_list);
			/* insert next pid from pid_list to hash */
			hlist_replace_rcu(&pid->pid_chain,
					  &pid_next->pid_chain);
		}
	}

	list_del_rcu(&pid->pid_list);
	pid->nr = 0;

	return nr;
}

void fastcall detach_pid(task_t *task, enum pid_type type)
{
	int tmp, nr;

	nr = __detach_pid(task, type);
	if (!nr)
		return;

	for (tmp = PIDTYPE_MAX; --tmp >= 0; )
		if (tmp != type && find_pid(tmp, nr))
			return;

#ifdef CONFIG_VE
	__free_vpid(task->pids[type].vnr, VE_TASK_INFO(task)->owner_env);
#endif
	free_pidmap(nr);
}

task_t *find_task_by_pid_type(int type, int nr)
{
	BUG();
	return NULL;
}

EXPORT_SYMBOL(find_task_by_pid_type);

task_t *find_task_by_pid_type_all(int type, int nr)
{
	struct pid *pid;

	BUG_ON(nr != -1 && is_virtual_pid(nr));

	pid = find_pid(type, nr);
	if (!pid)
		return NULL;

	return pid_task(&pid->pid_list, type);
}

EXPORT_SYMBOL(find_task_by_pid_type_all);

#ifdef CONFIG_VE

task_t *find_task_by_pid_type_ve(int type, int nr)
{
	task_t *tsk;
	int gnr = nr;
	struct pid *pid;

	if (is_virtual_pid(nr)) {
		gnr = __vpid_to_pid(nr);
		if (unlikely(gnr == -1))
			return NULL;
	}

	pid = find_pid(type, gnr);
	if (!pid)
		return NULL;

	tsk = pid_task(&pid->pid_list, type);
	if (!ve_accessible(VE_TASK_INFO(tsk)->owner_env, get_exec_env()))
		return NULL;
	return tsk;
}

EXPORT_SYMBOL(find_task_by_pid_type_ve);

#endif

/*
 * This function switches the PIDs if a non-leader thread calls
 * sys_execve() - this must be done without releasing the PID.
 * (which a detach_pid() would eventually do.)
 */
void switch_exec_pids(task_t *leader, task_t *thread)
{
	__detach_pid(leader, PIDTYPE_PID);
	__detach_pid(leader, PIDTYPE_TGID);
	__detach_pid(leader, PIDTYPE_PGID);
	__detach_pid(leader, PIDTYPE_SID);

	__detach_pid(thread, PIDTYPE_PID);
	__detach_pid(thread, PIDTYPE_TGID);

	leader->pid = leader->tgid = thread->pid;
	thread->pid = thread->tgid;
	set_virt_tgid(leader, virt_pid(thread));
	set_virt_pid(leader, virt_pid(thread));
	set_virt_pid(thread, virt_tgid(thread));

	attach_pid(thread, PIDTYPE_PID, thread->pid);
	attach_pid(thread, PIDTYPE_TGID, thread->tgid);
	attach_pid(thread, PIDTYPE_PGID, thread->signal->pgrp);
	attach_pid(thread, PIDTYPE_SID, thread->signal->session);
	list_add_tail(&thread->tasks, &init_task.tasks);

	attach_pid(leader, PIDTYPE_PID, leader->pid);
	attach_pid(leader, PIDTYPE_TGID, leader->tgid);
	attach_pid(leader, PIDTYPE_PGID, leader->signal->pgrp);
	attach_pid(leader, PIDTYPE_SID, leader->signal->session);
}

#ifdef CONFIG_VE

/* Virtual PID bits.
 *
 * At the moment all internal structures in kernel store real global pid.
 * The only place, where virtual PID is used, is at user frontend. We
 * remap virtual pids obtained from user to global ones (vpid_to_pid) and
 * map globals to virtuals before showing them to user (virt_pid_type).
 *
 * We hold virtual PIDs inside struct pid, so map global -> virtual is easy.
 */

pid_t _pid_type_to_vpid(int type, pid_t pid)
{
	struct pid * p;

	if (unlikely(is_virtual_pid(pid)))
		return -1;

	read_lock(&tasklist_lock);
	p = find_pid(type, pid);
	if (p) {
		pid = p->vnr;
	} else {
		pid = -1;
	}
	read_unlock(&tasklist_lock);
	return pid;
}

pid_t pid_type_to_vpid(int type, pid_t pid)
{
	int vpid;

	if (unlikely(pid <= 0))
		return pid;

	BUG_ON(is_virtual_pid(pid));

	if (ve_is_super(get_exec_env()))
		return pid;

	vpid = _pid_type_to_vpid(type, pid);
	if (unlikely(vpid == -1)) {
		/* It is allowed: global pid can be used everywhere.
		 * This can happen, when kernel remembers stray pids:
		 * signal queues, locks etc.
		 */
		vpid = pid;
	}
	return vpid;
}

/* To map virtual pids to global we maintain special hash table.
 *
 * Mapping entries are allocated when a process with non-trivial
 * mapping is forked, which is possible only after VE migrated.
 * Mappings are destroyed, when a global pid is removed from global
 * pidmap, which means we do not need to refcount mappings.
 */

static struct hlist_head *vpid_hash;

struct vpid_mapping
{
	int	vpid;
	int	veid;
	int	pid;
	struct hlist_node link;
};

static kmem_cache_t *vpid_mapping_cachep;

static inline int vpid_hashfn(int vnr, int veid)
{
	return hash_long((unsigned long)(vnr+(veid<<16)), pidhash_shift);
}

struct vpid_mapping *__lookup_vpid_mapping(int vnr, int veid)
{
	struct hlist_node *elem;
	struct vpid_mapping *map;

	hlist_for_each_entry(map, elem,
			&vpid_hash[vpid_hashfn(vnr, veid)], link) {
		if (map->vpid == vnr && map->veid == veid)
			return map;
	}
	return NULL;
}

/* __vpid_to_pid() is raw version of vpid_to_pid(). It is to be used
 * only under tasklist_lock. In some places we must use only this version
 * (f.e. __kill_pg_info is called under write lock!)
 *
 * Caller should pass virtual pid. This function returns an error, when
 * seeing a global pid.
 */
int __vpid_to_pid(int pid)
{
	struct vpid_mapping *map;

	if (unlikely(!is_virtual_pid(pid) || ve_is_super(get_exec_env())))
		return -1;

	if (!get_exec_env()->sparse_vpid) {
		if (pid != 1)
			return pid - VPID_DIV;
		return get_exec_env()->init_entry->pid;
	}

	map = __lookup_vpid_mapping(pid, VEID(get_exec_env()));
	if (map)
		return map->pid;
	return -1;
}

int vpid_to_pid(int pid)
{
	/* User gave bad pid. It is his problem. */
	if (unlikely(pid <= 0))
		return pid;

	if (!is_virtual_pid(pid))
		return pid;

	read_lock(&tasklist_lock);
	pid = __vpid_to_pid(pid);
	read_unlock(&tasklist_lock);
	return pid;
}

/* VEs which never migrated have trivial "arithmetic" mapping pid <-> vpid:
 *
 * vpid == 1 -> ve->init_task->pid
 * else	        pid & ~VPID_DIV
 *
 * In this case VE has ve->sparse_vpid = 0 and we do not use vpid hash table.
 *
 * When VE migrates and we see non-trivial mapping the first time, we
 * scan process table and populate mapping hash table.
 */

static int add_mapping(int pid, int vpid, int veid, struct hlist_head *cache)
{
        if (pid > 0 && vpid > 0 && !__lookup_vpid_mapping(vpid, veid)) {
		struct vpid_mapping *m;
		if (hlist_empty(cache)) {
			m = kmem_cache_alloc(vpid_mapping_cachep, GFP_ATOMIC);
			if (unlikely(m == NULL))
				return -ENOMEM;
		} else {
			m = hlist_entry(cache->first, struct vpid_mapping, link);
			hlist_del(&m->link);
		}
		m->pid = pid;
		m->vpid = vpid;
		m->veid = veid;
		hlist_add_head(&m->link,
			       &vpid_hash[vpid_hashfn(vpid, veid)]);
	}
	return 0;
}

static int switch_to_sparse_mapping(int pid)
{
	struct ve_struct *env = get_exec_env();
	struct hlist_head cache;
	task_t *g, *t;
	int pcount;
	int err;

	/* Transition happens under write_lock_irq, so we try to make
	 * it more reliable and fast preallocating mapping entries.
	 * pcounter may be not enough, we could have lots of orphaned
	 * process groups and sessions, which also require mappings.
	 */
	INIT_HLIST_HEAD(&cache);
	pcount = atomic_read(&env->pcounter);
	err = -ENOMEM;
	while (pcount > 0) {
		struct vpid_mapping *m;
		m = kmem_cache_alloc(vpid_mapping_cachep, GFP_KERNEL);
		if (!m)
			goto out;
		hlist_add_head(&m->link, &cache);
		pcount--;
	}

	write_lock_irq(&tasklist_lock);
	err = 0;
	if (env->sparse_vpid)
		goto out_unlock;

	err = -ENOMEM;
	do_each_thread_ve(g, t) {
		if (t->pid == pid)
			continue;
		if (add_mapping(t->pid, virt_pid(t), VEID(env), &cache))
			goto out_unlock;
	} while_each_thread_ve(g, t);

	for_each_process_ve(t) {
		if (t->pid == pid)
			continue;

		if (add_mapping(t->tgid, virt_tgid(t), VEID(env), &cache))
			goto out_unlock;
		if (add_mapping(t->signal->pgrp, virt_pgid(t), VEID(env), &cache))
			goto out_unlock;
		if (add_mapping(t->signal->session, virt_sid(t), VEID(env), &cache))
			goto out_unlock;
	}
	env->sparse_vpid = 1;
	err = 0;

out_unlock:
	if (err) {
		int i;

		for (i=0; i<(1<<pidhash_shift); i++) {
			struct hlist_node *elem, *next;
			struct vpid_mapping *map;

			hlist_for_each_entry_safe(map, elem, next, &vpid_hash[i], link) {
				if (map->veid == VEID(env)) {
					hlist_del(elem);
					hlist_add_head(elem, &cache);
				}
			}
		}
	}
	write_unlock_irq(&tasklist_lock);

out:
	while (!hlist_empty(&cache)) {
		struct vpid_mapping *m;
		m = hlist_entry(cache.first, struct vpid_mapping, link);
		hlist_del(&m->link);
		kmem_cache_free(vpid_mapping_cachep, m);
	}
	return err;
}

int alloc_vpid(int pid, int virt_pid)
{
	int result;
	struct vpid_mapping *m;
	struct ve_struct *env = get_exec_env();

	if (ve_is_super(env) || !env->virt_pids)
		return pid;

	if (!env->sparse_vpid) {
		if (virt_pid == -1)
			return pid + VPID_DIV;

		if (virt_pid == 1 || virt_pid == pid + VPID_DIV)
			return virt_pid;

		if ((result = switch_to_sparse_mapping(pid)) < 0)
			return result;
	}

	m = kmem_cache_alloc(vpid_mapping_cachep, GFP_KERNEL);
	if (!m)
		return -ENOMEM;

	m->pid = pid;
	m->veid = VEID(env);

	result = (virt_pid == -1) ? pid + VPID_DIV : virt_pid;

	write_lock_irq(&tasklist_lock);
	if (unlikely(__lookup_vpid_mapping(result, m->veid))) {
		if (virt_pid > 0) {
			result = -EEXIST;
			goto out;
		}

		/* No luck. Now we search for some not-existing vpid.
		 * It is weak place. We do linear search. */
		do {
			result++;
			if (!__is_virtual_pid(result))
				result += VPID_DIV;
			if (result >= pid_max)
				result = RESERVED_PIDS + VPID_DIV;
		} while (__lookup_vpid_mapping(result, m->veid) != NULL);

		/* And set last_pid in hope future alloc_pidmap to avoid
		 * collisions after future alloc_pidmap() */
		last_pid = result - VPID_DIV;
	}
	if (result > 0) {
		m->vpid = result;
		hlist_add_head(&m->link,
			       &vpid_hash[vpid_hashfn(result, m->veid)]);
	}
out:
	write_unlock_irq(&tasklist_lock);
	if (result < 0)
		kmem_cache_free(vpid_mapping_cachep, m);
	return result;
}
EXPORT_SYMBOL(alloc_vpid);

static void __free_vpid(int vpid, struct ve_struct *ve)
{
	struct vpid_mapping *m;

	if (!ve->sparse_vpid)
		return;

	if (!__is_virtual_pid(vpid) && (vpid != 1 || ve_is_super(ve)))
		return;

	m = __lookup_vpid_mapping(vpid, ve->veid);
	BUG_ON(m == NULL);
	hlist_del(&m->link);
	kmem_cache_free(vpid_mapping_cachep, m);
}

void free_vpid(int vpid, struct ve_struct *ve)
{
	write_lock_irq(&tasklist_lock);
	__free_vpid(vpid, ve);
	write_unlock_irq(&tasklist_lock);
}
EXPORT_SYMBOL(free_vpid);
#endif

/*
 * The pid hash table is scaled according to the amount of memory in the
 * machine.  From a minimum of 16 slots up to 4096 slots at one gigabyte or
 * more.
 */
void __init pidhash_init(void)
{
	int i, j, pidhash_size;
	unsigned long megabytes = nr_kernel_pages >> (20 - PAGE_SHIFT);

	pidhash_shift = max(4, fls(megabytes * 4));
	pidhash_shift = min(12, pidhash_shift);
	pidhash_size = 1 << pidhash_shift;

	printk("PID hash table entries: %d (order: %d, %Zd bytes)\n",
		pidhash_size, pidhash_shift,
		PIDTYPE_MAX * pidhash_size * sizeof(struct hlist_head));

	for (i = 0; i < PIDTYPE_MAX; i++) {
		pid_hash[i] = alloc_bootmem(pidhash_size *
					sizeof(*(pid_hash[i])));
		if (!pid_hash[i])
			panic("Could not alloc pidhash!\n");
		for (j = 0; j < pidhash_size; j++)
			INIT_HLIST_HEAD(&pid_hash[i][j]);
	}

#ifdef CONFIG_VE
	vpid_hash = alloc_bootmem(pidhash_size * sizeof(struct hlist_head));
	if (!vpid_hash)
		panic("Could not alloc vpid_hash!\n");
	for (j = 0; j < pidhash_size; j++)
		INIT_HLIST_HEAD(&vpid_hash[j]);
#endif
}

void __init pidmap_init(void)
{
	int i;

	pidmap_array->page = (void *)get_zeroed_page(GFP_KERNEL);
	set_bit(0, pidmap_array->page);
	atomic_dec(&pidmap_array->nr_free);

	/*
	 * Allocate PID 0, and hash it via all PID types:
	 */

	for (i = 0; i < PIDTYPE_MAX; i++)
		attach_pid(current, i, 0);

#ifdef CONFIG_VE
	vpid_mapping_cachep =
		kmem_cache_create("vpid_mapping",
				  sizeof(struct vpid_mapping),
				  __alignof__(struct vpid_mapping),
				  SLAB_PANIC|SLAB_UBC, NULL, NULL);
#endif
}
