/*
 *  kernel/ub/ub_dcache.c
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/config.h>
#include <linux/dcache.h>
#include <linux/slab.h>
#include <linux/kmem_cache.h>
#include <linux/fs.h>
#include <linux/err.h>

#include <ub/beancounter.h>
#include <ub/ub_mem.h>
#include <ub/ub_dcache.h>

/*
 * Locking
 *                          traverse  dcache_lock  d_lock
 *        ub_dentry_charge   +         +            +
 *      ub_dentry_uncharge   +         -            +
 * ub_dentry_charge_nofail   +         +            -
 *
 * d_inuse is atomic so that we can inc dentry's parent d_inuse in
 * ub_dentry_charhe with the only dentry's d_lock held.
 *
 * Race in uncharge vs charge_nofail is handled with dcache_lock.
 * Race in charge vs charge_nofail is inessential since they both inc d_inuse.
 * Race in uncharge vs charge is handled by altering d_inuse under d_lock.
 *
 * Race with d_move is handled this way:
 *  - charge_nofail and uncharge are protected by dcache_lock;
 *  - charge works only with dentry and dentry->d_parent->d_inuse, so
 *    it's enough to lock only the dentry.
 */

/*
 * Beancounting
 * UB argument must NOT be NULL
 */

static int do_charge_dcache(struct user_beancounter *ub, unsigned long size,
		enum severity sv)
{
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	if (__charge_beancounter_locked(ub, UB_KMEMSIZE, CHARGE_SIZE(size), sv))
		goto out_mem;
	if (__charge_beancounter_locked(ub, UB_DCACHESIZE, size, sv))
		goto out_dcache;
	spin_unlock_irqrestore(&ub->ub_lock, flags);
	return 0;

out_dcache:
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, CHARGE_SIZE(size));
out_mem:
	spin_unlock_irqrestore(&ub->ub_lock, flags);
	return -ENOMEM;
}

static void do_uncharge_dcache(struct user_beancounter *ub,
		unsigned long size)
{
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, CHARGE_SIZE(size));
	__uncharge_beancounter_locked(ub, UB_DCACHESIZE, size);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}

static int charge_dcache(struct user_beancounter *ub, unsigned long size,
		enum severity sv)
{
	struct user_beancounter *p, *q;

	for (p = ub; p != NULL; p = p->parent) {
		if (do_charge_dcache(p, size, sv))
			goto unroll;
	}
	return 0;

unroll:
	for (q = ub; q != p; q = q->parent)
		do_uncharge_dcache(q, size);
	return -ENOMEM;
}

void uncharge_dcache(struct user_beancounter *ub, unsigned long size)
{
	for (; ub != NULL; ub = ub->parent)
		do_uncharge_dcache(ub, size);
}

static inline void charge_dcache_forced(struct user_beancounter *ub,
		unsigned long size)
{
	charge_dcache(ub, size, UB_FORCE);
}

static inline void d_forced_charge(struct dentry_beancounter *d_bc)
{
	d_bc->d_ub = get_beancounter(get_exec_ub());
	if (d_bc->d_ub == NULL)
		return;

	charge_dcache_forced(d_bc->d_ub, d_bc->d_ubsize);
}

static inline void d_uncharge(struct dentry_beancounter *d_bc)
{
	if (d_bc->d_ub == NULL)
		return;

	uncharge_dcache(d_bc->d_ub, d_bc->d_ubsize);
	put_beancounter(d_bc->d_ub);
	d_bc->d_ub = NULL;
}

/*
 * Alloc / free dentry_beancounter
 */

static inline int d_alloc_beancounter(struct dentry *d)
{
	return 0;
}

static inline void d_free_beancounter(struct dentry_beancounter *d_bc)
{
}

static inline unsigned long d_charge_size(struct dentry *dentry)
{
	/* dentry's d_name is already set to appropriate value (see d_alloc) */
	return inode_cachep->objuse + dentry_cache->objuse +
		(dname_external(dentry) ?
		 kmem_obj_memusage((void *)dentry->d_name.name) : 0);
}

/*
 * dentry mark in use operation
 * d_lock is held
 */

static int d_inc_inuse(struct dentry *dentry)
{
	struct user_beancounter *ub;
	struct dentry_beancounter *d_bc;

	if (dentry != dentry->d_parent) {
		struct dentry *parent;

		/*
		 * Increment d_inuse of parent.
		 * It can't change since dentry->d_lock is held.
		 */
		parent = dentry->d_parent;
		if (ub_dget_testone(parent))
			BUG();
	}

	d_bc = &dentry->dentry_bc;
	ub = get_beancounter(get_exec_ub());

	if (ub != NULL && charge_dcache(ub, d_bc->d_ubsize, UB_SOFT))
		goto out_err;

	d_bc->d_ub = ub;
	return 0;

out_err:
	put_beancounter(ub);
	d_bc->d_ub = NULL;
	return -ENOMEM;
}

/*
 * no locks
 */
int ub_dentry_alloc(struct dentry *dentry)
{
	int err;
	struct dentry_beancounter *d_bc;

	err = d_alloc_beancounter(dentry);
	if (err < 0)
		return err;

	d_bc = &dentry->dentry_bc;
	d_bc->d_ub = get_beancounter(get_exec_ub());
	atomic_set(&d_bc->d_inuse, INUSE_INIT); /* see comment in ub_dcache.h */
	d_bc->d_ubsize = d_charge_size(dentry);

	err = 0;
	if (d_bc->d_ub != NULL &&
			charge_dcache(d_bc->d_ub, d_bc->d_ubsize, UB_HARD)) {
		put_beancounter(d_bc->d_ub);
		d_free_beancounter(d_bc);
		err = -ENOMEM;
	}

	return err;
}

/*
 * Charge / uncharge functions.
 *
 * We take d_lock to protect dentry_bc from concurrent acces
 * when simultaneous __d_lookup and d_put happens on one dentry.
 */

/*
 * no dcache_lock, d_lock and rcu_read_lock are held
 * drops d_lock, rcu_read_lock and returns error if any
 */
int ub_dentry_charge(struct dentry *dentry)
{
	int err;

	err = 0;
	if (ub_dget_testone(dentry))
		err = d_inc_inuse(dentry);

	/*
	 * d_lock and rcu_read_lock are dropped here
	 * (see also __d_lookup)
	 */
	spin_unlock(&dentry->d_lock);
	rcu_read_unlock();

	if (!err)
		return 0;

	/*
	 * d_invlaidate is required for real_lookup
	 * since it tries to create new dentry on
	 * d_lookup failure.
	 */
	if (!d_invalidate(dentry))
		return err;

	/* didn't succeeded, force dentry to be charged */
	d_forced_charge(&dentry->dentry_bc);
	return 0;
}

/*
 * dcache_lock is held
 * no d_locks, sequentaly takes and drops from dentry upward
 */
void ub_dentry_uncharge(struct dentry *dentry)
{
	struct dentry *parent;

	/* go up until status is changed and root is not reached */
	while (1) {
		/*
		 * We need d_lock here to handle
		 * the race with ub_dentry_charge
		 */
		spin_lock(&dentry->d_lock);
		if (!ub_dput_testzero(dentry)) {
			spin_unlock(&dentry->d_lock);
			break;
		}

		/* state transition 0 => -1 */
		d_uncharge(&dentry->dentry_bc);
		parent = dentry->d_parent;
		spin_unlock(&dentry->d_lock);

		/*
		 * dcache_lock is held (see comment in __dget_locked)
		 * so we can safely move upwards.
		 */
		if (dentry == parent)
			break;
		dentry = parent;
	}
}

/*
 * forced version. for dget in clean cache, when error is not an option
 *
 * dcache_lock is held
 * no d_locks
 */
void ub_dentry_charge_nofail(struct dentry *dentry)
{
	struct dentry *parent;

	/* go up until status is changed and root is not reached */
	while (1) {
		if (!ub_dget_testone(dentry))
			break;

		/*
		 * state transition -1 => 0
		 *
		 * No need to lock dentry before atomic_inc
		 * like we do in ub_dentry_uncharge.
		 * We can't race with ub_dentry_uncharge due
		 * to dcache_lock. The only possible race with
		 * ub_dentry_charge is OK since they both
		 * do atomic_inc.
		 */
		d_forced_charge(&dentry->dentry_bc);
		/*
		 * dcache_lock is held (see comment in __dget_locked)
		 * so we can safely move upwards.
		 */
		parent = dentry->d_parent;

		if (dentry == parent)
			break;
		dentry = parent;
	}
}
