/*
 *  kernel/ub/ub_mem.c
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/slab.h>
#include <linux/kmem_cache.h>
#include <linux/kmem_slab.h>
#include <linux/highmem.h>
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/gfp.h>
#include <linux/swap.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <ub/beancounter.h>
#include <ub/ub_mem.h>
#include <ub/ub_hash.h>

/*
 * Initialization
 */

/*
 * Slab accounting
 */

#ifdef CONFIG_UBC_DEBUG_KMEM

#define CC_HASH_SIZE	1024
static struct ub_cache_counter *cc_hash[CC_HASH_SIZE];
spinlock_t cc_lock;

static void __free_cache_counters(struct user_beancounter *ub,
		kmem_cache_t *cachep)
{
	struct ub_cache_counter *cc, **pprev, *del;
	int i;
	unsigned long flags;

	del = NULL;
	spin_lock_irqsave(&cc_lock, flags);
	for (i = 0; i < CC_HASH_SIZE; i++) {
		pprev = &cc_hash[i];
		cc = cc_hash[i];
		while (cc != NULL) {
			if (cc->ub != ub && cc->cachep != cachep) {
				pprev = &cc->next;
				cc = cc->next;
				continue;
			}

			list_del(&cc->ulist);
			*pprev = cc->next;
			cc->next = del;
			del = cc;
			cc = *pprev;
		}
	}
	spin_unlock_irqrestore(&cc_lock, flags);

	while (del != NULL) {
		cc = del->next;
		kfree(del);
		del = cc;
	}
}

void ub_free_counters(struct user_beancounter *ub)
{
	__free_cache_counters(ub, NULL);
}

void ub_kmemcache_free(kmem_cache_t *cachep)
{
	__free_cache_counters(NULL, cachep);
}

void __init init_cache_counters(void)
{
	memset(cc_hash, 0, CC_HASH_SIZE * sizeof(cc_hash[0]));
	spin_lock_init(&cc_lock);
}

#define cc_hash_fun(ub, cachep)	(				\
	(((unsigned long)(ub) >> L1_CACHE_SHIFT) ^		\
	 ((unsigned long)(ub) >> (BITS_PER_LONG / 2)) ^		\
	 ((unsigned long)(cachep) >> L1_CACHE_SHIFT) ^		\
	 ((unsigned long)(cachep) >> (BITS_PER_LONG / 2))	\
	) & (CC_HASH_SIZE - 1))

static int change_slab_charged(struct user_beancounter *ub, void *objp,
		unsigned long val, int mask)
{
	struct ub_cache_counter *cc, *new_cnt, **pprev;
	kmem_cache_t *cachep;
	unsigned long flags;

	cachep = virt_to_cache(objp);
	new_cnt = NULL;

again:
	spin_lock_irqsave(&cc_lock, flags);
	cc = cc_hash[cc_hash_fun(ub, cachep)];
	while (cc) {
		if (cc->ub == ub && cc->cachep == cachep)
			goto found;
		cc = cc->next;
	}

	if (new_cnt != NULL)
		goto insert;

	spin_unlock_irqrestore(&cc_lock, flags);

	new_cnt = kmalloc(sizeof(*new_cnt), mask & ~__GFP_UBC);
	if (new_cnt == NULL)
		return -ENOMEM;

	new_cnt->counter = 0;
	new_cnt->ub = ub;
	new_cnt->cachep = cachep;
	goto again;

insert:
	pprev = &cc_hash[cc_hash_fun(ub, cachep)];
	new_cnt->next = *pprev;
	*pprev = new_cnt;
	list_add(&new_cnt->ulist, &ub->ub_cclist);
	cc = new_cnt;
	new_cnt = NULL;

found:
	cc->counter += val;
	spin_unlock_irqrestore(&cc_lock, flags);
	if (new_cnt)
		kfree(new_cnt);
	return 0;
}

static inline int inc_slab_charged(struct user_beancounter *ub,
		void *objp, int mask)
{
	return change_slab_charged(ub, objp, 1, mask);
}

static inline void dec_slab_charged(struct user_beancounter *ub, void *objp)
{
	if (change_slab_charged(ub, objp, -1, 0) < 0)
		BUG();
}

#include <linux/vmalloc.h>

static inline int inc_pages_charged(struct user_beancounter *ub,
		struct page *pg, int order)
{
	int cpu;

	cpu = get_cpu();
	ub->ub_stat[cpu].pages_charged += (1 << order);
	put_cpu();
	return 0;
}

static inline void dec_pages_charged(struct user_beancounter *ub,
		struct page *pg, int order)
{
	int cpu;

	cpu = get_cpu();
	ub->ub_stat[cpu].pages_charged -= (1 << order);
	put_cpu();
}

void inc_vmalloc_charged(struct vm_struct *vm, int flags)
{
	int cpu;
	struct user_beancounter *ub;

	if (!(flags & __GFP_UBC))
		return;

	ub = get_exec_ub();
	if (ub == NULL)
		return;

	cpu = get_cpu();
	ub->ub_stat[cpu].vmalloc_charged += vm->nr_pages;
	put_cpu();
}

void dec_vmalloc_charged(struct vm_struct *vm)
{
	int cpu;
	struct user_beancounter *ub;

	ub = page_ub(vm->pages[0]);
	if (ub == NULL)
		return;

	cpu = get_cpu();
	ub->ub_stat[cpu].vmalloc_charged -= vm->nr_pages;
	put_cpu();
}

#else
#define inc_slab_charged(ub, o, m)	(0)
#define dec_slab_charged(ub, o)		do { } while (0)
#define inc_pages_charged(ub, pg, o) 	(0)
#define dec_pages_charged(ub, pg, o)	do { } while (0)
#endif

static inline struct user_beancounter **slab_ub_ref(void *objp)
{
	kmem_cache_t *cachep;
	struct slab *slabp;
	int objnr;

	cachep = virt_to_cache(objp);
	BUG_ON(!(cachep->flags & SLAB_UBC));
	slabp = virt_to_slab(objp);
	objnr = (objp - slabp->s_mem) / cachep->buffer_size;
	return slab_ubcs(cachep, slabp) + objnr;
}

struct user_beancounter *slab_ub(void *objp)
{
	struct user_beancounter **ub_ref;

	ub_ref = slab_ub_ref(objp);
	return *ub_ref;
}

EXPORT_SYMBOL(slab_ub);

static inline int should_charge(void *objp, int flags)
{
	kmem_cache_t *cachep;

	cachep = virt_to_cache(objp);
	if (!(cachep->flags & SLAB_UBC))
		return 0;
	if ((cachep->flags & SLAB_NO_CHARGE) && !(flags & __GFP_UBC))
		return 0;
	return 1;
}

#define should_uncharge(objp)	should_charge(objp, __GFP_UBC)

int ub_slab_charge(void *objp, int flags)
{
	unsigned int size;
	struct user_beancounter *ub;

	if (!should_charge(objp, flags))
		return 0;

	ub = get_beancounter(get_exec_ub());
	if (ub == NULL)
		return 0;

	size = CHARGE_SIZE(kmem_obj_memusage(objp));
	if (charge_beancounter(ub, UB_KMEMSIZE, size,
				(flags & __GFP_SOFT_UBC ? UB_SOFT : UB_HARD)))
		goto out_err;

	if (inc_slab_charged(ub, objp, flags) < 0) {
		uncharge_beancounter(ub, UB_KMEMSIZE, size);
		goto out_err;
	}
	*slab_ub_ref(objp) = ub;
	return 0;

out_err:
	put_beancounter(ub);
	return -ENOMEM;
}

void ub_slab_uncharge(void *objp)
{
	unsigned int size;
	struct user_beancounter **ub_ref;

	if (!should_uncharge(objp))
		return;

	ub_ref = slab_ub_ref(objp);
	if (*ub_ref == NULL)
		return;

	dec_slab_charged(*ub_ref, objp);
	size = CHARGE_SIZE(kmem_obj_memusage(objp));
	uncharge_beancounter(*ub_ref, UB_KMEMSIZE, size);
	put_beancounter(*ub_ref);
	*ub_ref = NULL;
}

/*
 * Pages accounting
 */

inline int ub_page_charge(struct page *page, int order, int mask)
{
	struct user_beancounter *ub;

	ub = NULL;
	if (!(mask & __GFP_UBC))
		goto out;

	ub = get_beancounter(get_exec_ub());
	if (ub == NULL)
		goto out;

	if (charge_beancounter(ub, UB_KMEMSIZE, CHARGE_ORDER(order),
				(mask & __GFP_SOFT_UBC ? UB_SOFT : UB_HARD)))
		goto err;
	if (inc_pages_charged(ub, page, order) < 0) {
		uncharge_beancounter(ub, UB_KMEMSIZE, CHARGE_ORDER(order));
		goto err;
	}
out:
	BUG_ON(page_ub(page) != NULL);
	page_ub(page) = ub;
	return 0;

err:
	BUG_ON(page_ub(page) != NULL);
	put_beancounter(ub);
	return -ENOMEM;
}

inline void ub_page_uncharge(struct page *page, int order)
{
	struct user_beancounter *ub;

	ub = page_ub(page);
	if (ub == NULL)
		return;

	dec_pages_charged(ub, page, order);
	BUG_ON(ub->ub_magic != UB_MAGIC);
	uncharge_beancounter(ub, UB_KMEMSIZE, CHARGE_ORDER(order));
	put_beancounter(ub);
	page_ub(page) = NULL;
}

/*
 * takes init_mm.page_table_lock
 * some outer lock to protect pages from vmalloced area must be held
 */
struct user_beancounter *vmalloc_ub(void *obj)
{
	struct page *pg;

	pg = vmalloc_to_page(obj);
	if (pg == NULL)
		return NULL;

	return page_ub(pg);
}

EXPORT_SYMBOL(vmalloc_ub);

struct user_beancounter *mem_ub(void *obj)
{
	struct user_beancounter *ub;

	if ((unsigned long)obj >= VMALLOC_START &&
	    (unsigned long)obj  < VMALLOC_END)
		ub = vmalloc_ub(obj);
	else
		ub = slab_ub(obj);

	return ub;
}

EXPORT_SYMBOL(mem_ub);
