/*
 *  kernel/ub/ub_misc.c
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/signal.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/kmem_cache.h>

#include <ub/beancounter.h>
#include <ub/ub_mem.h>

/*
 * Task staff
 */

static void init_task_sub(struct task_struct *tsk,
		struct task_beancounter *old_bc)
{
	struct task_beancounter *new_bc;
	struct user_beancounter *sub;

	new_bc = &tsk->task_bc;
	sub = old_bc->fork_sub;
	new_bc->fork_sub = get_beancounter(sub);
	new_bc->task_fnode = NULL;
	new_bc->task_freserv = old_bc->task_freserv;
	old_bc->task_freserv = NULL;
	memset(&new_bc->task_data, 0, sizeof(new_bc->task_data));
}

int ub_task_charge(struct task_struct *parent, struct task_struct *task)
{
	struct task_beancounter *old_bc;
	struct task_beancounter *new_bc;
	struct user_beancounter *ub;

	old_bc = &parent->task_bc;
#if 0
	if (old_bc->exec_ub == NULL) {
		/* FIXME: this won't work if task_bc is outside task_struct */
		init_task_sub(task, old_bc);
		return 0;
	}
#endif
	ub = old_bc->fork_sub;

	if (charge_beancounter(ub, UB_NUMPROC, 1, UB_HARD) < 0)
		return -ENOMEM;

	new_bc = &task->task_bc;
	new_bc->task_ub = get_beancounter(ub);
	new_bc->exec_ub = get_beancounter(ub);
	init_task_sub(task, old_bc);
	return 0;
}

void ub_task_uncharge(struct task_struct *task)
{
	struct task_beancounter *task_bc;

	task_bc = &task->task_bc;
	if (task_bc->task_ub != NULL)
		uncharge_beancounter(task_bc->task_ub, UB_NUMPROC, 1);

	put_beancounter(task_bc->exec_ub);
	put_beancounter(task_bc->task_ub);
	put_beancounter(task_bc->fork_sub);
	/* can't be freed elsewhere, failures possible in the middle of fork */
	if (task_bc->task_freserv != NULL)
		kfree(task_bc->task_freserv);

	task_bc->exec_ub = (struct user_beancounter *)0xdeadbcbc;
}

/*
 * Files and file locks.
 */

int ub_file_charge(struct file *f)
{
	struct user_beancounter *ub;

	/* No need to get_beancounter here since it's already got in slab */
	ub = slab_ub(f);
	if (ub == NULL)
		return 0;

	return charge_beancounter(ub, UB_NUMFILE, 1, UB_HARD);
}

void ub_file_uncharge(struct file *f)
{
	struct user_beancounter *ub;

	/* Ub will be put in slab */
	ub = slab_ub(f);
	if (ub == NULL)
		return;

	uncharge_beancounter(ub, UB_NUMFILE, 1);
}

int ub_flock_charge(struct file_lock *fl, int hard)
{
	struct user_beancounter *ub;
	int err;

	/* No need to get_beancounter here since it's already got in slab */
	ub = slab_ub(fl);
	if (ub == NULL)
		return 0;

	err = charge_beancounter(ub, UB_NUMFLOCK, 1, hard ? UB_HARD : UB_SOFT);
	if (!err)
		fl->fl_charged = 1;
	return err;
}

void ub_flock_uncharge(struct file_lock *fl)
{
	struct user_beancounter *ub;

	/* Ub will be put in slab */
	ub = slab_ub(fl);
	if (ub == NULL || !fl->fl_charged)
		return;

	uncharge_beancounter(ub, UB_NUMFLOCK, 1);
	fl->fl_charged = 0;
}

/*
 * Signal handling
 */

static int do_ub_siginfo_charge(struct user_beancounter *ub,
		unsigned long size)
{
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	if (__charge_beancounter_locked(ub, UB_KMEMSIZE, size, UB_HARD))
		goto out_kmem;

	if (__charge_beancounter_locked(ub, UB_NUMSIGINFO, 1, UB_HARD))
		goto out_num;

	spin_unlock_irqrestore(&ub->ub_lock, flags);
	return 0;

out_num:
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, size);
out_kmem:
	spin_unlock_irqrestore(&ub->ub_lock, flags);
	return -ENOMEM;
}

static void do_ub_siginfo_uncharge(struct user_beancounter *ub,
		unsigned long size)
{
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	__uncharge_beancounter_locked(ub, UB_KMEMSIZE, size);
	__uncharge_beancounter_locked(ub, UB_NUMSIGINFO, 1);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}

int ub_siginfo_charge(struct sigqueue *sq, struct user_beancounter *ub)
{
	unsigned long size;
	struct user_beancounter *p, *q;

	size = CHARGE_SIZE(kmem_obj_memusage(sq));
	for (p = ub; p != NULL; p = p->parent) {
		if (do_ub_siginfo_charge(p, size))
			goto unroll;
	}

	sq->sig_ub = get_beancounter(ub);
	return 0;

unroll:
	for (q = ub; q != p; q = q->parent)
		do_ub_siginfo_uncharge(q, size);
	return -ENOMEM;
}

void ub_siginfo_uncharge(struct sigqueue *sq)
{
	unsigned long size;
	struct user_beancounter *ub, *p;

	p = ub = sq->sig_ub;
	sq->sig_ub = NULL;
	size = CHARGE_SIZE(kmem_obj_memusage(sq));
	for (; ub != NULL; ub = ub->parent)
		do_ub_siginfo_uncharge(ub, size);
	put_beancounter(p);
}

/*
 * PTYs
 */

int ub_pty_charge(struct tty_struct *tty)
{
	struct user_beancounter *ub;
	int retval;

	ub = slab_ub(tty);
	retval = 0;
	if (ub && tty->driver->subtype == PTY_TYPE_MASTER &&
			!test_bit(TTY_CHARGED, &tty->flags)) {
		retval = charge_beancounter(ub, UB_NUMPTY, 1, UB_HARD);
		if (!retval)
			set_bit(TTY_CHARGED, &tty->flags);
	}
	return retval;
}

void ub_pty_uncharge(struct tty_struct *tty)
{
	struct user_beancounter *ub;

	ub = slab_ub(tty);
	if (ub && tty->driver->subtype == PTY_TYPE_MASTER &&
			test_bit(TTY_CHARGED, &tty->flags)) {
		uncharge_beancounter(ub, UB_NUMPTY, 1);
		clear_bit(TTY_CHARGED, &tty->flags);
	}
}
