/*
 *  kernel/ub/ub_page_bc.c
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/gfp.h>
#include <linux/vmalloc.h>

#include <ub/beancounter.h>
#include <ub/ub_hash.h>
#include <ub/ub_vmpages.h>
#include <ub/ub_page.h>

static kmem_cache_t *pb_cachep;
static spinlock_t pb_lock = SPIN_LOCK_UNLOCKED;
static struct page_beancounter **pb_hash_table;
static unsigned int pb_hash_mask;

/*
 * Auxiliary staff
 */

static inline struct page_beancounter *next_page_pb(struct page_beancounter *p)
{
	return list_entry(p->page_list.next, struct page_beancounter,
			page_list);
}

static inline struct page_beancounter *prev_page_pb(struct page_beancounter *p)
{
	return list_entry(p->page_list.prev, struct page_beancounter,
			page_list);
}

/*
 * Held pages manipulation
 */
static inline void set_held_pages(struct user_beancounter *bc)
{
	/* all three depend on ub_held_pages */
	__ub_update_physpages(bc);
	__ub_update_oomguarpages(bc);
	__ub_update_privvm(bc);
}

static inline void do_dec_held_pages(struct user_beancounter *ub, int value)
{
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	ub->ub_held_pages -= value;
	set_held_pages(ub);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}

static void dec_held_pages(struct user_beancounter *ub, int value)
{
	for (; ub != NULL; ub = ub->parent)
		do_dec_held_pages(ub, value);
}

static inline void do_inc_held_pages(struct user_beancounter *ub, int value)
{
	unsigned long flags;

	spin_lock_irqsave(&ub->ub_lock, flags);
	ub->ub_held_pages += value;
	set_held_pages(ub);
	spin_unlock_irqrestore(&ub->ub_lock, flags);
}

static void inc_held_pages(struct user_beancounter *ub, int value)
{
	for (; ub != NULL; ub = ub->parent)
		do_inc_held_pages(ub, value);
}

/*
 * Alloc - free
 */

inline int __pb_alloc(struct page_beancounter **pbc, gfp_t mask)
{
	*pbc = kmem_cache_alloc(pb_cachep, mask);
	if (*pbc != NULL)
		(*pbc)->pb_magic = PB_MAGIC;
	return (*pbc == NULL);
}

inline void pb_free(struct page_beancounter **pb)
{
	if (*pb != NULL) {
		kmem_cache_free(pb_cachep, *pb);
		*pb = NULL;
	}
}

void pb_free_list(struct page_beancounter **p_pb)
{
	struct page_beancounter *list = *p_pb, *pb;
	while (list) {
		pb = list;
	        list = list->next_hash;
		pb_free(&pb);
	}
	*p_pb = NULL;
}

/*
 * head -> <new objs> -> <old objs> -> ...
 */
static int __alloc_list(struct page_beancounter **head, int num)
{
	struct page_beancounter *pb;

	while (num > 0) {
		if (pb_alloc(&pb))
			return -1;
		pb->next_hash = *head;
		*head = pb;
		num--;
	}

	return num;
}

/*
 * Ensure that the list contains at least num elements.
 * p_pb points to an initialized list, may be of the zero length.
 *
 * mm->page_table_lock should be held
 */
int pb_alloc_list(struct page_beancounter **p_pb, int num)
{
	struct page_beancounter *list;

	for (list = *p_pb; list != NULL && num; list = list->next_hash, num--);
	if (!num)
		return 0;

	/*
	 *  *p_pb(after)       *p_pb (before)
	 *     \                  \
	 *     <new objs> -...-> <old objs> -> ...
	 */
	if (__alloc_list(p_pb, num) < 0)
		goto nomem;
	return 0;

nomem:
	pb_free_list(p_pb);
	return -ENOMEM;
}

/*
 * Hash routines
 */

static inline int pb_hash(struct user_beancounter *ub, struct page *page)
{
	return (((unsigned long)ub << 16) + ((unsigned long)ub >> 16) +
		(page_to_pfn(page) >> 7)) & pb_hash_mask;
}

/* pb_lock should be held */
static inline void insert_pb(struct page_beancounter *p, struct page *page,
		struct user_beancounter *ub, int hash)
{
	p->page = page;
	p->ub = get_beancounter(ub);
	p->next_hash = pb_hash_table[hash];
	pb_hash_table[hash] = p;
#ifdef CONFIG_UBC_DEBUG_KMEM
	ub->ub_stat[smp_processor_id()].pbcs++;
#endif
}

/*
 * Heart
 */

int pb_alloc_all(struct page_beancounter **pbs)
{
	int i, need_alloc;
	unsigned long flags;
	struct user_beancounter *ub;

	spin_lock_irqsave(&ub_hash_lock, flags);
	need_alloc = 0;
	for_each_beancounter(i, ub)
		need_alloc++;
	spin_unlock_irqrestore(&ub_hash_lock, flags);

	if (!__alloc_list(pbs, need_alloc))
		return 0;

	pb_free_list(pbs);
	return -ENOMEM;
}

int pb_add_ref(struct page *page, struct mm_struct *mm,
		struct page_beancounter **p_pb)
{
	int hash;
	struct user_beancounter *bc;
	struct page_beancounter *p;
	int shift;
	struct page_beancounter *head;

	bc = mm->mm_ub;
	if (bc == NULL)
		return 0;

	if (!PageAnon(page) && is_shmem_mapping(page->mapping))
		return 0;

	hash = pb_hash(bc, page);

	spin_lock(&pb_lock);
	for (p = pb_hash_table[hash];
			p != NULL && (p->page != page || p->ub != bc);
			p = p->next_hash);
	if (p != NULL) {
		/*
		 * This page is already associated with this beancounter,
		 * increment the usage counter.
		 */
		PB_COUNT_INC(p->refcount);
		spin_unlock(&pb_lock);
		return 0;
	}

	p = *p_pb;
	if (p == NULL) {
		spin_unlock(&pb_lock);
		return -1;
	}

	*p_pb = NULL;
	insert_pb(p, page, bc, hash);
	head = page_pbc(page);

	if (head != NULL) {
		/*
		 * Move the first element to the end of the list.
		 * List head (pb_head) is set to the next entry.
		 * Note that this code works even if head is the only element
		 * on the list (because it's cyclic).
		 */
		BUG_ON(head->pb_magic != PB_MAGIC);
		page_pbc(page) = next_page_pb(head);
		PB_SHIFT_INC(head->refcount);
		shift = PB_SHIFT_GET(head->refcount);
		/*
		 * Update user beancounter, the share of head has been changed.
		 * Note that the shift counter is taken after increment.
		 */
		dec_held_pages(head->ub, UB_PAGE_WEIGHT >> shift);
		/* add the new page beancounter to the end of the list */
		list_add_tail(&p->page_list, &page_pbc(page)->page_list);
	} else {
		page_pbc(page) = p;
		shift = 0;
		INIT_LIST_HEAD(&p->page_list);
	}

	p->refcount = PB_REFCOUNT_MAKE(shift, 1);
	spin_unlock(&pb_lock);

	/* update user beancounter for the new page beancounter */
	inc_held_pages(bc, UB_PAGE_WEIGHT >> shift);
	return 0;
}

void pb_remove_ref(struct page *page, struct mm_struct *mm)
{
	int hash;
	struct user_beancounter *bc;
	struct page_beancounter *p, **q;
	int shift, shiftt;

	bc = mm->mm_ub;
	if (bc == NULL)
		return;

	if (!PageAnon(page) && is_shmem_mapping(page->mapping))
		return;

	hash = pb_hash(bc, page);

	spin_lock(&pb_lock);
	BUG_ON(page_pbc(page) != NULL && page_pbc(page)->pb_magic != PB_MAGIC);
	for (q = pb_hash_table + hash, p = *q;
			p != NULL && (p->page != page || p->ub != bc);
			q = &p->next_hash, p = *q);
	if (p == NULL)
		goto out_unlock;

	PB_COUNT_DEC(p->refcount);
	if (PB_COUNT_GET(p->refcount))
		/*
		 * More references from the same user beancounter exist.
		 * Nothing needs to be done.
		 */
		goto out_unlock;

	/* remove from the hash list */
	*q = p->next_hash;

	shift = PB_SHIFT_GET(p->refcount);

	dec_held_pages(p->ub, UB_PAGE_WEIGHT >> shift);

	if (page_pbc(page) == p) {
		if (list_empty(&p->page_list))
			goto out_free;
		page_pbc(page) = next_page_pb(p);
	}
	list_del(&p->page_list);
	put_beancounter(p->ub);
#ifdef CONFIG_UBC_DEBUG_KMEM
	p->ub->ub_stat[smp_processor_id()].pbcs--;
#endif
	pb_free(&p);

	/* Now balance the list.  Move the tail and adjust its shift counter. */
	p = prev_page_pb(page_pbc(page));
	shiftt = PB_SHIFT_GET(p->refcount);
	page_pbc(page) = p;
	PB_SHIFT_DEC(p->refcount);

	inc_held_pages(p->ub, UB_PAGE_WEIGHT >> shiftt);

	/*
	 * If the shift counter of the moved beancounter is different from the
	 * removed one's, repeat the procedure for one more tail beancounter
	 */
	if (shiftt > shift) {
		p = prev_page_pb(page_pbc(page));
		page_pbc(page) = p;
		PB_SHIFT_DEC(p->refcount);
		inc_held_pages(p->ub, UB_PAGE_WEIGHT >> shiftt);
	}
	spin_unlock(&pb_lock);
	return;

out_free:
	page_pbc(page) = NULL;
#ifdef CONFIG_UBC_DEBUG_KMEM
	p->ub->ub_stat[smp_processor_id()].pbcs--;
#endif
	put_beancounter(p->ub);
	pb_free(&p);
out_unlock:
	spin_unlock(&pb_lock);
	return;
}

void pb_add_list_ref(struct page *page, struct mm_struct *mm,
		struct page_beancounter **p_pb)
{
	struct page_beancounter *list, *pb;

	pb = *p_pb;
	if (pb == NULL) {
		/* Typical case due to caller constraints */
		if (pb_add_ref(page, mm, &pb))
			BUG();
		return;
	}

	list = pb->next_hash;
	if (pb_add_ref(page, mm, &pb))
		BUG();
	if (pb != NULL) {
		pb->next_hash = list;
		list = pb;
	}
	*p_pb = list;
}

struct user_beancounter *pb_grab_page_ub(struct page *page)
{
	struct page_beancounter *pb;
	struct user_beancounter *ub;

	spin_lock(&pb_lock);
	pb = page_pbc(page);
	ub = (pb == NULL ? ERR_PTR(-EINVAL) :
			get_beancounter(pb->ub));
	spin_unlock(&pb_lock);
	return ub;
}

void __init ub_init_pbc(void)
{
	unsigned long hash_size;

	pb_cachep = kmem_cache_create("page_beancounter",
			sizeof(struct page_beancounter), 0,
			SLAB_HWCACHE_ALIGN | SLAB_PANIC, NULL, NULL);
	hash_size = num_physpages >> 2;
	for (pb_hash_mask = 1;
		(hash_size & pb_hash_mask) != hash_size;
		pb_hash_mask = (pb_hash_mask << 1) + 1);
	hash_size = pb_hash_mask + 1;
	printk(KERN_INFO "Page beancounter hash is %lu entries.\n", hash_size);
	pb_hash_table = vmalloc(hash_size * sizeof(struct page_beancounter *));
	memset(pb_hash_table, 0, hash_size * sizeof(struct page_beancounter *));
}
