/*
 *  linux/fs/proc/proc_ub.c
 *
 *  Copyright (C)  1998-2000  Andrey V. Savochkin <saw@saw.sw.com.sg>
 *  Copyright (C)  2005       SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 * TODO:
 *
 * Changes:
 */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>

#include <ub/beancounter.h>
#include <ub/ub_hash.h>
#include <ub/ub_debug.h>
#include <ub/ub_page.h>

#include <asm/page.h>
#include <asm/uaccess.h>

/*
 * we have 8 format strings depending on:
 * 1. BITS_PER_LONG
 * 2. CONFIG_UBC_KEEP_UNUSED
 * 3. resource number (see out_proc_beancounter)
 */

#ifdef CONFIG_UBC_KEEP_UNUSED
#define REF_FORMAT	"%5.5s %4i: %-12s "
#define UID_HEAD_STR	"uid ref"
#else
#define REF_FORMAT	"%10.10s: %-12s "
#define UID_HEAD_STR	"uid"
#endif
#define REF2_FORMAT	"%10s  %-12s "

#if BITS_PER_LONG == 32
#define RES_FORMAT	"%10lu %10lu %10lu %10lu %10lu"
#define HEAD_FORMAT	"%10s %10s %10s %10s %10s"
#define UB_PROC_LINE_TEXT	(10+2+12+1+10+1+10+1+10+1+10+1+10)
#else
#define RES_FORMAT	"%20lu %20lu %20lu %20lu %20lu"
#define HEAD_FORMAT	"%20s %20s %20s %20s %20s"
#define UB_PROC_LINE_TEXT	(10+2+12+1+20+1+20+1+20+1+20+1+20)
#endif

#define UB_PROC_LINE_LEN	(UB_PROC_LINE_TEXT + 1)

static void out_proc_version(char *buf)
{
	int len;

	len = sprintf(buf, "Version: 2.5");
	memset(buf + len, ' ', UB_PROC_LINE_TEXT - len);
	buf[UB_PROC_LINE_TEXT] = '\n';
}

static void out_proc_head(char *buf)
{
	sprintf(buf, REF2_FORMAT HEAD_FORMAT,
			UID_HEAD_STR, "resource", "held", "maxheld",
			"barrier", "limit", "failcnt");
	buf[UB_PROC_LINE_TEXT] = '\n';
}

static void out_proc_beancounter(char *buf, struct user_beancounter *ub, int r)
{
	if (r == 0) {
		char tmpbuf[64];
		print_ub_uid(ub, tmpbuf, sizeof(tmpbuf));
		sprintf(buf, REF_FORMAT RES_FORMAT,
			tmpbuf,
#ifdef CONFIG_UBC_KEEP_UNUSED
			atomic_read(&ub->ub_refcount),
#endif
			ub_rnames[r], ub->ub_parms[r].held,
			ub->ub_parms[r].maxheld, ub->ub_parms[r].barrier,
			ub->ub_parms[r].limit, ub->ub_parms[r].failcnt);
	} else
		sprintf(buf, REF2_FORMAT RES_FORMAT,
			"", ub_rnames[r],
			ub->ub_parms[r].held, ub->ub_parms[r].maxheld,
			ub->ub_parms[r].barrier, ub->ub_parms[r].limit,
			ub->ub_parms[r].failcnt);

	buf[UB_PROC_LINE_TEXT] = '\n';
}

static int ub_accessible(struct user_beancounter *ub,
		struct user_beancounter *exec_ub,
		struct file *file)
{
	struct user_beancounter *p, *q;

	for (p = exec_ub; p->parent != NULL; p = p->parent);
	for (q = ub; q->parent != NULL; q = q->parent);
	if (p != get_ub0() && q != p)
		return 0;
	if (ub->parent == NULL)
		return 1;
	return file->private_data == NULL ? 0 : 1;
}

static ssize_t ub_proc_read(struct file *file, char *usrbuf, size_t len,
		loff_t *poff)
{
	ssize_t retval;
	char *buf;
	unsigned long flags;
	int i, resource;
	struct ub_hash_slot *slot;
	struct user_beancounter *ub;
	struct user_beancounter *exec_ub = get_exec_ub();
	loff_t n, off;
	int rem, produced, job, tocopy;
	const int is_capable =
		(capable(CAP_DAC_OVERRIDE) || capable(CAP_DAC_READ_SEARCH));

	retval = -ENOBUFS;
	buf = (char *)__get_free_page(GFP_KERNEL);
	if (buf == NULL)
		goto out;

	retval = 0;
	if (!is_capable)
		goto out_free;

	off = *poff;
	if (off < 0) /* can't happen, just in case */
		goto inval;

again:
	i = 0;
	slot = ub_hash;
	n = off; /* The amount of data tp skip */
	produced = 0;
	if (n < (UB_PROC_LINE_LEN * 2)) {
		if (n < UB_PROC_LINE_LEN) {
			out_proc_version(buf);
			produced += UB_PROC_LINE_LEN;
			n += UB_PROC_LINE_LEN;
		}
		out_proc_head(buf + produced);
		produced += UB_PROC_LINE_LEN;
		n += UB_PROC_LINE_LEN;
	}
	n -= (2 * UB_PROC_LINE_LEN);
	spin_lock_irqsave(&ub_hash_lock, flags);
	while (1) {
		for (ub = slot->ubh_beans;
		     ub != NULL && n >= (UB_RESOURCES * UB_PROC_LINE_LEN);
		     ub = ub->ub_next)
			if (is_capable && ub_accessible(ub, exec_ub, file))
				n -= (UB_RESOURCES * UB_PROC_LINE_LEN);
		if (ub != NULL || ++i >= UB_HASH_SIZE)
			break;
		++slot;
	}
	rem = n; /* the amount of the data in the buffer to skip */
	job = PAGE_SIZE - UB_PROC_LINE_LEN + 1; /* end of buffer data */
	if (len < job - rem)
		job = rem + len;
	while (ub != NULL && produced < job) {
		if (is_capable && ub_accessible(ub, exec_ub, file))
			for (resource = 0;
				produced < job && resource < UB_RESOURCES;
				resource++, produced += UB_PROC_LINE_LEN)
			{
				out_proc_beancounter(buf + produced,
						ub, resource);
			}
		if (produced >= job)
			break;
		/* Find the next beancounter to produce more data. */
		ub = ub->ub_next;
		while (ub == NULL && ++i < UB_HASH_SIZE) {
			++slot;
			ub = slot->ubh_beans;
		}
	}

	spin_unlock_irqrestore(&ub_hash_lock, flags);
	ub_debug(UBD_ALLOC, KERN_DEBUG "UB_PROC: produced %d, job %d, rem %d\n",
			produced, job, rem);

	/*
	 * Temporary buffer `buf' contains `produced' bytes.
	 * Extract no more than `len' bytes at offset `rem'.
	 */
	if (produced <= rem)
		goto out_free;
	tocopy = produced - rem;
	if (len < tocopy)
		tocopy = len;
	if (!tocopy)
		goto out_free;
	if (copy_to_user(usrbuf, buf + rem, tocopy))
		goto fault;
	off += tocopy; /* can't overflow */
	*poff = off;
	len -= tocopy;
	retval += tocopy;
	if (!len)
		goto out_free;
	usrbuf += tocopy;
	goto again;

fault:
	retval = -EFAULT;
out_free:
	free_page((unsigned long)buf);
out:
	return retval;

inval:
	retval = -EINVAL;
	goto out_free;
}

static int ub_proc_open(struct inode *inode, struct file *file)
{
	file->private_data = strcmp(file->f_dentry->d_name.name,
						"user_beancounters") ?
						(void *)-1 : NULL;
	return 0;
}

static struct file_operations ub_file_operations = {
	.read = &ub_proc_read,
	.open = &ub_proc_open
};

#ifdef CONFIG_UBC_DEBUG_KMEM
#include <linux/seq_file.h>
#include <linux/kmem_cache.h>

static void *ubd_start(struct seq_file *m, loff_t *pos)
{
	loff_t n = *pos;
	struct user_beancounter *ub;
	long slot;

	spin_lock_irq(&ub_hash_lock);
	for (slot = 0; slot < UB_HASH_SIZE; slot++)
		for (ub = ub_hash[slot].ubh_beans; ub; ub = ub->ub_next) {
			if (n == 0) {
				m->private = (void *)slot;
				return (void *)ub;
			}
			n--;
		}
	return NULL;
}

static void *ubd_next(struct seq_file *m, void *p, loff_t *pos)
{
	struct user_beancounter *ub;
	long slot;

	ub = (struct user_beancounter *)p;
	slot = (long)m->private;

	++*pos;
	ub = ub->ub_next;
	while (1) {
		for (; ub; ub = ub->ub_next) {
			m->private = (void *)slot;
			return (void *)ub;
		}
		slot++;
		if (slot == UB_HASH_SIZE)
			break;
		ub = ub_hash[slot].ubh_beans;
	}
	return NULL;
}

static void ubd_stop(struct seq_file *m, void *p)
{
	spin_unlock_irq(&ub_hash_lock);
}

#define PROC_LINE_FMT	"\t%-17s\t%5lu\t%5lu\n"

static int ubd_show(struct seq_file *m, void *p)
{
	struct user_beancounter *ub;
	struct ub_cache_counter *cc;
	long pages, vmpages, pbc, swap, unmap;
	int i;
	char id[64];

	ub = (struct user_beancounter *)p;
	print_ub_uid(ub, id, sizeof(id));
	seq_printf(m, "%s:%d\n", id, atomic_read(&ub->ub_refcount));

	pages = vmpages = pbc = swap = unmap = 0;
	for (i = 0; i < NR_CPUS; i++) {
		pages += ub->ub_stat[i].pages_charged;
		vmpages += ub->ub_stat[i].vmalloc_charged;
		pbc += ub->ub_stat[i].pbcs;
		swap += ub->ub_stat[i].swapin;
		unmap += ub->ub_stat[i].unmap;
	}
	if (pages < 0)
		pages = 0;
	if (vmpages < 0)
		vmpages = 0;
	seq_printf(m, PROC_LINE_FMT, "pages", pages, PAGE_SIZE);
	seq_printf(m, PROC_LINE_FMT, "vmalloced", vmpages, PAGE_SIZE);

	seq_printf(m, PROC_LINE_FMT, ub_rnames[UB_UNUSEDPRIVVM],
			ub->ub_unused_privvmpages, PAGE_SIZE);
	seq_printf(m, PROC_LINE_FMT, ub_rnames[UB_TMPFSPAGES],
			ub->ub_tmpfs_respages, PAGE_SIZE);
	seq_printf(m, PROC_LINE_FMT, ub_rnames[UB_SWAPPAGES],
			ub->ub_swap_pages, PAGE_SIZE);
	seq_printf(m, PROC_LINE_FMT, "pbcs", pbc,
			(unsigned long)sizeof(struct page_beancounter));

	seq_printf(m, PROC_LINE_FMT, "swapin", swap, 0UL);
	seq_printf(m, PROC_LINE_FMT, "unmap", unmap, 0UL);
	/* interrupts are disabled by locking ub_hash_lock */
	spin_lock(&cc_lock);
	list_for_each_entry (cc, &ub->ub_cclist, ulist) {
		kmem_cache_t *cachep;

		cachep = cc->cachep;
		seq_printf(m, PROC_LINE_FMT,
				cachep->name,
				cc->counter,
				(unsigned long)cachep->objuse);
	}
	spin_unlock(&cc_lock);
	return 0;
}

static struct seq_operations kmemdebug_op = {
	.start	= ubd_start,
	.next	= ubd_next,
	.stop	= ubd_stop,
	.show	= ubd_show,
};

static int kmem_debug_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &kmemdebug_op);
}

static struct file_operations kmem_debug_ops = {
	.open		= kmem_debug_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};
#endif

void __init ub_init_proc(void)
{
	struct proc_dir_entry *entry;

	entry = create_proc_entry("user_beancounters", S_IRUGO, NULL);
	if (entry)
		entry->proc_fops = &ub_file_operations;
	else
		panic("Can't create /proc/user_beancounters entry!\n");

	entry = create_proc_entry("user_beancounters_sub", S_IRUGO, NULL);
	if (entry)
		entry->proc_fops = &ub_file_operations;
	else
		panic("Can't create /proc/user_beancounters2 entry!\n");

#ifdef CONFIG_UBC_DEBUG_KMEM
	entry = create_proc_entry("user_beancounters_debug", S_IRUGO, NULL);
	if (entry)
		entry->proc_fops = &kmem_debug_ops;
	else
		panic("Can't create /proc/user_beancounters_debug entry!\n");
#endif
}
