/*
 *  linux/kernel/vecalls.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 */

/*
 * 'vecalls.c' is file with basic VE support. It provides basic primities
 * along with initialization script
 */

#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/capability.h>
#include <linux/ve.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/ve_owner.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/sys.h>
#include <linux/fs.h>
#include <linux/namespace.h>
#include <linux/termios.h>
#include <linux/tty_driver.h>
#include <linux/netdevice.h>
#include <linux/wait.h>
#include <linux/inetdevice.h>
#include <linux/utsname.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/kernel_stat.h>
#include <linux/module.h>
#include <linux/suspend.h>
#include <linux/rcupdate.h>
#include <linux/in.h>
#include <linux/major.h>
#include <linux/kdev_t.h>
#include <linux/idr.h>
#include <linux/inetdevice.h>
#include <net/pkt_sched.h>
#include <linux/divert.h>
#include <ub/beancounter.h>

#include <net/route.h>
#include <net/ip_fib.h>

#include <linux/ve_proto.h>
#include <linux/venet.h>
#include <linux/vzctl.h>
#include <linux/vzcalluser.h>
#ifdef CONFIG_FAIRSCHED
#include <linux/fairsched.h>
#endif

#include <linux/nfcalls.h>

struct ve_struct *ve_list_head = NULL;
int nr_ve = 1;	/* One VE always exists. Compatibility with vestat */
rwlock_t ve_list_guard = RW_LOCK_UNLOCKED;
static rwlock_t devperms_hash_guard = RW_LOCK_UNLOCKED;

extern int glob_virt_pids;

static int	do_env_enter(struct ve_struct *ve);
int		real_env_create(envid_t veid, unsigned flags, u32 class_id,
				env_create_param_t *data, int datalen);
static void	do_clean_devperms(envid_t veid);
static int	alloc_ve_tty_drivers(struct ve_struct* ve);
static void	free_ve_tty_drivers(struct ve_struct* ve);
static int	register_ve_tty_drivers(struct ve_struct* ve);
static void	unregister_ve_tty_drivers(struct ve_struct* ve);
static int	init_ve_tty_drivers(struct ve_struct *);
static void	fini_ve_tty_drivers(struct ve_struct *);
static void	clear_termios(struct tty_driver* driver );
static void	ve_mapped_devs_cleanup(struct ve_struct *ve);

static int ve_get_cpu_stat(envid_t veid, struct vz_cpu_stat *buf);

static void vecalls_exit(void);

struct ve_struct *__find_ve_by_id(envid_t veid)
{
	struct ve_struct *ve;
	for (ve = ve_list_head;
	     ve != NULL && ve->veid != veid;
	     ve = ve->next);
	return ve;
}

struct ve_struct *get_ve_by_id(envid_t veid)
{
	struct ve_struct *ve;
	read_lock(&ve_list_guard);
	ve = __find_ve_by_id(veid);
	get_ve(ve);
	read_unlock(&ve_list_guard);
	return ve;
}

/*
 * real_put_ve() MUST be used instead of put_ve() inside vecalls.
 */
void real_do_env_free(struct ve_struct *ve);
static inline void real_put_ve(struct ve_struct *ve)
{
	if (ve && atomic_dec_and_test(&ve->counter)) {
		if (atomic_read(&ve->pcounter) > 0)
			BUG();
		if (ve->is_running)
			BUG();
		real_do_env_free(ve);
	}
}

extern struct file_system_type devpts_fs_type;
extern struct file_system_type sysfs_fs_type;
extern struct file_system_type tmpfs_fs_type;
extern struct file_system_type proc_fs_type;

extern spinlock_t task_capability_lock;
extern void ve_ipc_free(struct ve_struct * ve);
extern void ip_fragment_cleanup(struct ve_struct *ve);

static int ve_get_cpu_stat(envid_t veid, struct vz_cpu_stat *buf)
{
	struct ve_struct *ve;
	struct vz_cpu_stat *vstat;
	int retval;
	int i, cpu;
	unsigned long tmp;

	if (!ve_is_super(get_exec_env()) && (veid != get_exec_env()->veid))
		return -EPERM;
	if (veid == 0)
		return -ESRCH;

	vstat = kmalloc(sizeof(*vstat), GFP_KERNEL);
	if (!vstat)
		return -ENOMEM;
	memset(vstat, 0, sizeof(*vstat));

	retval = -ESRCH;
	read_lock(&ve_list_guard);
	ve = __find_ve_by_id(veid);
	if (ve == NULL)
		goto out_unlock;
	for (cpu = 0; cpu < NR_CPUS; cpu++) {
		struct ve_cpu_stats *st;

		st = VE_CPU_STATS(ve, cpu);
		vstat->user_jif += st->user;
		vstat->nice_jif += st->nice;
		vstat->system_jif += st->system;
		vstat->idle_clk += ve_sched_get_idle_time(ve, cpu);
	}
	vstat->uptime_clk = get_cycles() - ve->start_cycles;
	vstat->uptime_jif = jiffies - ve->start_jiffies;
	for (i = 0; i < 3; i++) {
		tmp = ve->avenrun[i] + (FIXED_1/200);
		vstat->avenrun[i].val_int = LOAD_INT(tmp);
		vstat->avenrun[i].val_frac = LOAD_FRAC(tmp);
	}
	read_unlock(&ve_list_guard);

	retval = 0;
	if (copy_to_user(buf, vstat, sizeof(*vstat)))
		retval = -EFAULT;
out_free:
	kfree(vstat);
	return retval;

out_unlock:
	read_unlock(&ve_list_guard);
	goto out_free;
}

/**********************************************************************
 * Devices permissions routines,
 * character and block devices separately
 **********************************************************************/

/* Rules applied in the following order:
   MAJOR!=0, MINOR!=0
   MAJOR!=0, MINOR==0
   MAJOR==0, MINOR==0
*/
struct devperms_struct
{
	dev_t   	dev;	/* device id */
	unsigned char	mask;
	unsigned 	type;
	envid_t	 	veid;

	struct devperms_struct *devhash_next;
	struct devperms_struct **devhash_pprev;
};

static struct devperms_struct original_perms[] =
{{
	MKDEV(0,0),	/*device*/
	S_IROTH | S_IWOTH,
	S_IFCHR,	/*type*/
	0,		/*veid*/
	NULL, NULL
},
{
	MKDEV(0,0),	/*device*/
	S_IXGRP | S_IROTH | S_IWOTH,
	S_IFBLK,	/*type*/
	0,		/*veid*/
	NULL, NULL
}};

static struct devperms_struct default_major_perms[] = {
	{MKDEV(UNIX98_PTY_MASTER_MAJOR, 0), S_IROTH | S_IWOTH, S_IFCHR},
	{MKDEV(UNIX98_PTY_SLAVE_MAJOR, 0), S_IROTH | S_IWOTH, S_IFCHR},
	{MKDEV(PTY_MASTER_MAJOR, 0), S_IROTH | S_IWOTH, S_IFCHR},
	{MKDEV(PTY_SLAVE_MAJOR, 0), S_IROTH | S_IWOTH, S_IFCHR},
};
static struct devperms_struct default_minor_perms[] = {
	{MKDEV(MEM_MAJOR, 3), S_IROTH | S_IWOTH, S_IFCHR},   /* null */
	{MKDEV(MEM_MAJOR, 5), S_IROTH | S_IWOTH, S_IFCHR},   /* zero */
	{MKDEV(MEM_MAJOR, 7), S_IROTH | S_IWOTH, S_IFCHR},   /* full */
	{MKDEV(TTYAUX_MAJOR, 0), S_IROTH | S_IWOTH, S_IFCHR},/* tty */
	{MKDEV(TTYAUX_MAJOR, 2), S_IROTH | S_IWOTH, S_IFCHR},/* ptmx */
	{MKDEV(MEM_MAJOR, 8), S_IROTH, S_IFCHR},  /* random */
	{MKDEV(MEM_MAJOR, 9), S_IROTH, S_IFCHR},  /* urandom */
};

static struct devperms_struct default_deny_perms = {
	MKDEV(0, 0), 0, S_IFCHR
};

static inline struct devperms_struct *find_default_devperms(int type,
						    dev_t dev)
{
	int i;

	/* XXX all defaults perms are S_IFCHR */
	if (type != S_IFCHR)
		return &default_deny_perms;

	for (i = 0;
	     i < sizeof(default_minor_perms)/sizeof(struct devperms_struct);
	     i++)
		if (MAJOR(dev) == MAJOR(default_minor_perms[i].dev) &&
		    MINOR(dev) == MINOR(default_minor_perms[i].dev))
			return &default_minor_perms[i];
	for (i = 0;
	     i < sizeof(default_major_perms)/sizeof(struct devperms_struct);
	     i++)
		if (MAJOR(dev) == MAJOR(default_major_perms[i].dev))
			return &default_major_perms[i];

	return &default_deny_perms;
}

#define DEVPERMS_HASH_SZ 512
struct devperms_struct *devperms_hash[DEVPERMS_HASH_SZ];

#define devperms_hashfn(id,dev) \
	( (id << 5) ^ (id >> 5) ^ (MAJOR(dev)) ^ MINOR(dev) ) & \
						(DEVPERMS_HASH_SZ - 1)

static inline void hash_devperms(struct devperms_struct *p)
{
	struct devperms_struct **htable =
		&devperms_hash[devperms_hashfn(p->veid,p->dev)];

	if ((p->devhash_next = *htable) != NULL)
		(*htable)->devhash_pprev = &p->devhash_next;
	*htable = p;
	p->devhash_pprev = htable;
}

static inline void unhash_devperms(struct devperms_struct *p)
{
	if (p->devhash_next)
		p->devhash_next->devhash_pprev = p->devhash_pprev;
	*p->devhash_pprev = p->devhash_next;
}

static int __init init_devperms_hash(void)
{
	write_lock_irq(&devperms_hash_guard);
	memset(devperms_hash, 0, sizeof(devperms_hash));
	hash_devperms(original_perms);
	hash_devperms(original_perms+1);
	write_unlock_irq(&devperms_hash_guard);
	return 0;
}

static inline void fini_devperms_hash(void)
{
}

static inline struct devperms_struct *find_devperms(envid_t veid,
						    int type,
						    dev_t dev)
{
	struct devperms_struct *p, **htable =
		&devperms_hash[devperms_hashfn(veid,dev)];

	for (p = *htable; p && !(p->type==type &&
				 MAJOR(dev)==MAJOR(p->dev) &&
				 MINOR(dev)==MINOR(p->dev) &&
				 p->veid==veid);
	     p = p->devhash_next)
		;
	return p;
}


static void do_clean_devperms(envid_t veid)
{
	int i;
	struct devperms_struct* ve;

	write_lock_irq(&devperms_hash_guard);
	for (i = 0; i < DEVPERMS_HASH_SZ; i++)
		for (ve = devperms_hash[i]; ve;) {
			struct devperms_struct *next = ve->devhash_next;
			if (ve->veid == veid) {
				unhash_devperms(ve);
				kfree(ve);
			}

			ve = next;
		}
	write_unlock_irq(&devperms_hash_guard);
}

/*
 * Mode is a mask of
 *	FMODE_READ	for read access (configurable by S_IROTH)
 *	FMODE_WRITE	for write access (configurable by S_IWOTH)
 *	FMODE_QUOTACTL	for quotactl access (configurable by S_IXGRP)
 */
int real_get_device_perms_ve(int dev_type, dev_t dev, int access_mode)
{
	struct devperms_struct *perms;
	struct ve_struct *ve;
	envid_t veid;

	perms = NULL;
	ve = get_exec_env();
	veid = ve->veid;

	read_lock(&devperms_hash_guard);

	perms = find_devperms(veid, dev_type|VE_USE_MINOR, dev);
	if (perms)
		goto end;

	perms = find_devperms(veid, dev_type|VE_USE_MAJOR, MKDEV(MAJOR(dev),0));
	if (perms)
		goto end;

	perms = find_devperms(veid, dev_type, MKDEV(0,0));
	if (perms)
		goto end;

	perms = find_default_devperms(dev_type, dev);

end:
	read_unlock(&devperms_hash_guard);

	access_mode = "\000\004\002\006\010\014\012\016"[access_mode];
	return perms ?
		(((perms->mask & access_mode) == access_mode) ? 0 : -EACCES) :
		-ENODEV;
}
EXPORT_SYMBOL(real_get_device_perms_ve);

int do_setdevperms(envid_t veid, unsigned type, dev_t dev, unsigned mask)
{
	struct devperms_struct   *perms;

	write_lock_irq(&devperms_hash_guard);
	perms = find_devperms(veid, type, dev);
	if (!perms) {
		struct devperms_struct   *perms_new;
		write_unlock_irq(&devperms_hash_guard);

		perms_new = kmalloc(sizeof(struct devperms_struct), GFP_KERNEL);
		if (!perms_new)
			return -ENOMEM;

		write_lock_irq(&devperms_hash_guard);
		perms = find_devperms(veid, type, dev);
		if (perms) {
			kfree(perms_new);
			perms_new = perms;
		}

		switch (type & VE_USE_MASK) {
		case 0:
			dev = 0;
			break;
		case VE_USE_MAJOR:
			dev = MKDEV(MAJOR(dev),0);
			break;
		}

		perms_new->veid = veid;
		perms_new->dev = dev;
		perms_new->type = type;
		perms_new->mask = mask & S_IALLUGO;
		hash_devperms(perms_new);
	} else
		perms->mask = mask & S_IALLUGO;
	write_unlock_irq(&devperms_hash_guard);
	return 0;
}
EXPORT_SYMBOL(do_setdevperms);

int real_setdevperms(envid_t veid, unsigned type, dev_t dev, unsigned mask)
{
	struct ve_struct *ve;
	int err;

	if (!capable(CAP_SETVEID) || veid == 0)
		return -EPERM;

	if ((ve = get_ve_by_id(veid)) == NULL)
		return -ESRCH;

	down_read(&ve->op_sem);
	err = -ESRCH;
	if (ve->is_running)
		err = do_setdevperms(veid, type, dev, mask);
	up_read(&ve->op_sem);
	real_put_ve(ve);
	return err;
}

void real_update_load_avg_ve(void)
{
	struct ve_struct *ve;
	unsigned long nr_active;

	read_lock(&ve_list_guard);
	for (ve = ve_list_head; ve != NULL; ve = ve->next) {
		nr_active = nr_running_ve(ve) + nr_uninterruptible_ve(ve);
		nr_active *= FIXED_1;
		CALC_LOAD(ve->avenrun[0], EXP_1, nr_active);
		CALC_LOAD(ve->avenrun[1], EXP_5, nr_active);
		CALC_LOAD(ve->avenrun[2], EXP_15, nr_active);
	}
	read_unlock(&ve_list_guard);
}


/**********************************************************************
 **********************************************************************
 *
 * FS-related helpers to VE start/stop
 *
 **********************************************************************
 **********************************************************************/

/*
 * DEVPTS needs a virtualization: each environment should see each own list of
 * pseudo-terminals.
 * To implement it we need to have separate devpts superblocks for each
 * VE, and each VE should mount its own one.
 * Thus, separate vfsmount structures are required.
 * To minimize intrusion into vfsmount lookup code, separate file_system_type
 * structures are created.
 *
 * In addition to this, patch fo character device itself is required, as file
 * system itself is used only for MINOR/MAJOR lookup.
 */
static int register_ve_fs_type(struct ve_struct *ve,
		struct file_system_type *template,
		struct file_system_type **p_fs_type, struct vfsmount **p_mnt)
{
	struct vfsmount *mnt;
	struct file_system_type *local_fs_type;
	int ret;

	VZTRACE("register_ve_fs_type(\"%s\")\n", template->name);

	local_fs_type = kmalloc(sizeof(*local_fs_type) + sizeof(void *),
					GFP_KERNEL);
	if (local_fs_type == NULL)
		return -ENOMEM;

	memset(local_fs_type, 0, sizeof(*local_fs_type));
	local_fs_type->name = template->name;
	local_fs_type->fs_flags = template->fs_flags;
	local_fs_type->get_sb = template->get_sb;
	local_fs_type->kill_sb = template->kill_sb;
	local_fs_type->owner = template->owner;
	/*
	 * 1. we do not have refcounter on fstype
	 * 2. fstype holds reference to ve using get_ve()/put_ve().
	 * so we free fstype when freeing ve and we are sure it's ok to free it
	 */
	SET_VE_OWNER_FSTYPE(local_fs_type, ve);
	get_filesystem(local_fs_type);	/* get_ve() inside */

	ret = register_filesystem(local_fs_type); /* does not get */
	if (ret)
		goto reg_err;

	mnt = kern_mount(local_fs_type);
	if (IS_ERR(mnt))
		goto mnt_err;

	/* Usage counters after succesful execution kern_mount:
	 * local_fs_type - +1 (get_fs_type,get_sb_single,put_filesystem)
	 * mnt - +1 == 1 (alloc_vfsmnt)
	 */

	*p_fs_type = local_fs_type;
	*p_mnt = mnt;
	return 0;

mnt_err:
	ret = PTR_ERR(mnt);
	unregister_filesystem(local_fs_type); /* does not put */

reg_err:
	put_filesystem(local_fs_type);
	kfree(local_fs_type);
	printk(KERN_DEBUG
	       "register_ve_fs_type(\"%s\") err=%d\n", template->name, ret);
	return ret;
}

static void umount_ve_fs_type(struct file_system_type *local_fs_type)
{
	struct vfsmount *mnt;
	struct list_head *p, *q;
	LIST_HEAD(kill);
	LIST_HEAD(umount_list);

	down_write(&namespace_sem);
	spin_lock(&vfsmount_lock);
	list_for_each_safe(p, q, &current->namespace->list) {
		mnt = list_entry(p, struct vfsmount, mnt_list);
		if (mnt->mnt_sb->s_type != local_fs_type)
			continue;
		list_del(p);
		list_add(p, &kill);
	}

	while (!list_empty(&kill)) {
		mnt = list_entry(kill.next, struct vfsmount, mnt_list);
		umount_tree(mnt, 1, &umount_list);
	}
	spin_unlock(&vfsmount_lock);
	up_write(&namespace_sem);
	release_mounts(&umount_list);
}

static void unregister_ve_fs_type(struct file_system_type *local_fs_type,
		struct vfsmount *local_fs_mount)
{
	if (local_fs_mount == NULL ||
	    local_fs_type == NULL) {
		if (local_fs_mount != NULL ||
		    local_fs_type != NULL)
			BUG();
		return;
	}

	VZTRACE("unregister_ve_fs_type(\"%s\")\n", local_fs_type->name);

	unregister_filesystem(local_fs_type);
	umount_ve_fs_type(local_fs_type);
	kern_umount(local_fs_mount); /* alias to mntput, drop our ref */
	put_filesystem(local_fs_type);
}


/**********************************************************************
 **********************************************************************
 *
 * FS-related helpers to VE start/stop
 *
 **********************************************************************
 **********************************************************************/

#ifdef CONFIG_SYSCTL
static ctl_table ve_sysctl_tables[] = {
	/* kernel */
	{
		.ctl_name	= CTL_KERN,
		.procname	= "kernel",
		.mode		= 0555,
		.child		= &ve_sysctl_tables[2],
	},
	{ .ctl_name = 0 },
	/* kernel/[vars] */
	{
		.ctl_name	= KERN_NODENAME,
		.procname	= "hostname",
		.maxlen 	= 64,
		.mode		= 0644,
		.proc_handler	= &proc_doutsstring,
		.strategy	= &sysctl_string,
	},
	{
		.ctl_name	= KERN_DOMAINNAME,
		.procname	= "domainname",
		.maxlen		= 64,
		.mode		= 0644,
		.proc_handler	= &proc_doutsstring,
		.strategy	= &sysctl_string,
	},
	{
		.ctl_name	= KERN_SHMMAX,
		.procname	= "shmmax",
		.maxlen		= sizeof(size_t),
		.mode		= 0644,
		.proc_handler	= &proc_doulongvec_minmax,
	},
	{
		.ctl_name	= KERN_SHMALL,
		.procname	= "shmall",
		.maxlen		= sizeof(size_t),
		.mode		= 0644,
		.proc_handler	= &proc_doulongvec_minmax,
	},
	{
		.ctl_name	= KERN_SHMMNI,
		.procname	= "shmmni",
		.maxlen		= sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{
		.ctl_name	= KERN_MSGMAX,
		.procname	= "msgmax",
		.maxlen		= sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{
		.ctl_name	= KERN_MSGMNI,
		.procname	= "msgmni",
		.maxlen		= sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{
		.ctl_name	= KERN_MSGMNB,
		.procname	= "msgmnb",
		.maxlen		= sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{
		.ctl_name	= KERN_SEM,
		.procname	= "sem",
		.maxlen		= 4 * sizeof(int),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec
	},
	{ .ctl_name = 0, }
};

static int register_ve_sysctltables(struct ve_struct *ve)
{
	struct ctl_table_header *header;
	ctl_table *root, *table;

	VZTRACE("register_ve_sysctltables\n");

	root = clone_sysctl_template(ve_sysctl_tables,
			sizeof(ve_sysctl_tables) / sizeof(ctl_table));
	if (root == NULL)
		goto out;

	table = root->child;
	table[0].data = &ve->utsname->nodename;
	table[1].data = &ve->utsname->domainname;
	table[2].data = &ve->_shm_ctlmax;
	table[3].data = &ve->_shm_ctlall;
	table[4].data = &ve->_shm_ctlmni;
	table[5].data = &ve->_msg_ctlmax;
	table[6].data = &ve->_msg_ctlmni;
	table[7].data = &ve->_msg_ctlmnb;
	table[8].data = &ve->_sem_ctls[0];

	/* insert at head to override kern entries */
	header = register_sysctl_table(root, 1);
	if (header == NULL)
		goto out_free;

	ve->kern_header = header;
	ve->kern_table = root;
	return 0;

out_free:
	free_sysctl_clone(root);
out:
	return -ENOMEM;
}

static inline void unregister_ve_sysctltables(struct ve_struct *ve)
{
	unregister_sysctl_table(ve->kern_header);
}

static inline void free_ve_sysctltables(struct ve_struct *ve)
{
	free_sysctl_clone(ve->kern_table);
}
#endif


/**********************************************************************
 **********************************************************************
 *
 * VE start: subsystems
 *
 **********************************************************************
 **********************************************************************/

#if defined(CONFIG_VE_NETDEV) || defined(CONFIG_VE_NETDEV_MODULE)
#include <net/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/icmp.h>

extern struct new_utsname virt_utsname;

static int init_ve_utsname(struct ve_struct *ve)
{
	ve->utsname = kmalloc(sizeof(*ve->utsname), GFP_KERNEL);
	if (ve->utsname == NULL)
		return -ENOMEM;

	down_read(&uts_sem); /* protect the source */
	memcpy(ve->utsname, &system_utsname, sizeof(*ve->utsname));
	memcpy(ve->utsname->release, virt_utsname.release,
			sizeof(virt_utsname.release));
	up_read(&uts_sem);

	return 0;
}

static void free_ve_utsname(struct ve_struct *ve)
{
	kfree(ve->utsname);
	ve->utsname = NULL;
}

static int init_fini_ve_mibs(struct ve_struct *ve, int fini)
{
	if (fini)
		goto fini;
	if (!(ve->_net_statistics[0] = alloc_percpu(struct linux_mib)))
		goto out1;
	if (!(ve->_net_statistics[1] = alloc_percpu(struct linux_mib)))
		goto out2;
	if (!(ve->_ip_statistics[0] = alloc_percpu(struct ipstats_mib)))
		goto out3;
	if (!(ve->_ip_statistics[1] = alloc_percpu(struct ipstats_mib)))
		goto out4;
	if (!(ve->_icmp_statistics[0] = alloc_percpu(struct icmp_mib)))
		goto out5;
	if (!(ve->_icmp_statistics[1] = alloc_percpu(struct icmp_mib)))
		goto out6;
	if (!(ve->_tcp_statistics[0] = alloc_percpu(struct tcp_mib)))
		goto out7;
	if (!(ve->_tcp_statistics[1] = alloc_percpu(struct tcp_mib)))
		goto out8;
	if (!(ve->_udp_statistics[0] = alloc_percpu(struct udp_mib)))
		goto out9;
	if (!(ve->_udp_statistics[1] = alloc_percpu(struct udp_mib)))
		goto out10;
	return 0;
fini:
	free_percpu(ve->_udp_statistics[1]);
out10:
	free_percpu(ve->_udp_statistics[0]);
out9:
	free_percpu(ve->_tcp_statistics[1]);
out8:
	free_percpu(ve->_tcp_statistics[0]);
out7:
	free_percpu(ve->_icmp_statistics[1]);
out6:
	free_percpu(ve->_icmp_statistics[0]);
out5:
	free_percpu(ve->_ip_statistics[1]);
out4:
	free_percpu(ve->_ip_statistics[0]);
out3:
	free_percpu(ve->_net_statistics[1]);
out2:
	free_percpu(ve->_net_statistics[0]);
out1:
	return -ENOMEM;
}

static inline int init_ve_mibs(struct ve_struct *ve)
{
	return init_fini_ve_mibs(ve, 0);
}

static inline void fini_ve_mibs(struct ve_struct *ve)
{
	(void)init_fini_ve_mibs(ve, 1);
}

extern struct net_device templ_loopback_dev;
static void veloop_setup(struct net_device *dev)
{
	int padded;
	padded = dev->padded;
	memcpy(dev, &templ_loopback_dev, sizeof(struct net_device));
	dev->padded = padded;
}

static int init_ve_netdev(void)
{
	struct ve_struct *ve;
	struct net_device_stats *stats;
	int err;

	ve = get_exec_env();
	INIT_HLIST_HEAD(&ve->_net_dev_head);
	ve->_net_dev_base = NULL;
	ve->_net_dev_tail = &ve->_net_dev_base;

	ve->_loopback_dev = alloc_netdev(0, templ_loopback_dev.name,
					 veloop_setup);
	if (ve->_loopback_dev == NULL)
		return -ENOMEM;
	if (loopback_dev.get_stats != NULL) {
		stats = kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
		if (stats != NULL) {
			memset(stats, 0, sizeof(struct net_device_stats));
			ve->_loopback_dev->priv = stats;
			ve->_loopback_dev->get_stats = loopback_dev.get_stats;
			ve->_loopback_dev->destructor = loopback_dev.destructor;
		}
	}
	err = register_netdev(ve->_loopback_dev);
	if (err) {
		if (ve->_loopback_dev->priv != NULL)
			kfree(ve->_loopback_dev->priv);
		free_netdev(ve->_loopback_dev);
	}
	return err;
}

static void fini_ve_netdev(void)
{
	struct ve_struct *ve;
	struct net_device *dev;

	ve = get_exec_env();
	while (1) {
		rtnl_lock();
		/*
		 * loopback is special, it can be referenced in  fib's,
		 * so it must be freed the last. Doing so is
		 * sufficient to guarantee absence of such references.
		 */
		if (dev_base == ve->_loopback_dev)
			dev = dev_base->next;
		else
			dev = dev_base;
		if (dev == NULL)
			break;
		unregister_netdevice(dev);
		rtnl_unlock();
		free_netdev(dev);
	}
	unregister_netdevice(ve->_loopback_dev);
	rtnl_unlock();
	free_netdev(ve->_loopback_dev);
	ve->_loopback_dev = NULL;
}
#else
#define init_ve_mibs(ve)	(0)
#define fini_ve_mibs(ve)	do { } while (0)
#define init_ve_netdev()	(0)
#define fini_ve_netdev()	do { } while (0)
#endif

static int prepare_proc_root(struct ve_struct *ve)
{
	struct proc_dir_entry *de;

	de = kmalloc(sizeof(struct proc_dir_entry) + 6, GFP_KERNEL);
	if (de == NULL)
		return -ENOMEM;
	memset(de, 0, sizeof(struct proc_dir_entry));
	memcpy(de + 1, "/proc", 6);
	de->name = (char *)(de + 1);
	de->namelen = 5;
	de->mode = S_IFDIR | S_IRUGO | S_IXUGO;
	de->nlink = 2;
	atomic_set(&de->count, 1);

	ve->proc_root = de;
	return 0;
}

#ifdef CONFIG_PROC_FS
static int init_ve_proc(struct ve_struct *ve)
{
	int err;
	struct proc_dir_entry *de;

	err = prepare_proc_root(ve);
	if (err)
		goto out_root;

	err = register_ve_fs_type(ve, &proc_fs_type,
			&ve->proc_fstype, &ve->proc_mnt);
	if (err)
		goto out_reg;

	/* create necessary /proc subdirs in VE local proc tree */
	err = -ENOMEM;
	de = create_proc_entry("vz", S_IFDIR|S_IRUGO|S_IXUGO, NULL);
	if (!de)
		goto out_vz;

#ifdef CONFIG_VE_IPTABLES
	proc_net = proc_mkdir("net", NULL);
	if (!proc_net)
		goto out_net;
#endif

	return 0;

#ifdef CONFIG_VE_IPTABLES
out_net:
	remove_proc_entry("vz", NULL);
#endif
out_vz:
	unregister_ve_fs_type(ve->proc_fstype, ve->proc_mnt);
	ve->proc_mnt = NULL;
out_reg:
	/* proc_fstype and proc_root are freed in real_put_ve -> free_ve_proc */
	;
out_root:
	return err;
}

static void fini_ve_proc(struct ve_struct *ve)
{
#ifdef CONFIG_VE_IPTABLES
	remove_proc_entry("net", NULL);
	proc_net =  NULL;
#endif
	remove_proc_entry("vz", NULL);
	unregister_ve_fs_type(ve->proc_fstype, ve->proc_mnt);
	ve->proc_mnt = NULL;
}

static void free_ve_proc(struct ve_struct *ve)
{
	/* proc filesystem frees proc_dir_entries on remove_proc_entry() only,
	   so we check that everything was removed and not lost */
	if (ve->proc_root && ve->proc_root->subdir) {
		struct proc_dir_entry *p = ve->proc_root;
		printk(KERN_WARNING "VPS: %d: proc entry /proc", ve->veid);
		while ((p = p->subdir) != NULL)
			printk("/%s", p->name);
		printk(" is not removed!\n");
	}

	kfree(ve->proc_root);
	kfree(ve->proc_fstype);

	ve->proc_fstype = NULL;
	ve->proc_root = NULL;
}
#else
#define init_ve_proc(ve)	(0)
#define fini_ve_proc(ve)	do { } while (0)
#define free_ve_proc(ve)	do { } while (0)
#endif

#ifdef CONFIG_SYSCTL
static int init_ve_sysctl(struct ve_struct *ve)
{
	int err;

#ifdef CONFIG_PROC_FS
	err = -ENOMEM;
	ve->proc_sys_root = proc_mkdir("sys", 0);
	if (ve->proc_sys_root == NULL)
		goto out_proc;
#endif
	INIT_LIST_HEAD(&ve->sysctl_lh);
	err = register_ve_sysctltables(ve);
	if (err)
		goto out_reg;

	err = devinet_sysctl_init(ve);
	if (err)
		goto out_dev;

	return 0;

out_dev:
	unregister_ve_sysctltables(ve);
	free_ve_sysctltables(ve);
out_reg:
#ifdef CONFIG_PROC_FS
	remove_proc_entry("sys", NULL);
out_proc:
#endif
	return err;
}

static void fini_ve_sysctl(struct ve_struct *ve)
{
	devinet_sysctl_fini(ve);
	unregister_ve_sysctltables(ve);
	remove_proc_entry("sys", NULL);
}

static void free_ve_sysctl(struct ve_struct *ve)
{
	devinet_sysctl_free(ve);
	free_ve_sysctltables(ve);
}
#else
#define init_ve_sysctl(ve)	(0)
#define fini_ve_sysctl(ve)	do { } while (0)
#define free_ve_sysctl(ve)	do { } while (0)
#endif

#ifdef CONFIG_UNIX98_PTYS
#include <linux/devpts_fs.h>

static int init_ve_devpts(struct ve_struct *ve)
{
	int err;

	err = -ENOMEM;
	ve->devpts_config = kmalloc(sizeof(struct devpts_config), GFP_KERNEL);
	if (ve->devpts_config == NULL)
		goto out;
	memset(ve->devpts_config, 0, sizeof(struct devpts_config));
	ve->devpts_config->mode = 0600;
	err = register_ve_fs_type(ve, &devpts_fs_type,
			&ve->devpts_fstype, &ve->devpts_mnt);
	if (err) {
		kfree(ve->devpts_config);
		ve->devpts_config = NULL;
	}
out:
	return err;
}

static void fini_ve_devpts(struct ve_struct *ve)
{
	unregister_ve_fs_type(ve->devpts_fstype, ve->devpts_mnt);
	/* devpts_fstype is freed in real_put_ve -> free_ve_filesystems */
	ve->devpts_mnt = NULL;
	kfree(ve->devpts_config);
	ve->devpts_config = NULL;
}
#else
#define init_ve_devpts(ve)	(0)
#define fini_ve_devpts(ve)	do { } while (0)
#endif

static int init_ve_shmem(struct ve_struct *ve)
{
	return register_ve_fs_type(ve,
				   &tmpfs_fs_type,
				   &ve->shmem_fstype,
				   &ve->shmem_mnt);
}

static void fini_ve_shmem(struct ve_struct *ve)
{
	unregister_ve_fs_type(ve->shmem_fstype, ve->shmem_mnt);
	/* shmem_fstype is freed in real_put_ve -> free_ve_filesystems */
	ve->shmem_mnt = NULL;
}

static int init_ve_sysfs(struct ve_struct *ve)
{
	struct subsystem *subsys;
	struct class *nc;
	int err;
	extern struct subsystem class_obj_subsys;
	extern struct subsystem class_subsys;
	extern struct class net_class;

#ifdef CONFIG_SYSFS
	err = 0;
	if (ve->features & VE_FEATURE_SYSFS)
		err = register_ve_fs_type(ve,
				   &sysfs_fs_type,
				   &ve->sysfs_fstype,
				   &ve->sysfs_mnt);
	if (err != 0)
		goto out_fs_type;
#endif
	err = -ENOMEM;
	subsys = kmalloc(sizeof(*subsys), GFP_KERNEL);
	if (subsys == NULL)
		goto out_class_obj;
	/* ick, this is ugly, the things we go through to keep from showing up
	 * in sysfs... */
	memset(subsys, 0, sizeof(*subsys));
	memcpy(&subsys->kset.kobj.name, &class_obj_subsys.kset.kobj.name,
			sizeof(subsys->kset.kobj.name));
	subsys->kset.ktype = class_obj_subsys.kset.ktype;
	subsys->kset.uevent_ops = class_obj_subsys.kset.uevent_ops;
	subsystem_init(subsys);
	if (!subsys->kset.subsys)
			subsys->kset.subsys = subsys;
	ve->class_obj_subsys = subsys;

	err = -ENOMEM;
	subsys = kmalloc(sizeof(*subsys), GFP_KERNEL);
	if (subsys == NULL)
		goto out_class_subsys;
	/* ick, this is ugly, the things we go through to keep from showing up
	 * in sysfs... */
	memset(subsys, 0, sizeof(*subsys));
	memcpy(&subsys->kset.kobj.name, &class_subsys.kset.kobj.name,
			sizeof(subsys->kset.kobj.name));
	subsys->kset.ktype = class_subsys.kset.ktype;
	subsys->kset.uevent_ops = class_subsys.kset.uevent_ops;
	ve->class_subsys = subsys;
	err = subsystem_register(subsys);
	if (err != 0)
		goto out_register;

	err = -ENOMEM;
	nc = kmalloc(sizeof(*nc), GFP_KERNEL);
	if (nc == NULL)
		goto out_nc;
	memset(nc, 0, sizeof(*nc));
	nc->name = net_class.name;
	nc->release = net_class.release;
	nc->uevent = net_class.uevent;
	err = class_register(nc);
	if (err != 0)
		goto out_class_register;
	ve->net_class = nc;

	return err;

out_class_register:
	kfree(nc);
out_nc:
	subsystem_unregister(subsys);
out_register:
	kfree(ve->class_subsys);
out_class_subsys:
	kfree(ve->class_obj_subsys);
out_class_obj:
#ifdef CONFIG_SYSFS
	unregister_ve_fs_type(ve->sysfs_fstype, ve->sysfs_mnt);
	/* sysfs_fstype is freed in real_put_ve -> free_ve_filesystems */
out_fs_type:
#endif
	ve->class_subsys = NULL;
	ve->class_obj_subsys = NULL;
	return err;
}

static void fini_ve_sysfs(struct ve_struct *ve)
{
	class_unregister(ve->net_class);
	subsystem_unregister(ve->class_subsys);

	kfree(ve->net_class);
	kfree(ve->class_subsys);
	kfree(ve->class_obj_subsys);

	ve->net_class = NULL;
	ve->class_subsys = NULL;
	ve->class_obj_subsys = NULL;
#ifdef CONFIG_SYSFS
	unregister_ve_fs_type(ve->sysfs_fstype, ve->sysfs_mnt);
	ve->sysfs_mnt = NULL;
	/* sysfs_fstype is freed in real_put_ve -> free_ve_filesystems */
#endif
}

static void free_ve_filesystems(struct ve_struct *ve)
{
#ifdef CONFIG_SYSFS
	kfree(ve->sysfs_fstype);
	ve->sysfs_fstype = NULL;
#endif
	kfree(ve->shmem_fstype);
	ve->shmem_fstype = NULL;

	kfree(ve->devpts_fstype);
	ve->devpts_fstype = NULL;

	free_ve_proc(ve);
}

static int init_printk(struct ve_struct *ve)
{
	struct ve_prep_printk {
		wait_queue_head_t       log_wait;
		unsigned long           log_start;
		unsigned long           log_end;
		unsigned long           logged_chars;
	} *tmp;

	tmp = kmalloc(sizeof(struct ve_prep_printk), GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;
	memset(tmp, 0, sizeof(struct ve_prep_printk));
	init_waitqueue_head(&tmp->log_wait);
	ve->_log_wait = &tmp->log_wait;
	ve->_log_start = &tmp->log_start;
	ve->_log_end = &tmp->log_end;
	ve->_logged_chars = &tmp->logged_chars;
	/* ve->log_buf will be initialized later by ve_log_init() */
	return 0;
}

static void fini_printk(struct ve_struct *ve)
{
	/*
	 * there is no spinlock protection here because nobody can use
	 * log_buf at the moments when this code is called.
	 */
	kfree(ve->log_buf);
	kfree(ve->_log_wait);
}

static void fini_venet(struct ve_struct *ve)
{
#if defined(CONFIG_VE_NETDEV) || defined(CONFIG_VE_NETDEV_MODULE)
	tcp_v4_kill_ve_sockets(ve);
#endif
#if defined(CONFIG_VE_NETDEV) || defined(CONFIG_VE_NETDEV_MODULE)
	ve_mapped_devs_cleanup(ve);
#endif
}

static int init_ve_sched(struct ve_struct *ve)
{
#ifdef CONFIG_FAIRSCHED
	int err;

	/*
	 * We refuse to switch to an already existing node since nodes
	 * keep a pointer to their ve_struct...
	 */
	err = sys_fairsched_mknod(0, 1, ve->veid);
	if (err < 0) {
		printk(KERN_WARNING "Can't create fairsched node %d\n",
				ve->veid);
		return err;
	}
	err = sys_fairsched_mvpr(current->pid, ve->veid);
	if (err) {
		printk(KERN_WARNING "Can't switch to fairsched node %d\n",
				ve->veid);
		if (sys_fairsched_rmnod(ve->veid))
			printk(KERN_ERR "Can't clean fairsched node %d\n",
					ve->veid);
		return err;
	}
#endif
	ve_sched_attach(ve);
	return 0;
}

static void fini_ve_sched(struct ve_struct *ve)
{
#ifdef CONFIG_FAIRSCHED
	if (task_vsched_id(current) == ve->veid)
		if (sys_fairsched_mvpr(current->pid, fairsched_init_node.id))
			printk(KERN_WARNING "Can't leave fairsched node %d\n",
					ve->veid);
	if (sys_fairsched_rmnod(ve->veid))
		printk(KERN_ERR "Can't remove fairsched node %d\n",
				ve->veid);
#endif
}

static int init_ve_struct(struct ve_struct *ve, envid_t veid,
		u32 class_id, env_create_param_t *data,
		struct task_struct *init_tsk)
{
	int n;

	memset(ve, 0, sizeof(*ve));
	(void)get_ve(ve);
	ve->veid = veid;
	ve->class_id = class_id;
	ve->init_entry = init_tsk;
	ve->features = data->feature_mask;
	INIT_LIST_HEAD(&ve->vetask_lh);
	init_rwsem(&ve->op_sem);
	ve->ifindex = -1;

	for(n = 0; n < UIDHASH_SZ_VE; ++n)
		INIT_LIST_HEAD(&ve->uidhash_table[n]);

	do_posix_clock_monotonic_gettime(&ve->start_timespec);
	ve->start_jiffies = jiffies;
	ve->start_cycles = get_cycles();
	ve->virt_pids = glob_virt_pids;

	return 0;
}

static void set_ve_root(struct ve_struct *ve, struct task_struct *tsk)
{
	read_lock(&tsk->fs->lock);
	ve->fs_rootmnt = tsk->fs->rootmnt;
	ve->fs_root = tsk->fs->root;
	read_unlock(&tsk->fs->lock);
	mark_tree_virtual(ve->fs_rootmnt, ve->fs_root);
}

static void set_ve_caps(struct ve_struct *ve, struct task_struct *tsk)
{
	/* required for real_setdevperms from register_ve_<fs> above */
	memcpy(&ve->cap_default, &tsk->cap_effective, sizeof(kernel_cap_t));
	cap_lower(ve->cap_default, CAP_SETVEID);
}

static int ve_list_add(struct ve_struct *ve)
{
	write_lock_irq(&ve_list_guard);
	if (__find_ve_by_id(ve->veid) != NULL)
		goto err_exists;

	ve->prev = NULL;
	ve->next = ve_list_head;
	if (ve_list_head)
		ve_list_head->prev = ve;
	ve_list_head = ve;
	nr_ve++;
	write_unlock_irq(&ve_list_guard);
	return 0;

err_exists:
	write_unlock_irq(&ve_list_guard);
	return -EEXIST;
}

static void ve_list_del(struct ve_struct *ve)
{
	write_lock_irq(&ve_list_guard);
	if (ve->prev)
		ve->prev->next = ve->next;
	else
		ve_list_head = ve->next;
	if (ve->next)
		ve->next->prev = ve->prev;
	nr_ve--;
	write_unlock_irq(&ve_list_guard);
}

static void set_task_ve_caps(struct task_struct *tsk, struct ve_struct *ve)
{
	spin_lock(&task_capability_lock);
	cap_mask(tsk->cap_effective, ve->cap_default);
	cap_mask(tsk->cap_inheritable, ve->cap_default);
	cap_mask(tsk->cap_permitted, ve->cap_default);
	spin_unlock(&task_capability_lock);
}

static void move_task(struct task_struct *tsk, struct ve_struct *new,
		struct ve_struct *old)
{
	/* this probihibts ptracing of task entered to VPS from host system */
	tsk->mm->vps_dumpable = 0;
	/* setup capabilities before enter */
	set_task_ve_caps(tsk, new);

	write_lock_irq(&tasklist_lock);
	VE_TASK_INFO(tsk)->owner_env = new;
	VE_TASK_INFO(tsk)->exec_env = new;
	REMOVE_VE_LINKS(tsk);
	SET_VE_LINKS(tsk);

	atomic_dec(&old->pcounter);
	atomic_inc(&new->pcounter);
	real_put_ve(old);
	get_ve(new);
	write_unlock_irq(&tasklist_lock);
}

#if (defined(CONFIG_VE_NETDEV) || defined(CONFIG_VE_NETDEV_MODULE)) && \
	defined(CONFIG_NETFILTER) && defined(CONFIG_VE_IPTABLES)
#define init_ve_netfilter()	init_netfilter()
#define fini_ve_netfilter()	fini_netfilter()
#else
#define init_ve_netfilter()	(0)
#define fini_ve_netfilter()	do { } while (0)
#endif

#define KSYMIPTINIT(mask, ve, full_mask, mod, name, args)	\
({								\
	int ret = 0;						\
	if (VE_IPT_CMP(mask, full_mask) &&			\
		VE_IPT_CMP((ve)->_iptables_modules, 		\
			full_mask & ~(full_mask##_MOD))) {	\
		ret = KSYMERRCALL(1, mod, name, args);		\
		if (ret == 0)					\
			(ve)->_iptables_modules |=		\
					full_mask##_MOD;	\
		if (ret == 1)					\
			ret = 0;				\
	}							\
	ret;							\
})

#define KSYMIPTFINI(mask, full_mask, mod, name, args)		\
({								\
 	if (VE_IPT_CMP(mask, full_mask##_MOD))			\
		KSYMSAFECALL_VOID(mod, name, args);		\
})


#if defined(CONFIG_VE_NETDEV) || defined(CONFIG_VE_NETDEV_MODULE)
static int do_ve_iptables(struct ve_struct *ve, __u64 init_mask,
		int init_or_cleanup)
{
	int err;

	err = 0;
	if (!init_or_cleanup)
		goto cleanup;

	/* init part */
#if defined(CONFIG_NETFILTER_XTABLES) || \
    defined(CONFIG_NETFILTER_XTABLES_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_IPTABLES,
			x_tables, init_xtables, ());
	if (err < 0)
		goto err_xtables;
#endif
#if defined(CONFIG_IP_NF_IPTABLES) || \
    defined(CONFIG_IP_NF_IPTABLES_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_IPTABLES,
			ip_tables, init_iptables, ());
	if (err < 0)
		goto err_iptables;
#endif
#if defined(CONFIG_IP_NF_CONNTRACK) || \
    defined(CONFIG_IP_NF_CONNTRACK_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_CONNTRACK,
			ip_conntrack, init_iptable_conntrack, ());
	if (err < 0)
		goto err_iptable_conntrack;
#endif
#if defined(CONFIG_IP_NF_FTP) || \
    defined(CONFIG_IP_NF_FTP_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_CONNTRACK_FTP,
			ip_conntrack_ftp, init_iptable_ftp, ());
	if (err < 0)
		goto err_iptable_ftp;
#endif
#if defined(CONFIG_IP_NF_IRC) || \
    defined(CONFIG_IP_NF_IRC_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_CONNTRACK_IRC,
			ip_conntrack_irc, init_iptable_irc, ());
	if (err < 0)
		goto err_iptable_irc;
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_CONNTRACK) || \
    defined(CONFIG_NETFILTER_XT_MATCH_CONNTRACK_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_CONNTRACK,
			xt_conntrack, init_xt_conntrack_match, ());
	if (err < 0)
		goto err_xt_conntrack_match;
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_STATE) || \
    defined(CONFIG_NETFILTER_XT_MATCH_STATE_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_STATE,
			xt_state, init_xt_state, ());
	if (err < 0)
		goto err_xt_state;
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_HELPER) || \
    defined(CONFIG_NETFILTER_XT_MATCH_HELPER_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_HELPER,
			xt_helper, init_xt_helper, ());
	if (err < 0)
		goto err_xt_helper;
#endif
#if defined(CONFIG_IP_NF_NAT) || \
    defined(CONFIG_IP_NF_NAT_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_NAT,
			ip_nat, ip_nat_init, ());
	if (err < 0)
		goto err_iptable_nat;
	err = KSYMIPTINIT(init_mask, ve, VE_IP_NAT,
			iptable_nat, init_iptable_nat, ());
	if (err < 0)
		goto err_iptable_nat2;
#endif
#if defined(CONFIG_IP_NF_NAT_FTP) || \
    defined(CONFIG_IP_NF_NAT_FTP_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_NAT_FTP,
			ip_nat_ftp, init_iptable_nat_ftp, ());
	if (err < 0)
		goto err_iptable_nat_ftp;
#endif
#if defined(CONFIG_IP_NF_NAT_IRC) || \
    defined(CONFIG_IP_NF_NAT_IRC_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_NAT_IRC,
			ip_nat_irc, init_iptable_nat_irc, ());
	if (err < 0)
		goto err_iptable_nat_irc;
#endif
#if defined(CONFIG_IP_NF_FILTER) || \
    defined(CONFIG_IP_NF_FILTER_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_FILTER,
			iptable_filter,	init_iptable_filter, ());
	if (err < 0)
		goto err_iptable_filter;
#endif
#if defined(CONFIG_IP_NF_MANGLE) || \
    defined(CONFIG_IP_NF_MANGLE_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MANGLE,
			iptable_mangle,	init_iptable_mangle, ());
	if (err < 0)
		goto err_iptable_mangle;
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_LIMIT) || \
    defined(CONFIG_NETFILTER_XT_MATCH_LIMIT_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_LIMIT,
			xt_limit, init_xt_limit, ());
	if (err < 0)
		goto err_xt_limit;
#endif
#if defined(CONFIG_IP_NF_MATCH_MULTIPORT) || \
    defined(CONFIG_IP_NF_MATCH_MULTIPORT_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_MULTIPORT,
			ipt_multiport, init_iptable_multiport, ());
	if (err < 0)
		goto err_iptable_multiport;
#endif
#if defined(CONFIG_IP_NF_MATCH_TOS) || \
    defined(CONFIG_IP_NF_MATCH_TOS_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_TOS,
			ipt_tos, init_iptable_tos, ());
	if (err < 0)
		goto err_iptable_tos;
#endif
#if defined(CONFIG_IP_NF_TARGET_TOS) || \
    defined(CONFIG_IP_NF_TARGET_TOS_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_TARGET_TOS,
			ipt_TOS, init_iptable_TOS, ());
	if (err < 0)
		goto err_iptable_TOS;
#endif
#if defined(CONFIG_IP_NF_TARGET_REJECT) || \
    defined(CONFIG_IP_NF_TARGET_REJECT_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_TARGET_REJECT,
			ipt_REJECT, init_iptable_REJECT, ());
	if (err < 0)
		goto err_iptable_REJECT;
#endif
#if defined(CONFIG_IP_NF_TARGET_TCPMSS) || \
    defined(CONFIG_IP_NF_TARGET_TCPMSS_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_TARGET_TCPMSS,
			ipt_TCPMSS, init_iptable_TCPMSS, ());
	if (err < 0)
		goto err_iptable_TCPMSS;
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_TCPMSS) || \
    defined(CONFIG_NETFILTER_XT_MATCH_TCPMSS_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_TCPMSS,
			xt_tcpmss, init_xt_tcpmss, ());
	if (err < 0)
		goto err_xt_tcpmss;
#endif
#if defined(CONFIG_IP_NF_MATCH_TTL) || \
    defined(CONFIG_IP_NF_MATCH_TTL_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_TTL,
			ipt_ttl, init_iptable_ttl, ());
	if (err < 0)
		goto err_iptable_ttl;
#endif
#if defined(CONFIG_IP_NF_TARGET_LOG) || \
    defined(CONFIG_IP_NF_TARGET_LOG_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_TARGET_LOG,
			ipt_LOG, init_iptable_LOG, ());
	if (err < 0)
		goto err_iptable_LOG;
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_LENGTH) || \
    defined(CONFIG_NETFILTER_XT_MATCH_LENGTH_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_MATCH_LENGTH,
			xt_length, init_xt_length, ());
	if (err < 0)
		goto err_xt_length;
#endif
#if defined(CONFIG_IP_NF_TARGET_REDIRECT) || \
    defined(CONFIG_IP_NF_TARGET_REDIRECT_MODULE)
	err = KSYMIPTINIT(init_mask, ve, VE_IP_TARGET_REDIRECT,
			ipt_REDIRECT, init_iptable_REDIRECT, ());
	if (err < 0)
		goto err_iptable_REDIRECT;
#endif
	return 0;

/* ------------------------------------------------------------------------- */

cleanup:
#if defined(CONFIG_IP_NF_TARGET_REDIRECT) || \
    defined(CONFIG_IP_NF_TARGET_REDIRECT_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_TARGET_REDIRECT,
			ipt_REDIRECT, fini_iptable_REDIRECT, ());
err_iptable_REDIRECT:
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_LENGTH) || \
    defined(CONFIG_NETFILTER_XT_MATCH_LENGTH_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_LENGTH,
			xt_length, fini_xt_length, ());
err_xt_length:
#endif
#if defined(CONFIG_IP_NF_TARGET_LOG) || \
    defined(CONFIG_IP_NF_TARGET_LOG_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_TARGET_LOG,
			ipt_LOG, fini_iptable_LOG, ());
err_iptable_LOG:
#endif
#if defined(CONFIG_IP_NF_MATCH_TTL) || \
    defined(CONFIG_IP_NF_MATCH_TTL_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_TTL,
			ipt_ttl, fini_iptable_ttl, ());
err_iptable_ttl:
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_TCPMSS) || \
    defined(CONFIG_NETFILTER_XT_MATCH_TCPMSS_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_TCPMSS,
			xt_tcpmss, fini_xt_tcpmss, ());
err_xt_tcpmss:
#endif
#if defined(CONFIG_IP_NF_TARGET_TCPMSS) || \
    defined(CONFIG_IP_NF_TARGET_TCPMSS_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_TARGET_TCPMSS,
			ipt_TCPMSS, fini_iptable_TCPMSS, ());
err_iptable_TCPMSS:
#endif
#if defined(CONFIG_IP_NF_TARGET_REJECT) || \
    defined(CONFIG_IP_NF_TARGET_REJECT_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_TARGET_REJECT,
			ipt_REJECT, fini_iptable_REJECT, ());
err_iptable_REJECT:
#endif
#if defined(CONFIG_IP_NF_TARGET_TOS) || \
    defined(CONFIG_IP_NF_TARGET_TOS_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_TARGET_TOS,
			ipt_TOS, fini_iptable_TOS, ());
err_iptable_TOS:
#endif
#if defined(CONFIG_IP_NF_MATCH_TOS) || \
    defined(CONFIG_IP_NF_MATCH_TOS_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_TOS,
			ipt_tos, fini_iptable_tos, ());
err_iptable_tos:
#endif
#if defined(CONFIG_IP_NF_MATCH_MULTIPORT) || \
    defined(CONFIG_IP_NF_MATCH_MULTIPORT_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_MULTIPORT,
			ipt_multiport, fini_iptable_multiport, ());
err_iptable_multiport:
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_LIMIT) || \
    defined(CONFIG_NETFILTER_XT_MATCH_LIMIT_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_LIMIT,
			xt_limit, fini_xt_limit, ());
err_xt_limit:
#endif
#if defined(CONFIG_IP_NF_MANGLE) || \
    defined(CONFIG_IP_NF_MANGLE_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MANGLE,
			iptable_mangle,	fini_iptable_mangle, ());
err_iptable_mangle:
#endif
#if defined(CONFIG_IP_NF_FILTER) || \
    defined(CONFIG_IP_NF_FILTER_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_FILTER,
			iptable_filter,	fini_iptable_filter, ());
err_iptable_filter:
#endif
#if defined(CONFIG_IP_NF_NAT_IRC) || \
    defined(CONFIG_IP_NF_NAT_IRC_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_NAT_IRC,
			ip_nat_irc, fini_iptable_nat_irc, ());
err_iptable_nat_irc:
#endif
#if defined(CONFIG_IP_NF_NAT_FTP) || \
    defined(CONFIG_IP_NF_NAT_FTP_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_NAT_FTP,
			ip_nat_ftp, fini_iptable_nat_ftp, ());
err_iptable_nat_ftp:
#endif
#if defined(CONFIG_IP_NF_NAT) || \
    defined(CONFIG_IP_NF_NAT_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_NAT,
			iptable_nat, fini_iptable_nat, ());
err_iptable_nat2:
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_NAT,
			ip_nat, ip_nat_cleanup, ());
err_iptable_nat:
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_HELPER) || \
    defined(CONFIG_NETFILTER_XT_MATCH_HELPER_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_HELPER,
			xt_helper, fini_xt_helper, ());
err_xt_helper:
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_STATE) || \
    defined(CONFIG_NETFILTER_XT_MATCH_STATE_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_STATE,
			xt_state, fini_xt_state, ());
err_xt_state:
#endif
#if defined(CONFIG_NETFILTER_XT_MATCH_CONNTRACK) || \
    defined(CONFIG_NETFILTER_XT_MATCH_CONNTRACK_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MATCH_CONNTRACK,
			xt_conntrack, fini_xt_conntrack_match, ());
err_xt_conntrack_match:
#endif
#if defined(CONFIG_IP_NF_IRC) || \
    defined(CONFIG_IP_NF_IRC_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_CONNTRACK_IRC,
			ip_conntrack_irc, fini_iptable_irc, ());
err_iptable_irc:
#endif
#if defined(CONFIG_IP_NF_FTP) || \
    defined(CONFIG_IP_NF_FTP_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_CONNTRACK_FTP,
			ip_conntrack_ftp, fini_iptable_ftp, ());
err_iptable_ftp:
#endif
#if defined(CONFIG_IP_NF_CONNTRACK) || \
    defined(CONFIG_IP_NF_CONNTRACK_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_CONNTRACK,
			ip_conntrack, fini_iptable_conntrack, ());
err_iptable_conntrack:
#endif
#if defined(CONFIG_IP_NF_IPTABLES) || \
    defined(CONFIG_IP_NF_IPTABLES_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_IPTABLES,
			ip_tables, fini_iptables, ());
err_iptables:
#endif
#if defined(CONFIG_NETFILTER_XTABLES) || \
    defined(CONFIG_NETFILTER_XTABLES_MODULE)
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_IPTABLES,
			x_tables, fini_xtables, ());
err_xtables:
#endif
	ve->_iptables_modules = 0;

	return err;
}
#else
#define do_ve_iptables(ve, initmask, init)	(0)
#endif

static inline int init_ve_iptables(struct ve_struct *ve, __u64 init_mask)
{
	return do_ve_iptables(ve, init_mask, 1);
}

static inline void fini_ve_iptables(struct ve_struct *ve, __u64 init_mask)
{
	(void)do_ve_iptables(ve, init_mask, 0);
}

static void flush_ve_iptables(struct ve_struct *ve)
{
	/*
	 * flush all rule tables first,
	 * this helps us to avoid refs to freed objs
	 */
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_MANGLE, ip_tables,
			ipt_flush_table, (ve->_ipt_mangle_table));
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_FILTER, ip_tables,
			ipt_flush_table, (ve->_ve_ipt_filter_pf));
	KSYMIPTFINI(ve->_iptables_modules, VE_IP_NAT, ip_tables,
			ipt_flush_table, (ve->_ip_conntrack->_ip_nat_table));
}

static struct list_head ve_hooks[VE_MAX_HOOKS];
static DECLARE_RWSEM(ve_hook_sem);

int ve_hook_register(struct ve_hook *vh)
{
	struct list_head *lh;
	struct ve_hook *tmp;

	down_write(&ve_hook_sem);
	list_for_each(lh, &ve_hooks[vh->hooknum]) {
		tmp = list_entry(lh, struct ve_hook, list);
		if (vh->priority < tmp->priority)
			break;
	}
	list_add_tail(&vh->list, lh);
	up_write(&ve_hook_sem);
	return 0;
}
EXPORT_SYMBOL(ve_hook_register);

void ve_hook_unregister(struct ve_hook *vh)
{
	down_write(&ve_hook_sem);
	list_del(&vh->list);
	up_write(&ve_hook_sem);
}
EXPORT_SYMBOL(ve_hook_unregister);

static int ve_hook_iterate(unsigned int hooknum, void *data)
{
	struct ve_hook *vh;
	int err;

	err = 0;
	down_read(&ve_hook_sem);
	list_for_each_entry(vh, &ve_hooks[hooknum], list) {
		if (!try_module_get(vh->owner))
			continue;
		err = vh->hook(hooknum, data);
		module_put(vh->owner);
		if (err)
			break;
	}

	if (err) {
		list_for_each_entry_continue_reverse(vh,
					&ve_hooks[hooknum], list) {
			if (!try_module_get(vh->owner))
				continue;
			if (vh->undo)
				vh->undo(hooknum, data);
			module_put(vh->owner);
		}
	}
	up_read(&ve_hook_sem);
	return err;
}

static void ve_hook_iterate_cleanup(unsigned int hooknum, void *data)
{
	struct ve_hook *vh;

	down_read(&ve_hook_sem);
	list_for_each_entry_reverse(vh, &ve_hooks[hooknum], list) {
		if (!try_module_get(vh->owner))
			continue;
		(void)vh->hook(hooknum, data);
		module_put(vh->owner);
	}
	up_read(&ve_hook_sem);
}

static int do_env_create(envid_t veid, u32 class_id,
			 env_create_param_t *data, int datalen)
{
	struct task_struct *tsk;
	struct ve_struct *old;
	struct ve_struct *old_exec;
	struct ve_struct *ve;
 	__u64 init_mask;
	int err;

	tsk = current;
	old = VE_TASK_INFO(tsk)->owner_env;

	if (!thread_group_leader(tsk))
		return -EINVAL;

	if (tsk->signal->tty) {
		printk("ERR: VE init has controlling terminal\n");
		return -EINVAL;
	}
	if (tsk->signal->pgrp != tsk->pid || tsk->signal->session != tsk->pid) {
		int may_setsid;
		read_lock(&tasklist_lock);
		may_setsid = (find_pid(PIDTYPE_PGID, tsk->pid) == NULL);
		read_unlock(&tasklist_lock);
		if (!may_setsid) {
			printk("ERR: VE init is process group leader\n");
			return -EINVAL;
		}
	}


	VZTRACE("%s: veid=%d classid=%d pid=%d\n",
		__FUNCTION__, veid, class_id, current->pid);

	err = -ENOMEM;
	ve = kmalloc(sizeof(struct ve_struct), GFP_KERNEL);
	if (ve == NULL)
		goto err_struct;

	init_ve_struct(ve, veid, class_id, data, tsk);
	__module_get(THIS_MODULE);
	down_write(&ve->op_sem);
	if ((err = ve_list_add(ve)) < 0)
		goto err_exist;

	/* this should be done before context switching */
	if ((err = init_printk(ve)) < 0)
		goto err_log_wait;

	old_exec = set_exec_env(ve);

	if ((err = init_ve_sched(ve)) < 0)
		goto err_sched;

	/* move user to VE */
	if ((err = set_user(0, 0)) < 0)
		goto err_set_user;

	set_ve_root(ve, tsk);

	if ((err = init_ve_utsname(ve)))
		goto err_utsname;

	if ((err = init_ve_mibs(ve)))
		goto err_mibs;

	if ((err = init_ve_proc(ve)))
		goto err_proc;

	if ((err = init_ve_sysctl(ve)))
		goto err_sysctl;

	if ((err = init_ve_sysfs(ve)))
		goto err_sysfs;

	if ((err = init_ve_route(ve)) < 0)
		goto err_route;

	if ((err = init_ve_netdev()))
		goto err_dev;

	if ((err = init_ve_tty_drivers(ve)) < 0)
		goto err_tty;

	if ((err = init_ve_shmem(ve)))
		goto err_shmem;

	if ((err = init_ve_devpts(ve)))
		goto err_devpts;

	/* init SYSV IPC variables */
	if ((err = init_ve_ipc(ve)) < 0)
		goto err_ipc;

	set_ve_caps(ve, tsk);

	/* It is safe to initialize netfilter here as routing initialization and
	   interface setup will be done below. This means that NO skb can be
	   passed inside. Den */
	/* iptables ve initialization for non ve0;
	   ve0 init is in module_init */
	if ((err = init_ve_netfilter()) < 0)
		goto err_netfilter;

	init_mask = data ? data->iptables_mask : VE_IP_DEFAULT;
	if ((err = init_ve_iptables(ve, init_mask)) < 0)
		goto err_iptables;

	if ((err = alloc_vpid(tsk->pid, 1)) < 0)
		goto err_vpid;

	if ((err = ve_hook_iterate(VE_HOOK_INIT, (void *)ve)) < 0)
		goto err_ve_hook;

	/* finally: set vpids and move inside */
	move_task(tsk, ve, old);

	set_virt_pid(tsk, 1);
	set_virt_tgid(tsk, 1);

	set_special_pids(tsk->pid, tsk->pid);
	current->signal->tty_old_pgrp = 0;
	set_virt_pgid(tsk, 1);
	set_virt_sid(tsk, 1);

	ve->is_running = 1;
	up_write(&ve->op_sem);

	printk(KERN_INFO "VPS: %d: started\n", veid);
	return veid;

err_ve_hook:
	free_vpid(1, ve);
err_vpid:
	fini_venet(ve);
	fini_ve_iptables(ve, init_mask);
err_iptables:
	fini_ve_netfilter();
err_netfilter:
	fini_ve_ipc(ve);
err_ipc:
	fini_ve_devpts(ve);
err_devpts:
	fini_ve_shmem(ve);
err_shmem:
	fini_ve_tty_drivers(ve);
err_tty:
	fini_ve_netdev();
err_dev:
	fini_ve_route(ve);
err_route:
	fini_ve_sysfs(ve);
err_sysfs:
	fini_ve_sysctl(ve);
err_sysctl:
	fini_ve_proc(ve);
err_proc:
	do_clean_devperms(ve->veid); /* register procfs adds devperms */
	fini_ve_mibs(ve);
err_mibs:
	/* free_ve_utsname() is called inside real_put_ve() */ ;
err_utsname:
	/* It is safe to restore current->envid here because
	 * ve_fairsched_detach does not use current->envid. */
	/* Really fairsched code uses current->envid in sys_fairsched_mknod
	 * only.  It is correct if sys_fairsched_mknod is called from
	 * userspace.  If sys_fairsched_mknod is called from
	 * ve_fairsched_attach, then node->envid and node->parent_node->envid
	 * are explicitly set to valid value after the call. */
	/* FIXME */
	VE_TASK_INFO(tsk)->owner_env = old;
	VE_TASK_INFO(tsk)->exec_env = old_exec;
	/* move user back */
	if (set_user(0, 0) < 0)
		printk(KERN_WARNING"Can't restore UID\n");

err_set_user:
	fini_ve_sched(ve);
err_sched:
	(void)set_exec_env(old_exec);

	/* we can jump here having incorrect envid */
	VE_TASK_INFO(tsk)->owner_env = old;
	fini_printk(ve);
err_log_wait:
	ve_list_del(ve);
	up_write(&ve->op_sem);

	real_put_ve(ve);
err_struct:
	printk(KERN_INFO "VPS: %d: failed to start with err=%d\n", veid, err);
	return err;

err_exist:
	kfree(ve);
	goto err_struct;
}


/**********************************************************************
 **********************************************************************
 *
 * VE start/stop callbacks
 *
 **********************************************************************
 **********************************************************************/

int real_env_create(envid_t veid, unsigned flags, u32 class_id,
			env_create_param_t *data, int datalen)
{
	int status;
	struct ve_struct *ve;

	if (!flags) {
		status = get_exec_env()->veid;
		goto out;
	}

	status = -EPERM;
	if (!capable(CAP_SETVEID))
		goto out;

	status = -EINVAL;
	if ((flags & VE_TEST) && (flags & (VE_ENTER|VE_CREATE)))
		goto out;

	status = -EINVAL;
	ve = get_ve_by_id(veid);
	if (ve) {
		if (flags & VE_TEST) {
			status = 0;
			goto out_put;
		}
		if (flags & VE_EXCLUSIVE) {
			status = -EACCES;
			goto out_put;
		}
		if (flags & VE_CREATE) {
			flags &= ~VE_CREATE;
			flags |= VE_ENTER;
		}
	} else {
		if (flags & (VE_TEST|VE_ENTER)) {
			status = -ESRCH;
			goto out;
		}
	}

	if (flags & VE_CREATE) {
		status = do_env_create(veid, class_id, data, datalen);
		goto out;
	} else if (flags & VE_ENTER)
		status = do_env_enter(ve);

	/* else: returning EINVAL */

out_put:
	real_put_ve(ve);
out:
	return status;
}

static int do_env_enter(struct ve_struct *ve)
{
	struct task_struct *tsk = current;
	int err;

	VZTRACE("%s: veid=%d\n", __FUNCTION__, ve->veid);

	err = -EBUSY;
	down_read(&ve->op_sem);
	if (!ve->is_running)
		goto out_up;

#ifdef CONFIG_FAIRSCHED
	err = sys_fairsched_mvpr(current->pid, ve->veid);
	if (err)
		goto out_up;
#endif

	ve_sched_attach(ve);
	move_task(current, ve, VE_TASK_INFO(tsk)->owner_env);
	err = VE_TASK_INFO(tsk)->owner_env->veid;

out_up:
	up_read(&ve->op_sem);
	return err;
}

static void env_cleanup(struct ve_struct *ve)
{
	struct ve_struct *old_ve;

	VZTRACE("real_do_env_cleanup\n");

	down_read(&ve->op_sem);
	old_ve = set_exec_env(ve);

	ve_hook_iterate_cleanup(VE_HOOK_FINI, (void *)ve);

	fini_venet(ve);

	/* no new packets in flight beyond this point */
	synchronize_net();
	/* skb hold dst_entry, and in turn lies in the ip fragment queue */
	ip_fragment_cleanup(ve);

	fini_ve_netdev();
	fini_ve_route(ve);

	/* kill iptables */
	/* No skb belonging to VE can exist at this point as unregister_netdev
	   is an operation awaiting until ALL skb's gone */
	flush_ve_iptables(ve);
	fini_ve_iptables(ve, ve->_iptables_modules);
	fini_ve_netfilter();

	ve_ipc_cleanup();

	fini_ve_sched(ve);
	do_clean_devperms(ve->veid);

	fini_ve_devpts(ve);
	fini_ve_shmem(ve);
	fini_ve_sysfs(ve);
	unregister_ve_tty_drivers(ve);
	fini_ve_sysctl(ve);
	fini_ve_proc(ve);

	fini_ve_mibs(ve);

	(void)set_exec_env(old_ve);
	fini_printk(ve);	/* no printk can happen in ve context anymore */

	ve_list_del(ve);
	up_read(&ve->op_sem);

	real_put_ve(ve);
}

static struct list_head ve_cleanup_list;
static spinlock_t ve_cleanup_lock;

static DECLARE_COMPLETION(vzmond_complete);
static struct task_struct *vzmond_thread;
static volatile int stop_vzmond;

void real_do_env_cleanup(struct ve_struct *ve)
{
	spin_lock(&ve_cleanup_lock);
	list_add_tail(&ve->cleanup_list, &ve_cleanup_list);
	spin_unlock(&ve_cleanup_lock);
	wake_up_process(vzmond_thread);
}

static void do_pending_env_cleanups(void)
{
	struct ve_struct *ve;

	spin_lock(&ve_cleanup_lock);
	while (1) {
		if (list_empty(&ve_cleanup_list) || need_resched())
			break;
		ve = list_entry(ve_cleanup_list.next, struct ve_struct,
				cleanup_list);
		list_del(&ve->cleanup_list);
		spin_unlock(&ve_cleanup_lock);
		env_cleanup(ve);
		spin_lock(&ve_cleanup_lock);
	}
	spin_unlock(&ve_cleanup_lock);
}

static int have_pending_cleanups(void)
{
	return !list_empty(&ve_cleanup_list);
}

static int vzmond(void *arg)
{
	daemonize("vzmond");
	vzmond_thread = current;
	set_current_state(TASK_INTERRUPTIBLE);

	while (!stop_vzmond) {
		schedule();
		try_to_freeze();
		if (signal_pending(current))
			flush_signals(current);

		do_pending_env_cleanups();
		set_current_state(TASK_INTERRUPTIBLE);
		if (have_pending_cleanups())
			__set_current_state(TASK_RUNNING);
	}

	__set_task_state(current, TASK_RUNNING);
	complete_and_exit(&vzmond_complete, 0);
}

static int __init init_vzmond(void)
{
	INIT_LIST_HEAD(&ve_cleanup_list);
	spin_lock_init(&ve_cleanup_lock);
	stop_vzmond = 0;
	return kernel_thread(vzmond, NULL, 0);
}

static void fini_vzmond(void)
{
	stop_vzmond = 1;
	wake_up_process(vzmond_thread);
	wait_for_completion(&vzmond_complete);
	WARN_ON(!list_empty(&ve_cleanup_list));
}

void real_do_env_free(struct ve_struct *ve)
{
	VZTRACE("real_do_env_free\n");

	ve_ipc_free(ve); /* free SYSV IPC resources */
	free_ve_tty_drivers(ve);
	free_ve_utsname(ve);
	free_ve_sysctl(ve); /* free per ve sysctl data */
	free_ve_filesystems(ve);
	printk(KERN_INFO "VPS: %d: stopped\n", VEID(ve));
	kfree(ve);

	module_put(THIS_MODULE);
}
EXPORT_SYMBOL(real_do_env_free);


/**********************************************************************
 **********************************************************************
 *
 * VE TTY handling
 *
 **********************************************************************
 **********************************************************************/

DCL_VE_OWNER(TTYDRV, struct tty_driver, owner_env)

static struct tty_driver *alloc_ve_tty_driver(struct tty_driver *base,
					   struct ve_struct *ve)
{
	size_t size;
	struct tty_driver *driver;

	driver = kmalloc(sizeof(struct tty_driver), GFP_KERNEL);
	if (!driver)
		goto out;

	memcpy(driver, base, sizeof(struct tty_driver));

	driver->driver_state = NULL;

	size = base->num * 3 * sizeof(void *);
	if (!(driver->flags & TTY_DRIVER_DEVPTS_MEM)) {
		void **p;
		p = kmalloc(size, GFP_KERNEL);
		if (!p)
			goto out_free;
		memset(p, 0, size);
		driver->ttys = (struct tty_struct **)p;
		driver->termios = (struct termios **)(p + driver->num);
		driver->termios_locked = (struct termios **)(p + driver->num * 2);
	} else {
		driver->ttys = NULL;
		driver->termios = NULL;
		driver->termios_locked = NULL;
	}

	SET_VE_OWNER_TTYDRV(driver, ve);
	driver->flags |= TTY_DRIVER_INSTALLED;

	return driver;

out_free:
	kfree(driver);
out:
	return NULL;
}

static void free_ve_tty_driver(struct tty_driver *driver)
{
	if (!driver)
		return;

	clear_termios(driver);
	kfree(driver->ttys);
	kfree(driver);
}

static int alloc_ve_tty_drivers(struct ve_struct* ve)
{
#ifdef CONFIG_LEGACY_PTYS
	/* Traditional BSD devices */
	ve->pty_driver = alloc_ve_tty_driver(pty_driver, ve);
	if (!ve->pty_driver)
		goto out_mem;

	ve->pty_slave_driver = alloc_ve_tty_driver(pty_slave_driver, ve);
	if (!ve->pty_slave_driver)
		goto out_mem;

	ve->pty_driver->other       = ve->pty_slave_driver;
	ve->pty_slave_driver->other = ve->pty_driver;
#endif

#ifdef CONFIG_UNIX98_PTYS
	ve->ptm_driver = alloc_ve_tty_driver(ptm_driver, ve);
	if (!ve->ptm_driver)
		goto out_mem;

	ve->pts_driver = alloc_ve_tty_driver(pts_driver, ve);
	if (!ve->pts_driver)
		goto out_mem;

	ve->ptm_driver->other = ve->pts_driver;
	ve->pts_driver->other = ve->ptm_driver;

	ve->allocated_ptys = kmalloc(sizeof(*ve->allocated_ptys), GFP_KERNEL);
	if (!ve->allocated_ptys)
		goto out_mem;
	idr_init(ve->allocated_ptys);
#endif
	return 0;

out_mem:
	free_ve_tty_drivers(ve);
	return -ENOMEM;
}

static void free_ve_tty_drivers(struct ve_struct* ve)
{
#ifdef CONFIG_LEGACY_PTYS
	free_ve_tty_driver(ve->pty_driver);
	free_ve_tty_driver(ve->pty_slave_driver);
	ve->pty_driver = ve->pty_slave_driver = NULL;
#endif
#ifdef CONFIG_UNIX98_PTYS
	free_ve_tty_driver(ve->ptm_driver);
	free_ve_tty_driver(ve->pts_driver);
	kfree(ve->allocated_ptys);
	ve->ptm_driver = ve->pts_driver = NULL;
	ve->allocated_ptys = NULL;
#endif
}

static inline void __register_tty_driver(struct tty_driver *driver)
{
	list_add(&driver->tty_drivers, &tty_drivers);
}

static inline void __unregister_tty_driver(struct tty_driver *driver)
{
	if (!driver)
		return;
	list_del(&driver->tty_drivers);
}

static int register_ve_tty_drivers(struct ve_struct* ve)
{
	write_lock_irq(&tty_driver_guard);
#ifdef CONFIG_UNIX98_PTYS
	__register_tty_driver(ve->ptm_driver);
	__register_tty_driver(ve->pts_driver);
#endif
#ifdef CONFIG_LEGACY_PTYS
	__register_tty_driver(ve->pty_driver);
	__register_tty_driver(ve->pty_slave_driver);
#endif
	write_unlock_irq(&tty_driver_guard);

	return 0;
}

static void unregister_ve_tty_drivers(struct ve_struct* ve)
{
	VZTRACE("unregister_ve_tty_drivers\n");

	write_lock_irq(&tty_driver_guard);
	__unregister_tty_driver(ve->pty_driver);
	__unregister_tty_driver(ve->pty_slave_driver);
#ifdef CONFIG_UNIX98_PTYS
	__unregister_tty_driver(ve->ptm_driver);
	__unregister_tty_driver(ve->pts_driver);
#endif
	write_unlock_irq(&tty_driver_guard);
}

static int init_ve_tty_drivers(struct ve_struct *ve)
{
	int err;

	if ((err = alloc_ve_tty_drivers(ve)))
		goto err_ttyalloc;
	if ((err = register_ve_tty_drivers(ve)))
		goto err_ttyreg;
	return 0;

err_ttyreg:
	free_ve_tty_drivers(ve);
err_ttyalloc:
	return err;
}

static void fini_ve_tty_drivers(struct ve_struct *ve)
{
	unregister_ve_tty_drivers(ve);
	free_ve_tty_drivers(ve);
}

/*
 * Free the termios and termios_locked structures because
 * we don't want to get memory leaks when modular tty
 * drivers are removed from the kernel.
 */
static void clear_termios(struct tty_driver *driver)
{
	int i;
	struct termios *tp;

	if (driver->termios == NULL)
		return;
	for (i = 0; i < driver->num; i++) {
		tp = driver->termios[i];
		if (tp) {
			driver->termios[i] = NULL;
			kfree(tp);
		}
		tp = driver->termios_locked[i];
		if (tp) {
			driver->termios_locked[i] = NULL;
			kfree(tp);
		}
	}
}


/**********************************************************************
 **********************************************************************
 *
 * Pieces of VE network
 *
 **********************************************************************
 **********************************************************************/

#if defined(CONFIG_VE_NETDEV) || defined(CONFIG_VE_NETDEV_MODULE)
#include <asm/uaccess.h>
#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <net/route.h>
#include <net/ip_fib.h>
#endif

#if defined(CONFIG_VE_NETDEV) || defined(CONFIG_VE_NETDEV_MODULE)
static void ve_del_ip_addrs(struct net_device *dev)
{
	struct in_device *in_dev;

	in_dev = in_dev_get(dev);
	if (in_dev == NULL)
		return;

	while (in_dev->ifa_list != NULL) {
		inet_del_ifa(in_dev, &in_dev->ifa_list, 1);
	}
	in_dev_put(in_dev);
}

static int ve_netdev_cleanup(struct net_device *dev, int to_ve)
{
	int err;

	err = 0;
	ve_del_ip_addrs(dev);
	if ((dev->flags & IFF_UP) != 0)
		err = dev_close(dev);
	synchronize_net();
	dev_shutdown(dev);
	dev_mc_discard(dev);
	free_divert_blk(dev);
	synchronize_net();

	if (to_ve)
		dev->orig_mtu = dev->mtu;
	else {
		int rc = dev_set_mtu(dev, dev->orig_mtu);
		if (err == 0)
			err = rc;
	}

	return err;
}

static void __ve_dev_move(struct net_device *dev, struct ve_struct *ve_src,
	struct ve_struct *ve_dst, struct user_beancounter *exec_ub)
{
	struct net_device **dp, *d;
	struct user_beancounter *ub;

	for (d = ve_src->_net_dev_base, dp = NULL; d != NULL;
	     dp = &d->next, d = d->next) {
		if (d == dev) {
			hlist_del(&dev->name_hlist);
			hlist_del(&dev->index_hlist);
			if (ve_src->_net_dev_tail == &dev->next)
				ve_src->_net_dev_tail = dp;
			if (dp)
				*dp = dev->next;
			dev->next = NULL;
			break;
		}
	}
	*ve_dst->_net_dev_tail = dev;
	ve_dst->_net_dev_tail = &dev->next;
	hlist_add_head(&dev->name_hlist, dev_name_hash(dev->name, ve_dst));
	hlist_add_head(&dev->index_hlist, dev_index_hash(dev->ifindex, ve_dst));
	dev->owner_env = ve_dst;

	ub = netdev_bc(dev)->exec_ub;
	netdev_bc(dev)->exec_ub = get_beancounter(exec_ub);
	put_beancounter(ub);
}

static int ve_dev_add(envid_t veid, char *dev_name)
{
	int err;
	struct net_device *dev;
	struct ve_struct *ve;
	struct hlist_node *p;

	dev = NULL;
	err = -ESRCH;

	ve = get_ve_by_id(veid);
	if (ve == NULL)
		goto out;

	rtnl_lock();

	read_lock(&dev_base_lock);
	hlist_for_each(p, dev_name_hash(dev_name, get_ve0())) {
		struct net_device *d = hlist_entry(p, struct net_device,
						   name_hlist);
		if (strncmp(d->name, dev_name, IFNAMSIZ) == 0) {
			dev = d;
			break;
		}
	}
	read_unlock(&dev_base_lock);
	if (dev == NULL)
		goto out_unlock;

	err = -EPERM;
	if (!ve_is_dev_movable(dev))
		goto out_unlock;

	err = -EINVAL;
	if (dev->flags & (IFF_SLAVE|IFF_MASTER))
		goto out_unlock;

	ve_netdev_cleanup(dev, 1);

	write_lock_bh(&dev_base_lock);
	__ve_dev_move(dev, get_ve0(), ve, get_exec_ub());
	write_unlock_bh(&dev_base_lock);

	err = 0;

out_unlock:
	rtnl_unlock();
	real_put_ve(ve);

	if (dev == NULL)
		printk(KERN_WARNING "Device %s not found\n", dev_name);

out:
	return err;
}

static int ve_dev_del(envid_t veid, char *dev_name)
{
	int err;
	struct net_device *dev;
	struct ve_struct *ve, *old_exec;
	struct hlist_node *p;

	dev = NULL;
	err = -ESRCH;

	ve = get_ve_by_id(veid);
	if (ve == NULL)
		goto out;

	rtnl_lock();

	read_lock(&dev_base_lock);
	hlist_for_each(p, dev_name_hash(dev_name, ve)) {
		struct net_device *d = hlist_entry(p, struct net_device,
						   name_hlist);
		if (strncmp(d->name, dev_name, IFNAMSIZ) == 0) {
			dev = d;
			break;
		}
	}
	read_unlock(&dev_base_lock);
	if (dev == NULL)
		goto out_unlock;

	err = -EPERM;
	if (!ve_is_dev_movable(dev))
		goto out_unlock;

	old_exec = set_exec_env(ve);
	ve_netdev_cleanup(dev, 0);
	(void)set_exec_env(old_exec);

	write_lock_bh(&dev_base_lock);
	__ve_dev_move(dev, ve, get_ve0(), netdev_bc(dev)->owner_ub);
	write_unlock_bh(&dev_base_lock);

	err = 0;

out_unlock:
	rtnl_unlock();
	real_put_ve(ve);

	if (dev == NULL)
		printk(KERN_WARNING "Device %s not found\n", dev_name);

out:
	return err;
}

int real_ve_dev_map(envid_t veid, int op, char *dev_name)
{
	int err;
	err = -EPERM;
	if (!capable(CAP_SETVEID))
		goto out;
	switch (op)
	{
		case VE_NETDEV_ADD:
			err = ve_dev_add(veid, dev_name);
			break;
		case VE_NETDEV_DEL:
			err = ve_dev_del(veid, dev_name);
			break;
		default:
			err = -EINVAL;
			break;
	}
out:
	return err;
}

static void ve_mapped_devs_cleanup(struct ve_struct *ve)
{
	struct net_device *dev;

	rtnl_lock();
	write_lock_bh(&dev_base_lock);
restart:
	for (dev = ve->_net_dev_base; dev != NULL; dev = dev->next)
	{
		if ((dev->features & NETIF_F_VENET) ||
		    (dev == ve->_loopback_dev)) /* Skip loopback dev */
			continue;
		write_unlock_bh(&dev_base_lock);
		ve_netdev_cleanup(dev, 0);
		write_lock_bh(&dev_base_lock);
		__ve_dev_move(dev, ve, get_ve0(), netdev_bc(dev)->owner_ub);
		goto restart;
	}
	write_unlock_bh(&dev_base_lock);
	rtnl_unlock();
}
#endif


/**********************************************************************
 **********************************************************************
 *
 * VE information via /proc
 *
 **********************************************************************
 **********************************************************************/
#ifdef CONFIG_PROC_FS
static int devperms_seq_show(struct seq_file *m, void *v)
{
	struct devperms_struct *dp;
	char dev_s[32], type_c;
	unsigned use, type;
	dev_t dev;

	dp = (struct devperms_struct *)v;
	if (dp == (struct devperms_struct *)1L) {
		seq_printf(m, "Version: 2.7\n");
		return 0;
	}

	use = dp->type & VE_USE_MASK;
	type = dp->type & S_IFMT;
	dev = dp->dev;

	if ((use | VE_USE_MINOR) == use)
		snprintf(dev_s, sizeof(dev_s), "%d:%d", MAJOR(dev), MINOR(dev));
	else if ((use | VE_USE_MAJOR) == use)
		snprintf(dev_s, sizeof(dev_s), "%d:*", MAJOR(dp->dev));
	else
		snprintf(dev_s, sizeof(dev_s), "*:*");

	if (type == S_IFCHR)
		type_c = 'c';
	else if (type == S_IFBLK)
		type_c = 'b';
	else
		type_c = '?';

	seq_printf(m, "%10u %c %03o %s\n", dp->veid, type_c, dp->mask, dev_s);
	return 0;
}

static void *devperms_seq_start(struct seq_file *m, loff_t *pos)
{
	loff_t cpos;
	long slot;
	struct devperms_struct *dp;

	cpos = *pos;
	read_lock(&devperms_hash_guard);
	if (cpos-- == 0)
		return (void *)1L;

	for (slot = 0; slot < DEVPERMS_HASH_SZ; slot++)
		for (dp = devperms_hash[slot]; dp; dp = dp->devhash_next)
			if (cpos-- == 0) {
				m->private = (void *)slot;
				return dp;
			}
	return NULL;
}

static void *devperms_seq_next(struct seq_file *m, void *v, loff_t *pos)
{
	long slot;
	struct devperms_struct *dp;

	dp = (struct devperms_struct *)v;

	if (dp == (struct devperms_struct *)1L)
		slot = 0;
	else if (dp->devhash_next == NULL)
		slot = (long)m->private + 1;
	else {
		(*pos)++;
		return dp->devhash_next;
	}

	for (; slot < DEVPERMS_HASH_SZ; slot++)
		if (devperms_hash[slot]) {
			(*pos)++;
			m->private = (void *)slot;
			return devperms_hash[slot];
		}
	return NULL;
}

static void devperms_seq_stop(struct seq_file *m, void *v)
{
	read_unlock(&devperms_hash_guard);
}

static struct seq_operations devperms_seq_op = {
	.start	= devperms_seq_start,
	.next	= devperms_seq_next,
	.stop	= devperms_seq_stop,
	.show	= devperms_seq_show,
};

static int devperms_open(struct inode *inode, struct file *file)
{
        return seq_open(file, &devperms_seq_op);
}

static struct file_operations proc_devperms_ops = {
	.open		= devperms_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release,
};

#if BITS_PER_LONG == 32
#define VESTAT_LINE_WIDTH (6 * 11 + 6 * 21)
#define VESTAT_LINE_FMT "%10u %10lu %10lu %10lu %10lu %20Lu %20Lu %20Lu %20Lu %20Lu %20Lu %10lu\n"
#define VESTAT_HEAD_FMT "%10s %10s %10s %10s %10s %20s %20s %20s %20s %20s %20s %10s\n"
#else
#define VESTAT_LINE_WIDTH (12 * 21)
#define VESTAT_LINE_FMT "%20u %20lu %20lu %20lu %20lu %20Lu %20Lu %20Lu %20Lu %20Lu %20Lu %20lu\n"
#define VESTAT_HEAD_FMT "%20s %20s %20s %20s %20s %20s %20s %20s %20s %20s %20s %20s\n"
#endif

static int vestat_seq_show(struct seq_file *m, void *v)
{
	struct ve_struct *ve = (struct ve_struct *)v;
	struct ve_struct *curve;
	int cpu;
	unsigned long user_ve, nice_ve, system_ve, uptime;
	cycles_t uptime_cycles, idle_time, strv_time, used;

	curve = get_exec_env();
	if (ve == ve_list_head ||
	    (!ve_is_super(curve) && ve == curve)) {
		/* print header */
		seq_printf(m, "%-*s\n",
			VESTAT_LINE_WIDTH - 1,
			"Version: 2.2");
		seq_printf(m, VESTAT_HEAD_FMT, "VEID",
					"user", "nice", "system",
					"uptime", "idle",
					"strv", "uptime", "used",
					"maxlat", "totlat", "numsched");
	}

	if (ve == get_ve0())
		return 0;

	user_ve = nice_ve = system_ve = 0;
	idle_time = strv_time = used = 0;

	for (cpu = 0; cpu < NR_CPUS; cpu++) {
		struct ve_cpu_stats *st;

		st = VE_CPU_STATS(ve, cpu);
		user_ve += st->user;
		nice_ve += st->nice;
		system_ve += st->system;
		used += VE_CPU_STATS(ve, cpu)->used_time;
		idle_time += ve_sched_get_idle_time(ve, cpu);
	}
	uptime_cycles = get_cycles() - ve->start_cycles;
	uptime = jiffies - ve->start_jiffies;

	seq_printf(m, VESTAT_LINE_FMT, ve->veid,
				user_ve, nice_ve, system_ve,
				uptime, idle_time,
				strv_time, uptime_cycles, used,
				ve->sched_lat_ve.last.maxlat,
				ve->sched_lat_ve.last.totlat,
				ve->sched_lat_ve.last.count);
	return 0;
}

static void *ve_seq_start(struct seq_file *m, loff_t *pos)
{
	struct ve_struct *ve, *curve;
	loff_t l;

	curve = get_exec_env();
	read_lock(&ve_list_guard);
	if (!ve_is_super(curve)) {
		if (*pos != 0)
			return NULL;
		return curve;
	}
	for (ve = ve_list_head, l = *pos;
	     ve != NULL && l > 0;
	     ve = ve->next, l--);
	return ve;
}

static void *ve_seq_next(struct seq_file *m, void *v, loff_t *pos)
{
	struct ve_struct *ve = (struct ve_struct *)v;

	if (!ve_is_super(get_exec_env()))
		return NULL;
	(*pos)++;
	return ve->next;
}

static void ve_seq_stop(struct seq_file *m, void *v)
{
	read_unlock(&ve_list_guard);
}

static struct seq_operations vestat_seq_op = {
        start:  ve_seq_start,
        next:   ve_seq_next,
        stop:   ve_seq_stop,
        show:   vestat_seq_show
};

static int vestat_open(struct inode *inode, struct file *file)
{
        return seq_open(file, &vestat_seq_op);
}

static struct file_operations proc_vestat_operations = {
        open:           vestat_open,
        read:           seq_read,
        llseek:         seq_lseek,
        release:        seq_release
};

static int __init init_vecalls_proc(void)
{
	struct proc_dir_entry *de;

	de = create_proc_glob_entry("vz/vestat",
			S_IFREG|S_IRUSR, NULL);
	if (de == NULL) {
		/* create "vz" subdirectory, if not exist */
		(void) create_proc_glob_entry("vz",
					      S_IFDIR|S_IRUGO|S_IXUGO, NULL);
		de = create_proc_glob_entry("vz/vestat",
				S_IFREG|S_IRUSR, NULL);
	}
	if (de)
		de->proc_fops = &proc_vestat_operations;
	else
		printk(KERN_WARNING
				"VZMON: can't make vestat proc entry\n");

	de = create_proc_entry("vz/devperms", S_IFREG | S_IRUSR, NULL);
	if (de)
		de->proc_fops = &proc_devperms_ops;
	else
		printk(KERN_WARNING
				"VZMON: can't make devperms proc entry\n");
	return 0;
}

static void fini_vecalls_proc(void)
{
	remove_proc_entry("vz/devperms", NULL);
	remove_proc_entry("vz/vestat", NULL);
}
#else
#define init_vecalls_proc()	(0)
#define fini_vecalls_proc()	do { } while (0)
#endif /* CONFIG_PROC_FS */


/**********************************************************************
 **********************************************************************
 *
 * User ctl
 *
 **********************************************************************
 **********************************************************************/

int vzcalls_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static struct vzioctlinfo vzcalls = {
	type: VZCTLTYPE,
	func: vzcalls_ioctl,
	owner: THIS_MODULE,
};

int vzcalls_ioctl(struct inode *ino, struct file *file, unsigned int cmd,
		unsigned long arg)
{
	int err;

	err = -ENOTTY;
	switch(cmd) {
	    case VZCTL_MARK_ENV_TO_DOWN: {
		        /* Compatibility issue */
		        err = 0;
		}
		break;
	    case VZCTL_SETDEVPERMS: {
			/* Device type was mistakenly declared as dev_t
			 * in the old user-kernel interface.
			 * That's wrong, dev_t is a kernel internal type.
			 * I use `unsigned' not having anything better in mind.
			 * 2001/08/11  SAW  */
			struct vzctl_setdevperms s;
			err = -EFAULT;
			if (copy_from_user(&s, (void *)arg, sizeof(s)))
				break;
			err = real_setdevperms(s.veid, s.type,
					new_decode_dev(s.dev), s.mask);
		}
		break;
#if defined(CONFIG_VE_NETDEV) || defined(CONFIG_VE_NETDEV_MODULE)
	    case VZCTL_VE_NETDEV: {
			struct vzctl_ve_netdev d;
			char *s;
			err = -EFAULT;
			if (copy_from_user(&d, (void *)arg, sizeof(d)))
				break;
			err = -ENOMEM;
			s = kmalloc(IFNAMSIZ+1, GFP_KERNEL);
			if (s == NULL)
				break;
			err = -EFAULT;
			if (strncpy_from_user(s, d.dev_name, IFNAMSIZ) > 0) {
				s[IFNAMSIZ] = 0;
				err = real_ve_dev_map(d.veid, d.op, s);
			}
			kfree(s);
		}
		break;
#endif
	    case VZCTL_ENV_CREATE: {
			struct vzctl_env_create s;
			err = -EFAULT;
			if (copy_from_user(&s, (void *)arg, sizeof(s)))
				break;
			err = real_env_create(s.veid, s.flags, s.class_id,
				NULL, 0);
		}
		break;
	    case VZCTL_ENV_CREATE_DATA: {
			struct vzctl_env_create_data s;
			env_create_param_t *data;
			err = -EFAULT;
			if (copy_from_user(&s, (void *)arg, sizeof(s)))
				break;
			err=-EINVAL;
			if (s.datalen < VZCTL_ENV_CREATE_DATA_MINLEN ||
			    s.datalen > VZCTL_ENV_CREATE_DATA_MAXLEN ||
			    s.data == 0)
				break;
			err = -ENOMEM;
			data = kmalloc(sizeof(*data), GFP_KERNEL);
			if (!data)
				break;
			memset(data, 0, sizeof(*data));
			err = -EFAULT;
			if (copy_from_user(data, (void *)s.data, s.datalen))
				goto free_data;
			err = real_env_create(s.veid, s.flags, s.class_id,
				data, s.datalen);
free_data:
			kfree(data);
		}
		break;
	    case VZCTL_GET_CPU_STAT: {
			struct vzctl_cpustatctl s;
			err = -EFAULT;
			if (copy_from_user(&s, (void *)arg, sizeof(s)))
				break;
			err = ve_get_cpu_stat(s.veid, s.cpustat);
		}
		break;
	}
	return err;
}


/**********************************************************************
 **********************************************************************
 *
 * Init/exit stuff
 *
 **********************************************************************
 **********************************************************************/

#ifdef CONFIG_VE_CALLS_MODULE
static int __init init_vecalls_symbols(void)
{
	KSYMRESOLVE(real_get_device_perms_ve);
	KSYMRESOLVE(real_do_env_cleanup);
	KSYMRESOLVE(real_do_env_free);
	KSYMRESOLVE(real_update_load_avg_ve);
	KSYMMODRESOLVE(vzmon);
	return 0;
}

static void fini_vecalls_symbols(void)
{
	KSYMMODUNRESOLVE(vzmon);
	KSYMUNRESOLVE(real_get_device_perms_ve);
	KSYMUNRESOLVE(real_do_env_cleanup);
	KSYMUNRESOLVE(real_do_env_free);
	KSYMUNRESOLVE(real_update_load_avg_ve);
}
#else
#define init_vecalls_symbols()	(0)
#define fini_vecalls_symbols()	do { } while (0)
#endif

static inline __init int init_vecalls_ioctls(void)
{
	vzioctl_register(&vzcalls);
	return 0;
}

static inline void fini_vecalls_ioctls(void)
{
	vzioctl_unregister(&vzcalls);
}

static int __init vecalls_init(void)
{
	int err;
	int i;

	ve_list_head = get_ve0();

	err = init_vzmond();
	if (err < 0)
		goto out_vzmond;

	err = init_devperms_hash();
	if (err < 0)
		goto out_perms;

	err = init_vecalls_symbols();
	if (err < 0)
		goto out_sym;

	err = init_vecalls_proc();
	if (err < 0)
		goto out_proc;

	err = init_vecalls_ioctls();
	if (err < 0)
		goto out_ioctls;

	for (i = 0; i < VE_MAX_HOOKS; i++)
		INIT_LIST_HEAD(&ve_hooks[i]);

	return 0;

out_ioctls:
	fini_vecalls_proc();
out_proc:
	fini_vecalls_symbols();
out_sym:
	fini_devperms_hash();
out_perms:
	fini_vzmond();
out_vzmond:
	return err;
}

static void vecalls_exit(void)
{
	fini_vecalls_ioctls();
	fini_vecalls_proc();
	fini_vecalls_symbols();
	fini_devperms_hash();
	fini_vzmond();
}

EXPORT_SYMBOL(get_ve_by_id);
EXPORT_SYMBOL(__find_ve_by_id);
EXPORT_SYMBOL(ve_list_guard);
EXPORT_SYMBOL(ve_list_head);
EXPORT_SYMBOL(nr_ve);

MODULE_AUTHOR("SWsoft <info@sw-soft.com>");
MODULE_DESCRIPTION("Virtuozzo Control");
MODULE_LICENSE("GPL v2");

module_init(vecalls_init)
module_exit(vecalls_exit)
