/*
 *  kernel/vzdev.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/fs.h>
#include <linux/list.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/vzctl.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/vzcalluser.h>
#include <asm/uaccess.h>
#include <asm/pgalloc.h>
#include <linux/device.h>

#define VZCTL_MAJOR 126
#define VZCTL_NAME "vzctl"

MODULE_AUTHOR("SWsoft <info@sw-soft.com>");
MODULE_DESCRIPTION("Virtuozzo Interface");
MODULE_LICENSE("GPL v2");

static LIST_HEAD(ioctls);
static spinlock_t ioctl_lock = SPIN_LOCK_UNLOCKED;

int vzctl_ioctl(struct inode *ino, struct file *file, unsigned int cmd,
		unsigned long arg)
{
	int err;
	struct list_head *p;
	struct vzioctlinfo *inf;

	err = -ENOTTY;
	spin_lock(&ioctl_lock);
	list_for_each(p, &ioctls) {
		inf = list_entry(p, struct vzioctlinfo, list);
		if (inf->type != _IOC_TYPE(cmd))
			continue;

		err = try_module_get(inf->owner) ? 0 : -EBUSY;
		spin_unlock(&ioctl_lock);
		if (!err) {
			err = (*inf->func)(ino, file, cmd, arg);
			module_put(inf->owner);
		}
		return err;
	}
	spin_unlock(&ioctl_lock);
	return err;
}

void vzioctl_register(struct vzioctlinfo *inf)
{
	spin_lock(&ioctl_lock);
	list_add(&inf->list, &ioctls);
	spin_unlock(&ioctl_lock);
}

void vzioctl_unregister(struct vzioctlinfo *inf)
{
	spin_lock(&ioctl_lock);
	list_del_init(&inf->list);
	spin_unlock(&ioctl_lock);
}

EXPORT_SYMBOL(vzioctl_register);
EXPORT_SYMBOL(vzioctl_unregister);

/*
 * Init/exit stuff.
 */
static struct file_operations vzctl_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= vzctl_ioctl,
};

static struct class *vzctl_class;

static void __exit vzctl_exit(void)
{
	class_device_destroy(vzctl_class, MKDEV(VZCTL_MAJOR, 0));
	class_destroy(vzctl_class);
	unregister_chrdev(VZCTL_MAJOR, VZCTL_NAME);
}

static int __init vzctl_init(void)
{
	int ret;
	struct class_device *class_err;

	ret = register_chrdev(VZCTL_MAJOR, VZCTL_NAME, &vzctl_fops);
	if (ret < 0)
		goto out;

	vzctl_class = class_create(THIS_MODULE, "vzctl");
	if (IS_ERR(vzctl_class)) {
		ret = PTR_ERR(vzctl_class);
		goto out_cleandev;
	}

	class_err = class_device_create(vzctl_class, NULL, MKDEV(VZCTL_MAJOR, 0),
				NULL, VZCTL_NAME);
	if (IS_ERR(class_err)) {
		ret = PTR_ERR(class_err);
		goto out_rmclass;
	}

	goto out;

out_rmclass:
	class_destroy(vzctl_class);
out_cleandev:
	unregister_chrdev(VZCTL_MAJOR, VZCTL_NAME);
out:
	return ret;
}

module_init(vzctl_init)
module_exit(vzctl_exit);
