/* IP tables module for matching the value of the TTL 
 *
 * ipt_ttl.c,v 1.5 2000/11/13 11:16:08 laforge Exp
 *
 * (C) 2000,2001 by Harald Welte <laforge@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/nfcalls.h>

#include <linux/netfilter_ipv4/ipt_ttl.h>
#include <linux/netfilter_ipv4/ip_tables.h>

MODULE_AUTHOR("Harald Welte <laforge@netfilter.org>");
MODULE_DESCRIPTION("IP tables TTL matching module");
MODULE_LICENSE("GPL");

static int match(const struct sk_buff *skb, const struct net_device *in,
		 const struct net_device *out, const void *matchinfo,
		 int offset, unsigned int protoff, int *hotdrop)
{
	const struct ipt_ttl_info *info = matchinfo;

	switch (info->mode) {
		case IPT_TTL_EQ:
			return (skb->nh.iph->ttl == info->ttl);
			break;
		case IPT_TTL_NE:
			return (!(skb->nh.iph->ttl == info->ttl));
			break;
		case IPT_TTL_LT:
			return (skb->nh.iph->ttl < info->ttl);
			break;
		case IPT_TTL_GT:
			return (skb->nh.iph->ttl > info->ttl);
			break;
		default:
			printk(KERN_WARNING "ipt_ttl: unknown mode %d\n", 
				info->mode);
			return 0;
	}

	return 0;
}

static int checkentry(const char *tablename, const void  *ip,
		      void *matchinfo, unsigned int matchsize,
		      unsigned int hook_mask)
{
	if (matchsize != IPT_ALIGN(sizeof(struct ipt_ttl_info)))
		return 0;

	return 1;
}

#ifdef CONFIG_COMPAT
static int compat(void *match,
		void **dstptr, int *size, int convert)
{
	int off;

	off = IPT_ALIGN(sizeof(struct ipt_ttl_info)) -
		COMPAT_IPT_ALIGN(sizeof(struct ipt_ttl_info));
	return ipt_match_align_compat(match, dstptr, size, off, convert);
}
#endif

static struct ipt_match ttl_match = {
	.name		= "ttl",
	.match		= &match,
	.checkentry	= &checkentry,
#ifdef CONFIG_COMPAT
	.compat		= &compat,
#endif
	.me		= THIS_MODULE,
};

int init_iptable_ttl(void)
{
	return ipt_register_match(&ttl_match);
}

void fini_iptable_ttl(void)
{
	ipt_unregister_match(&ttl_match);
}

static int __init init(void)
{
	int err;

	err = init_iptable_ttl();
	if (err < 0)
		return err;

	KSYMRESOLVE(init_iptable_ttl);
	KSYMRESOLVE(fini_iptable_ttl);
	KSYMMODRESOLVE(ipt_ttl);
	return 0;
}

static void __exit fini(void)
{
	KSYMMODUNRESOLVE(ipt_ttl);
	KSYMUNRESOLVE(init_iptable_ttl);
	KSYMUNRESOLVE(fini_iptable_ttl);
	fini_iptable_ttl();
}

module_init(init);
module_exit(fini);
