/* Kernel module to match connection tracking information.
 * Superset of Rusty's minimalistic state match.
 *
 * (C) 2001  Marc Boucher (marc@mbsi.ca).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>

#if defined(CONFIG_IP_NF_CONNTRACK) || defined(CONFIG_IP_NF_CONNTRACK_MODULE)
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_tuple.h>
#else
#include <net/netfilter/nf_conntrack.h>
#endif

#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_conntrack.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/nfcalls.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Marc Boucher <marc@mbsi.ca>");
MODULE_DESCRIPTION("iptables connection tracking match module");
MODULE_ALIAS("ipt_conntrack");

#if defined(CONFIG_IP_NF_CONNTRACK) || defined(CONFIG_IP_NF_CONNTRACK_MODULE)

static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      unsigned int protoff,
      int *hotdrop)
{
	const struct xt_conntrack_info *sinfo = matchinfo;
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;
	unsigned int statebit;

	ct = ip_conntrack_get((struct sk_buff *)skb, &ctinfo);

#define FWINV(bool,invflg) ((bool) ^ !!(sinfo->invflags & invflg))

	if (ct == &ip_conntrack_untracked)
		statebit = XT_CONNTRACK_STATE_UNTRACKED;
	else if (ct)
 		statebit = XT_CONNTRACK_STATE_BIT(ctinfo);
 	else
 		statebit = XT_CONNTRACK_STATE_INVALID;
 
	if(sinfo->flags & XT_CONNTRACK_STATE) {
		if (ct) {
			if(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip !=
			    ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip)
				statebit |= XT_CONNTRACK_STATE_SNAT;

			if(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip !=
			    ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip)
				statebit |= XT_CONNTRACK_STATE_DNAT;
		}

		if (FWINV((statebit & sinfo->statemask) == 0, XT_CONNTRACK_STATE))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_PROTO) {
		if (!ct || FWINV(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum != sinfo->tuple[IP_CT_DIR_ORIGINAL].dst.protonum, XT_CONNTRACK_PROTO))
                	return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_ORIGSRC) {
		if (!ct || FWINV((ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip&sinfo->sipmsk[IP_CT_DIR_ORIGINAL].s_addr) != sinfo->tuple[IP_CT_DIR_ORIGINAL].src.ip, XT_CONNTRACK_ORIGSRC))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_ORIGDST) {
		if (!ct || FWINV((ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip&sinfo->dipmsk[IP_CT_DIR_ORIGINAL].s_addr) != sinfo->tuple[IP_CT_DIR_ORIGINAL].dst.ip, XT_CONNTRACK_ORIGDST))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_REPLSRC) {
		if (!ct || FWINV((ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip&sinfo->sipmsk[IP_CT_DIR_REPLY].s_addr) != sinfo->tuple[IP_CT_DIR_REPLY].src.ip, XT_CONNTRACK_REPLSRC))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_REPLDST) {
		if (!ct || FWINV((ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip&sinfo->dipmsk[IP_CT_DIR_REPLY].s_addr) != sinfo->tuple[IP_CT_DIR_REPLY].dst.ip, XT_CONNTRACK_REPLDST))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_STATUS) {
		if (!ct || FWINV((ct->status & sinfo->statusmask) == 0, XT_CONNTRACK_STATUS))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_EXPIRES) {
		unsigned long expires;

		if(!ct)
			return 0;

		expires = timer_pending(&ct->timeout) ? (ct->timeout.expires - jiffies)/HZ : 0;

		if (FWINV(!(expires >= sinfo->expires_min && expires <= sinfo->expires_max), XT_CONNTRACK_EXPIRES))
			return 0;
	}

	return 1;
}

#else /* CONFIG_IP_NF_CONNTRACK */
static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      unsigned int protoff,
      int *hotdrop)
{
	const struct xt_conntrack_info *sinfo = matchinfo;
	struct nf_conn *ct;
	enum ip_conntrack_info ctinfo;
	unsigned int statebit;

	ct = nf_ct_get((struct sk_buff *)skb, &ctinfo);

#define FWINV(bool,invflg) ((bool) ^ !!(sinfo->invflags & invflg))

	if (ct == &nf_conntrack_untracked)
		statebit = XT_CONNTRACK_STATE_UNTRACKED;
	else if (ct)
 		statebit = XT_CONNTRACK_STATE_BIT(ctinfo);
 	else
 		statebit = XT_CONNTRACK_STATE_INVALID;
 
	if(sinfo->flags & XT_CONNTRACK_STATE) {
		if (ct) {
			if(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u3.ip !=
			    ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u3.ip)
				statebit |= XT_CONNTRACK_STATE_SNAT;

			if(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u3.ip !=
			    ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u3.ip)
				statebit |= XT_CONNTRACK_STATE_DNAT;
		}

		if (FWINV((statebit & sinfo->statemask) == 0, XT_CONNTRACK_STATE))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_PROTO) {
		if (!ct || FWINV(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum != sinfo->tuple[IP_CT_DIR_ORIGINAL].dst.protonum, XT_CONNTRACK_PROTO))
                	return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_ORIGSRC) {
		if (!ct || FWINV((ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u3.ip&sinfo->sipmsk[IP_CT_DIR_ORIGINAL].s_addr) != sinfo->tuple[IP_CT_DIR_ORIGINAL].src.ip, XT_CONNTRACK_ORIGSRC))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_ORIGDST) {
		if (!ct || FWINV((ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u3.ip&sinfo->dipmsk[IP_CT_DIR_ORIGINAL].s_addr) != sinfo->tuple[IP_CT_DIR_ORIGINAL].dst.ip, XT_CONNTRACK_ORIGDST))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_REPLSRC) {
		if (!ct || FWINV((ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u3.ip&sinfo->sipmsk[IP_CT_DIR_REPLY].s_addr) != sinfo->tuple[IP_CT_DIR_REPLY].src.ip, XT_CONNTRACK_REPLSRC))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_REPLDST) {
		if (!ct || FWINV((ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u3.ip&sinfo->dipmsk[IP_CT_DIR_REPLY].s_addr) != sinfo->tuple[IP_CT_DIR_REPLY].dst.ip, XT_CONNTRACK_REPLDST))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_STATUS) {
		if (!ct || FWINV((ct->status & sinfo->statusmask) == 0, XT_CONNTRACK_STATUS))
			return 0;
	}

	if(sinfo->flags & XT_CONNTRACK_EXPIRES) {
		unsigned long expires;

		if(!ct)
			return 0;

		expires = timer_pending(&ct->timeout) ? (ct->timeout.expires - jiffies)/HZ : 0;

		if (FWINV(!(expires >= sinfo->expires_min && expires <= sinfo->expires_max), XT_CONNTRACK_EXPIRES))
			return 0;
	}

	return 1;
}

#endif /* CONFIG_NF_IP_CONNTRACK */

static int check(const char *tablename,
		 const void *ip,
		 void *matchinfo,
		 unsigned int matchsize,
		 unsigned int hook_mask)
{
	if (matchsize != XT_ALIGN(sizeof(struct xt_conntrack_info)))
		return 0;

	return 1;
}

#ifdef CONFIG_COMPAT
static int compat_to_user(void *match, void **dstptr,
		int *size, int off)
{
	struct ipt_entry_match *pm;
	struct xt_conntrack_info *pinfo;
	struct compat_xt_conntrack_info info;
	u_int16_t msize;

	pm = (struct ipt_entry_match *)match;
	msize = pm->u.user.match_size;
	if (__copy_to_user(*dstptr, pm, sizeof(struct ipt_entry_match)))
		return -EFAULT;
	pinfo = (struct xt_conntrack_info *)pm->data;
	memset(&info, 0, sizeof(struct compat_xt_conntrack_info));
	info.statemask = pinfo->statemask;
	info.statusmask = pinfo->statusmask;
	memcpy(info.tuple, pinfo->tuple, IP_CT_DIR_MAX *
			sizeof(struct ip_conntrack_tuple));
	memcpy(info.sipmsk, pinfo->sipmsk,
			IP_CT_DIR_MAX * sizeof(struct in_addr));
	memcpy(info.dipmsk, pinfo->dipmsk,
			IP_CT_DIR_MAX * sizeof(struct in_addr));
	info.expires_min = pinfo->expires_min;
	info.expires_max = pinfo->expires_max;
	info.flags = pinfo->flags;
	info.invflags = pinfo->invflags;
	if (__copy_to_user(*dstptr + sizeof(struct ipt_entry_match),
				&info, sizeof(struct compat_xt_conntrack_info)))
		return -EFAULT;
	msize -= off;
	if (put_user(msize, (u_int16_t *)*dstptr))
		return -EFAULT;
	*size -= off;
	*dstptr += msize;
	return 0;
}

static int compat_from_user(void *match, void **dstptr,
		int *size, int off)
{
	struct compat_ipt_entry_match *pm;
	struct ipt_entry_match *dstpm;
	struct compat_xt_conntrack_info *pinfo;
	struct xt_conntrack_info info;
	u_int16_t msize;

	pm = (struct compat_ipt_entry_match *)match;
	dstpm = (struct ipt_entry_match *)*dstptr;
	msize = pm->u.user.match_size;
	memcpy(*dstptr, pm, sizeof(struct compat_ipt_entry_match));
	pinfo = (struct compat_xt_conntrack_info *)pm->data;
	memset(&info, 0, sizeof(struct xt_conntrack_info));
	info.statemask = pinfo->statemask;
	info.statusmask = pinfo->statusmask;
	memcpy(info.tuple, pinfo->tuple, IP_CT_DIR_MAX *
			sizeof(struct ip_conntrack_tuple));
	memcpy(info.sipmsk, pinfo->sipmsk,
			IP_CT_DIR_MAX * sizeof(struct in_addr));
	memcpy(info.dipmsk, pinfo->dipmsk,
			IP_CT_DIR_MAX * sizeof(struct in_addr));
	info.expires_min = pinfo->expires_min;
	info.expires_max = pinfo->expires_max;
	info.flags = pinfo->flags;
	info.invflags = pinfo->invflags;
	memcpy(*dstptr + sizeof(struct compat_ipt_entry_match),
				&info, sizeof(struct xt_conntrack_info));
	msize += off;
	dstpm->u.user.match_size = msize;
	*size += off;
	*dstptr += msize;
	return 0;
}

static int compat(void *match, void **dstptr, int *size, int convert)
{
	int ret, off;

	off = XT_ALIGN(sizeof(struct xt_conntrack_info)) -
		COMPAT_XT_ALIGN(sizeof(struct compat_xt_conntrack_info));
	switch (convert) {
		case COMPAT_TO_USER:
			ret = compat_to_user(match, dstptr, size, off);
			break;
		case COMPAT_FROM_USER:
			ret = compat_from_user(match, dstptr, size, off);
			break;
		case COMPAT_CALC_SIZE:
			*size += off;
			ret = 0;
			break;
		default:
			ret = -ENOPROTOOPT;
			break;
	}
	return ret;
}
#endif

static struct xt_match conntrack_match = {
	.name		= "conntrack",
	.match		= &match,
	.checkentry	= &check,
#ifdef CONFIG_COMPAT
	.compat		= &compat,
#endif
	.me		= THIS_MODULE,
};

int init_xt_conntrack_match(void)
{
	return xt_register_match(AF_INET, &conntrack_match);
}

void fini_xt_conntrack_match(void)
{
	xt_unregister_match(AF_INET, &conntrack_match);
}

static int __init init(void)
{
	int ret;
	need_conntrack();
	ret = init_xt_conntrack_match();
	if (ret < 0)
		return ret;

	KSYMRESOLVE(init_xt_conntrack_match);
	KSYMRESOLVE(fini_xt_conntrack_match);
	KSYMMODRESOLVE(xt_conntrack);
	return ret;
}

static void __exit fini(void)
{
	KSYMMODUNRESOLVE(xt_conntrack);
	KSYMUNRESOLVE(init_xt_conntrack_match);
	KSYMUNRESOLVE(fini_xt_conntrack_match);
	fini_xt_conntrack_match();
}

module_init(init);
module_exit(fini);
