/* Kernel module to match packet length. */
/* (C) 1999-2001 James Morris <jmorros@intercode.com.au>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ipv6.h>
#include <net/ip.h>

#include <linux/netfilter/xt_length.h>
#include <linux/netfilter/x_tables.h>
#include <linux/nfcalls.h>

MODULE_AUTHOR("James Morris <jmorris@intercode.com.au>");
MODULE_DESCRIPTION("IP tables packet length matching module");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_length");
MODULE_ALIAS("ip6t_length");

static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      unsigned int protoff,
      int *hotdrop)
{
	const struct xt_length_info *info = matchinfo;
	u_int16_t pktlen = ntohs(skb->nh.iph->tot_len);
	
	return (pktlen >= info->min && pktlen <= info->max) ^ info->invert;
}

static int
match6(const struct sk_buff *skb,
       const struct net_device *in,
       const struct net_device *out,
       const void *matchinfo,
       int offset,
       unsigned int protoff,
       int *hotdrop)
{
	const struct xt_length_info *info = matchinfo;
	u_int16_t pktlen = ntohs(skb->nh.ipv6h->payload_len) + sizeof(struct ipv6hdr);
	
	return (pktlen >= info->min && pktlen <= info->max) ^ info->invert;
}

static int
checkentry(const char *tablename,
           const void *ip,
           void *matchinfo,
           unsigned int matchsize,
           unsigned int hook_mask)
{
	if (matchsize != XT_ALIGN(sizeof(struct xt_length_info)))
		return 0;

	return 1;
}

#ifdef CONFIG_COMPAT
static int compat(void *match,
		void **dstptr, int *size, int convert)
{
	int off;

	off = XT_ALIGN(sizeof(struct xt_length_info)) -
		COMPAT_XT_ALIGN(sizeof(struct xt_length_info));
	return ipt_match_align_compat(match, dstptr, size, off, convert);
}
#endif

static struct xt_match length_match = {
	.name		= "length",
	.match		= &match,
	.checkentry	= &checkentry,
#ifdef CONFIG_COMPAT
	.compat		= &compat,
#endif
	.me		= THIS_MODULE,
};
static struct xt_match length6_match = {
	.name		= "length",
	.match		= &match6,
	.checkentry	= &checkentry,
#ifdef CONFIG_COMPAT
	.compat		= &compat,
#endif
	.me		= THIS_MODULE,
};

int init_xt_length(void)
{
	int ret;
	ret = xt_register_match(AF_INET, &length_match);
	if (ret)
		return ret;
	ret = xt_register_match(AF_INET6, &length6_match);
	if (ret)
		xt_unregister_match(AF_INET, &length_match);

	return ret;
}

void fini_xt_length(void)
{
	xt_unregister_match(AF_INET, &length_match);
	xt_unregister_match(AF_INET6, &length6_match);
}

static int __init init(void)
{
	int err;

	err = init_xt_length();
	if (err < 0)
		return err;

	KSYMRESOLVE(init_xt_length);
	KSYMRESOLVE(fini_xt_length);
	KSYMMODRESOLVE(xt_length);
	return 0;
}

static void __exit fini(void)
{
	KSYMMODUNRESOLVE(xt_length);
	KSYMUNRESOLVE(init_xt_length);
	KSYMUNRESOLVE(fini_xt_length);
	fini_xt_length();
}

module_init(init);
module_exit(fini);
