/* Kernel module to control the rate
 *
 * 2 September 1999: Changed from the target RATE to the match
 *                   `limit', removed logging.  Did I mention that
 *                   Alexey is a fucking genius?
 *                   Rusty Russell (rusty@rustcorp.com.au).  */

/* (C) 1999 Jrme de Vivie <devivie@info.enserb.u-bordeaux.fr>
 * (C) 1999 Herv Eychenne <eychenne@info.enserb.u-bordeaux.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/nfcalls.h>

#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_limit.h>
#include <linux/netfilter_ipv4/ip_tables.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Herve Eychenne <rv@wallfire.org>");
MODULE_DESCRIPTION("iptables rate limit match");
MODULE_ALIAS("ipt_limit");
MODULE_ALIAS("ip6t_limit");

#ifdef CONFIG_VE_IPTABLES
#include <linux/sched.h>
#define ve_ipt_limit_reg	(*(get_exec_env()->_ipt_limit_reg))
#else
#define ve_ipt_limit_reg	ipt_limit_reg
#endif

/* The algorithm used is the Simple Token Bucket Filter (TBF)
 * see net/sched/sch_tbf.c in the linux source tree
 */

static DEFINE_SPINLOCK(limit_lock);

/* Rusty: This is my (non-mathematically-inclined) understanding of
   this algorithm.  The `average rate' in jiffies becomes your initial
   amount of credit `credit' and the most credit you can ever have
   `credit_cap'.  The `peak rate' becomes the cost of passing the
   test, `cost'.

   `prev' tracks the last packet hit: you gain one credit per jiffy.
   If you get credit balance more than this, the extra credit is
   discarded.  Every time the match passes, you lose `cost' credits;
   if you don't have that many, the test fails.

   See Alexey's formal explanation in net/sched/sch_tbf.c.

   To get the maxmum range, we multiply by this factor (ie. you get N
   credits per jiffy).  We want to allow a rate as low as 1 per day
   (slowest userspace tool allows), which means
   CREDITS_PER_JIFFY*HZ*60*60*24 < 2^32. ie. */
#define MAX_CPJ (0xFFFFFFFF / (HZ*60*60*24))

/* Repeated shift and or gives us all 1s, final shift and add 1 gives
 * us the power of 2 below the theoretical max, so GCC simply does a
 * shift. */
#define _POW2_BELOW2(x) ((x)|((x)>>1))
#define _POW2_BELOW4(x) (_POW2_BELOW2(x)|_POW2_BELOW2((x)>>2))
#define _POW2_BELOW8(x) (_POW2_BELOW4(x)|_POW2_BELOW4((x)>>4))
#define _POW2_BELOW16(x) (_POW2_BELOW8(x)|_POW2_BELOW8((x)>>8))
#define _POW2_BELOW32(x) (_POW2_BELOW16(x)|_POW2_BELOW16((x)>>16))
#define POW2_BELOW32(x) ((_POW2_BELOW32(x)>>1) + 1)

#define CREDITS_PER_JIFFY POW2_BELOW32(MAX_CPJ)

static int
ipt_limit_match(const struct sk_buff *skb,
		const struct net_device *in,
		const struct net_device *out,
		const void *matchinfo,
		int offset,
		unsigned int protoff,
		int *hotdrop)
{
	struct xt_rateinfo *r = ((struct xt_rateinfo *)matchinfo)->master;
	unsigned long now = jiffies;

	spin_lock_bh(&limit_lock);
	r->credit += (now - xchg(&r->prev, now)) * CREDITS_PER_JIFFY;
	if (r->credit > r->credit_cap)
		r->credit = r->credit_cap;

	if (r->credit >= r->cost) {
		/* We're not limited. */
		r->credit -= r->cost;
		spin_unlock_bh(&limit_lock);
		return 1;
	}

       	spin_unlock_bh(&limit_lock);
	return 0;
}

/* Precision saver. */
static u_int32_t
user2credits(u_int32_t user)
{
	/* If multiplying would overflow... */
	if (user > 0xFFFFFFFF / (HZ*CREDITS_PER_JIFFY))
		/* Divide first. */
		return (user / XT_LIMIT_SCALE) * HZ * CREDITS_PER_JIFFY;

	return (user * HZ * CREDITS_PER_JIFFY) / XT_LIMIT_SCALE;
}

static int
ipt_limit_checkentry(const char *tablename,
		     const void *inf,
		     void *matchinfo,
		     unsigned int matchsize,
		     unsigned int hook_mask)
{
	struct xt_rateinfo *r = matchinfo;

	if (matchsize != XT_ALIGN(sizeof(struct xt_rateinfo)))
		return 0;

	/* Check for overflow. */
	if (r->burst == 0
	    || user2credits(r->avg * r->burst) < user2credits(r->avg)) {
		printk("Overflow in xt_limit, try lower: %u/%u\n",
		       r->avg, r->burst);
		return 0;
	}

	/* User avg in seconds * XT_LIMIT_SCALE: convert to jiffies *
	   128. */
	r->prev = jiffies;
	r->credit = user2credits(r->avg * r->burst);	 /* Credits full. */
	r->credit_cap = user2credits(r->avg * r->burst); /* Credits full. */
	r->cost = user2credits(r->avg);

	/* For SMP, we only want to use one set of counters. */
	r->master = r;

	return 1;
}

#ifdef CONFIG_COMPAT
static int ipt_limit_compat_to_user(void *match, void **dstptr,
		int *size, int off)
{
	struct ipt_entry_match *pm;
	struct xt_rateinfo *pinfo;
	struct compat_xt_rateinfo rinfo;
	u_int16_t msize;

	pm = (struct ipt_entry_match *)match;
	msize = pm->u.user.match_size;
	if (__copy_to_user(*dstptr, pm, sizeof(struct ipt_entry_match)))
		return -EFAULT;
	pinfo = (struct xt_rateinfo *)pm->data;
	memset(&rinfo, 0, sizeof(struct compat_xt_rateinfo));
	rinfo.avg = pinfo->avg;
	rinfo.burst = pinfo->burst;
	if (__copy_to_user(*dstptr + sizeof(struct ipt_entry_match),
				&rinfo, sizeof(struct compat_xt_rateinfo)))
		return -EFAULT;
	msize -= off;
	if (put_user(msize, (u_int16_t *)*dstptr))
		return -EFAULT;
	*size -= off;
	*dstptr += msize;
	return 0;
}

static int ipt_limit_compat_from_user(void *match, void **dstptr,
		int *size, int off)
{
	struct compat_ipt_entry_match *pm;
	struct ipt_entry_match *dstpm;
	struct compat_xt_rateinfo *pinfo;
	struct xt_rateinfo rinfo;
	u_int16_t msize;

	pm = (struct compat_ipt_entry_match *)match;
	dstpm = (struct ipt_entry_match *)*dstptr;
	msize = pm->u.user.match_size;
	memcpy(*dstptr, pm, sizeof(struct compat_ipt_entry_match));
	pinfo = (struct compat_xt_rateinfo *)pm->data;
	memset(&rinfo, 0, sizeof(struct xt_rateinfo));
	rinfo.avg = pinfo->avg;
	rinfo.burst = pinfo->burst;
	memcpy(*dstptr + sizeof(struct compat_ipt_entry_match),
				&rinfo, sizeof(struct xt_rateinfo));
	msize += off;
	dstpm->u.user.match_size = msize;
	*size += off;
	*dstptr += msize;
	return 0;
}

static int ipt_limit_compat(void *match, void **dstptr,
		int *size, int convert)
{
	int ret, off;

	off = XT_ALIGN(sizeof(struct xt_rateinfo)) -
		COMPAT_XT_ALIGN(sizeof(struct compat_xt_rateinfo));
	switch (convert) {
		case COMPAT_TO_USER:
			ret = ipt_limit_compat_to_user(match,
					dstptr, size, off);
			break;
		case COMPAT_FROM_USER:
			ret = ipt_limit_compat_from_user(match,
					dstptr, size, off);
			break;
		case COMPAT_CALC_SIZE:
			*size += off;
			ret = 0;
			break;
		default:
			ret = -ENOPROTOOPT;
			break;
	}
	return ret;
}
#endif

static struct xt_match ipt_limit_reg = {
	.name		= "limit",
	.match		= ipt_limit_match,
	.checkentry	= ipt_limit_checkentry,
#ifdef CONFIG_COMPAT
	.compat		= ipt_limit_compat,
#endif
	.me		= THIS_MODULE,
};
static struct xt_match limit6_reg = {
	.name		= "limit",
	.match		= ipt_limit_match,
	.checkentry	= ipt_limit_checkentry,
#ifdef CONFIG_COMPAT
	.compat		= ipt_limit_compat,
#endif
	.me		= THIS_MODULE,
};

int init_xt_limit(void)
{
	int ret;
	
	ret = xt_register_match(AF_INET, &ipt_limit_reg);
	if (ret)
		return ret;
	
	ret = xt_register_match(AF_INET6, &limit6_reg);
	if (ret)
		xt_unregister_match(AF_INET, &ipt_limit_reg);

	return ret;
}

void fini_xt_limit(void)
{
	xt_unregister_match(AF_INET, &ipt_limit_reg);
	xt_unregister_match(AF_INET6, &limit6_reg);
}

static int __init init(void)
{
	int err;

	err = init_xt_limit();
	if (err < 0)
		return err;

	KSYMRESOLVE(init_xt_limit);
	KSYMRESOLVE(fini_xt_limit);
	KSYMMODRESOLVE(xt_limit);
	return 0;
}

static void __exit fini(void)
{
	KSYMMODUNRESOLVE(xt_limit);
	KSYMUNRESOLVE(init_xt_limit);
	KSYMUNRESOLVE(fini_xt_limit);
	fini_xt_limit();
}

module_init(init);
module_exit(fini);
