/* Kernel module to match connection tracking information. */

/* (C) 1999-2001 Paul `Rusty' Russell
 * (C) 2002-2005 Netfilter Core Team <coreteam@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/nfcalls.h>
#include <net/netfilter/nf_conntrack_compat.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_state.h>
#include <linux/netfilter_ipv4/ip_tables.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Rusty Russell <rusty@rustcorp.com.au>");
MODULE_DESCRIPTION("ip[6]_tables connection tracking state match module");
MODULE_ALIAS("ipt_state");
MODULE_ALIAS("ip6t_state");

static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      unsigned int protoff,
      int *hotdrop)
{
	const struct xt_state_info *sinfo = matchinfo;
	enum ip_conntrack_info ctinfo;
	unsigned int statebit;

	if (nf_ct_is_untracked(skb))
		statebit = XT_STATE_UNTRACKED;
	else if (!nf_ct_get_ctinfo(skb, &ctinfo))
		statebit = XT_STATE_INVALID;
	else
		statebit = XT_STATE_BIT(ctinfo);

	return (sinfo->statemask & statebit);
}

static int check(const char *tablename,
		 const void *ip,
		 void *matchinfo,
		 unsigned int matchsize,
		 unsigned int hook_mask)
{
	if (matchsize != XT_ALIGN(sizeof(struct xt_state_info)))
		return 0;

	return 1;
}

#ifdef CONFIG_COMPAT
static int compat_to_user(void *match, void **dstptr,
		int *size, int off)
{
	struct ipt_entry_match *pm;
	struct xt_state_info *pinfo;
	struct compat_xt_state_info info;
	u_int16_t msize;

	pm = (struct ipt_entry_match *)match;
	msize = pm->u.user.match_size;
	if (__copy_to_user(*dstptr, pm, sizeof(struct ipt_entry_match)))
		return -EFAULT;
	pinfo = (struct xt_state_info *)pm->data;
	memset(&info, 0, sizeof(struct compat_xt_state_info));
	info.statemask = pinfo->statemask;
	if (__copy_to_user(*dstptr + sizeof(struct ipt_entry_match),
				&info, sizeof(struct compat_xt_state_info)))
		return -EFAULT;
	msize -= off;
	if (put_user(msize, (u_int16_t *)*dstptr))
		return -EFAULT;
	*size -= off;
	*dstptr += msize;
	return 0;
}

static int compat_from_user(void *match, void **dstptr,
		int *size, int off)
{
	struct compat_ipt_entry_match *pm;
	struct ipt_entry_match *dstpm;
	struct compat_xt_state_info *pinfo;
	struct xt_state_info info;
	u_int16_t msize;

	pm = (struct compat_ipt_entry_match *)match;
	dstpm = (struct ipt_entry_match *)*dstptr;
	msize = pm->u.user.match_size;
	memcpy(*dstptr, pm, sizeof(struct compat_ipt_entry_match));
	pinfo = (struct compat_xt_state_info *)pm->data;
	memset(&info, 0, sizeof(struct xt_state_info));
	info.statemask = pinfo->statemask;
	memcpy(*dstptr + sizeof(struct compat_ipt_entry_match),
				&info, sizeof(struct xt_state_info));
	msize += off;
	dstpm->u.user.match_size = msize;
	*size += off;
	*dstptr += msize;
	return 0;
}

static int compat(void *match, void **dstptr, int *size, int convert)
{
	int ret, off;

	off = XT_ALIGN(sizeof(struct xt_state_info)) -
		COMPAT_XT_ALIGN(sizeof(struct compat_xt_state_info));
	switch (convert) {
		case COMPAT_TO_USER:
			ret = compat_to_user(match, dstptr, size, off);
			break;
		case COMPAT_FROM_USER:
			ret = compat_from_user(match, dstptr, size, off);
			break;
		case COMPAT_CALC_SIZE:
			*size += off;
			ret = 0;
			break;
		default:
			ret = -ENOPROTOOPT;
			break;
	}
	return ret;
}
#endif

static struct xt_match state_match = {
	.name		= "state",
	.match		= &match,
	.checkentry	= &check,
#ifdef CONFIG_COMPAT
	.compat		= &compat,
#endif
	.me		= THIS_MODULE,
};

static struct xt_match state6_match = {
	.name		= "state",
	.match		= &match,
	.checkentry	= &check,
#ifdef CONFIG_COMPAT
	.compat		= &compat,
#endif
	.me		= THIS_MODULE,
};

int init_xt_state(void)
{
	int ret;

	ret = xt_register_match(AF_INET, &state_match);
	if (ret < 0)
		return ret;

	ret = xt_register_match(AF_INET6, &state6_match);
	if (ret < 0)
		xt_unregister_match(AF_INET,&state_match);

	return ret;
}

void fini_xt_state(void)
{
	xt_unregister_match(AF_INET, &state_match);
	xt_unregister_match(AF_INET6, &state6_match);
}

static int __init init(void)
{
	int err;

	need_conntrack();
	err = init_xt_state();
	if (err < 0)
		return err;

	KSYMRESOLVE(init_xt_state);
	KSYMRESOLVE(fini_xt_state);
	KSYMMODRESOLVE(xt_state);
	return 0;
}

static void __exit fini(void)
{
	KSYMMODUNRESOLVE(xt_state);
	KSYMUNRESOLVE(init_xt_state);
	KSYMUNRESOLVE(fini_xt_state);
	fini_xt_state();
}

module_init(init);
module_exit(fini);
